//++++++++++++++++++++++++++++++++++++++++++++++++++++++
/*! \file	RuleEngine.cpp
 *  \brief Implementuje engine pre spracovanie pravidiel
 *  \author Rado Kostelnik
 *  \date 1.12.2001 16:00:10
 *  \version 

 *  \bug 
 */
//++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Historia zmien:
//++++++++++++++++++++++++++++++++++++++++++++++++++++++
/*
 */
//++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Rules.cpp: implementation of the CRules class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ruleparser.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRuleEngine::CRuleEngine()
{
}

CRuleEngine::~CRuleEngine()
{
	m_rules.erase( m_rules.begin(), m_rules.end() );
}

//------------------------------------------------------
/*! \fn		bool CRuleEngine::loadXML( string sFileName )
 *  \brief	 
 *	\param	filename Nazov vstupneho suboru
 *  \return Vracia true, ak sa subor nacital v poriadku
 *  \author	Rado Kostelnik
 *  \date	30.11.2001 23:36:59
 */
//------------------------------------------------------
bool CRuleEngine::LoadXML( string sFileName )
{
	CoInitializeEx( 0, COINIT_MULTITHREADED );

	IXMLDOMDocumentPtr dom;
	IXMLDOMNodePtr e;
	_bstr_t s;
	_variant_t filename;
	_variant_t retval((bool) true);

	IXMLDOMNodeListPtr l;
	IXMLDOMElementPtr el;

	dom.CreateInstance( "MSXML2.DOMDocument.3.0" );

	s = sFileName.c_str();
	filename.vt = VT_BSTR;
	filename.bstrVal = s;

	retval = dom->load( filename );
	
	if ( (bool)retval == false )
		return false;

	e = dom->documentElement;

	l = e->selectNodes( L"rule" );

	l->reset();

	el = l->nextNode(); 

	while ( el )
	{
		CRule *r = new CRule( this );
		r->set_XML ( el );
		m_rules.push_back( *r );

		el = l->nextNode(); 
	}

	if(l != NULL)
		l.Release();

	if(el != NULL)
		el.Release();

	if(e != NULL)
		e.Release();

	if(dom != NULL)
		dom.Release();

	//CoUninitialize(); //rado otipka

	return true;
}

//------------------------------------------------------
/*! \fn		TRuleVector* CRuleEngine::GetRulesByName(string sName)
 *  \brief	 
 *	\param	sName Nazov pravidla
 *  \return Zoznam vsetkych pravidiel s danym menom
 *  \author	Rado Kostelnik
 *  \date	30.11.2001 23:36:59
 */
//------------------------------------------------------
TRuleVector* CRuleEngine::GetRulesByName(string sName)
{
	TRuleVectorIterator ri = m_rules.begin();
	TRuleVectorIterator re = m_rules.end();
	TRuleVector* rules = new TRuleVector;

	for ( ; ri != re; ri ++ )
	{
		CRule& r = *ri;
		if ( r.Name() == sName )
			rules->push_back( r );
	}

	return rules;
}

//------------------------------------------------------
/*! \fn		TActionVector* CRuleEngine::Evaluate(string sRuleName, CQuery &query)
 *  \brief	Zacina vyhodnocovanie stromu pravidiel 
 *	\param	sRuleName Nazov zaciatocneho pravidla
 *	\param	query objekt, ktory vracia hodnoty premmenych
 *  \return Vracia vector akcii, ktore sa maju vykonat
 *  \author	Rado Kostelnik
 *  \date	30.11.2001 23:36:59
 */
//------------------------------------------------------
TActionVector* CRuleEngine::Evaluate(string sRuleName, CQuery &query)
{
	TRuleVectorIterator ri = m_rules.begin();
	TRuleVectorIterator re = m_rules.end();
	TActionVector* retval = NULL;

	for ( ; ri != re; ri++ )
	{
		CRule& r = *ri;
		if ( r.Name() == sRuleName )
		{
			retval = r.Evaluate( query );
			if ( retval )
				return retval;
		}

	}

	return retval;
}
