//++++++++++++++++++++++++++++++++++++++++++++++++++++++
/*! \file		HigherSkills.cpp
 *  \brief		Implementacia HigherSkills - vyssie schopnosti hraca
 *  \author		Rado Kostelnik
 *  \date		1.12.2001 17:28:47
 *  \version	0.1

 *  \bug 
 */
//++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Historia zmien:
//++++++++++++++++++++++++++++++++++++++++++++++++++++++
/*
 */
//++++++++++++++++++++++++++++++++++++++++++++++++++++++
#include "stdafx.h"
#include "higherskills.h"
#include "abstractbasicskills.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHigherSkills::CHigherSkills()
{
	char tmp[1024];
	string fname;
	GetCurrentDirectory( 1024, tmp );

	fname = tmp;
	fname += "\\rules.xml";

	m_ruleengine.LoadXML( fname );
}

//------------------------------------------------------
/*! \fn		CHigherSkills::CHigherSkills( string sFileName )
 *  \brief	Konstruktor s parametrom - nazov suboru, ktory sa ma precitat
 *  \param  string sFileName
 *  \return
 *  \author Rado Kostelnik
 *  \date	9.12.2001 22:02:41

 *  \bug 
 */
//------------------------------------------------------
// Historia zmien:
//------------------------------------------------------
/*
 */
//------------------------------------------------------
CHigherSkills::CHigherSkills( string sFileName )
{
	m_ruleengine.LoadXML( sFileName );
}

CHigherSkills::CHigherSkills( CLocalAgent* pAgent, string sFileName )
:	m_pAgent ( pAgent ), 
	m_pCondQ(new CConditionQuery(pAgent)), 
	m_pParamQ(new CParameterQuery(pAgent))
{
	m_ruleengine.LoadXML( sFileName );
}

CHigherSkills::~CHigherSkills()
{
	if(m_pCondQ != NULL)
	{
		delete m_pCondQ;
		m_pCondQ = NULL;
	}
	if(m_pParamQ != NULL)
	{
		delete m_pParamQ;
		m_pParamQ = NULL;
	}
}

//------------------------------------------------------
/*! \fn			void CHigherSkills::deliberate( CLocalAgent* p )
 *  \brief		Rozhodovanie o nasledujucej cinnosti na vyssej urovni
 *  \param		CLocalAgent* p agent, ktoremu urcujeme cinnosti
 *  \return		nic
 *  \author		Rado Kostelnik
 *  \date		1.12.2001 17:16:09

 *  \bug 
 */
//------------------------------------------------------
// Historia zmien:
//------------------------------------------------------
/*
 */
//------------------------------------------------------
void CHigherSkills::Deliberate( )
{
	TActionVector* av;
	CBasicSkillList* sl = m_pAgent->listSkl;  
	CBasicSkill* s = NULL;
	TParameterBSkillsVector v;
	int priority = 0;
	string tmps;
	char buf[100];

	av = m_ruleengine.Evaluate( "default", *m_pCondQ );

	if ( av )
	{
		TActionVectorIterator ai = av->begin();
		TActionVectorIterator ae = av->end();

		for ( ; ai != ae; ai++ )
		{
			CAction& a = *ai;
			//
			// Vyhodnotenie realnych hodnot parametrov akcii
			//
			TRuleParameterVectorIterator pi = a.parameters().begin();
			TRuleParameterVectorIterator pe = a.parameters().end();

			v.clear();
			for ( ; pi != pe; pi++ )
			{
				CRuleParameter& p = *pi;
				
				long pos;

				// location
				if ( p.type == "location" )
				{
					tmps = m_pParamQ->getValue( p.value );
					pos = tmps.find( " " );
					v.push_back( CParameter( VALUE_X, tmps.substr( 0, pos ) ) );
					v.push_back( CParameter( VALUE_Y, tmps.substr( pos+1 ) ) );
				}
				// playerid
				else if ( p.type == "playerid" )
				{
					tmps = m_pParamQ->getValue( p.value );
					v.push_back ( CParameter( VALUE_PLAYER, tmps ) );
				}
				// effort
				else if ( p.type == "effort" )
				{
					if ( p.value == "Maximum" )
						priority = 3;
					else if ( p.value == "Medium" )
						priority = 2;
					else if ( p.value == "Minimum" )
						priority = 1;
				}
				// angle
				else if ( p.type == "angle" )
				{
					//tmps = m_paramq.getValue( p.value );
					v.push_back ( CParameter( VALUE_ANGLE, p.value ) );
				}
				else
					cout << "Error: Unknown type: " << p.type << endl;
			}

			//
			// Kick
			//
			if ( a.name == "Kick" )
			{
				s = new CKick( v, m_pAgent, priority );
			}
			else
			//
			// GoToPosition
			//
			if ( a.name == "GoToPosition" )
			{
				
				sprintf( buf, "%d", BSKILL_MOVE_GOTO );
				tmps = buf;
				v.push_back ( CParameter( VALUE_TYPE, tmps ) );
				s = new CMove( v, m_pAgent, priority );
			}
			else 
			//
			// DribbleTo
			//
			if ( a.name == "DribbleTo" )
			{
				s = new CDribble( v, m_pAgent, priority );
			}
			else 
			//
			// TurnTo
			//
			if ( a.name == "Turn" )
			{
				s = new CTurn( v, m_pAgent, priority );
			}
			else
			//
			// PassTo
			//
			if ( a.name == "PassTo" )
			{
				s = new CPass( v, m_pAgent, priority );
			}
			else 
			//
			// Catch
			//
			if ( a.name == "Catch" )
			{
				s = new CCatch( v, m_pAgent, priority );
			}

			if ( s )
			{
				sl->AddSkill( s );
				TRACE("Adding action: %s", a.name.c_str() );
			}
			s = NULL;
		
		}
	}
}