// ConditionQuery.cpp: implementation of the CConditionQuery class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ruleparser.h"
#include "conditionquery.h"
#include "worldobjects.h"
#include "worldobjects.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CConditionQuery::CConditionQuery(CLocalAgent *pAgent) : CQuery()
{
	addTerm( new CHasBallTerm(pAgent) );
	addTerm( new CNearGoalTerm(pAgent) );
	addTerm( new CFreePathToGoalTerm(pAgent) );
	addTerm( new CHasFreeTeamMateTerm(pAgent) );
	addTerm( new CCanIPassTerm(pAgent) );
	addTerm( new CHasBetterLocatedTeamMateTerm(pAgent) );
	addTerm( new CSeeBallTerm(pAgent) );
	addTerm( new CSomeoneHasBallTerm(pAgent) );
	addTerm( new CTeamMateHasBallTerm(pAgent) );
	addTerm( new CAmIFreeTerm(pAgent) );
	addTerm( new CCanIGetPassTerm(pAgent) );
	addTerm( new CBallKickableTerm(pAgent) );
	addTerm( new CBallNearTerm(pAgent) );
	addTerm( new CTeamMateNearBallTerm(pAgent) );
}

string CHasBallTerm::Evaluate()
{
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pOpponent = NULL;
    
	double dBallDist = m_pAgent->bskl->RelVectorTo(
		m_pAgent, 
		m_pAgent->GetWorldModel()->GetBall()->GetPosition()
	).GetDistance();

    double dOpponentBallDist;
	double dBestOpponentBallDist = MAX_BALL_DIST;
	double dStepDist;
	double dOpponentStepDist;

	if(dBallDist < m_pAgent->cfg->kickable_area)
		return KPDL_TRUE;

	// Vzdialenost lopty od superov
	for(int i=0; i<iPlayerCount; i++)
	{
		pOpponent = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		if(pOpponent->GetPlayerSide() == S_OPPONENT)
		{
			dOpponentBallDist = m_pAgent->bskl->RelVectorTo(
				pOpponent, 
				m_pAgent->GetWorldModel()->GetBall()->GetPosition()
			).GetDistance();

			if(dBestOpponentBallDist < dOpponentBallDist)
				dBestOpponentBallDist = dOpponentBallDist;
		}
	}

	// Dalej predpokladajme, ze hrac sa bude k lopte priblizovat priamo,
	// maximalnou rychlostou
	dStepDist = dBallDist / m_pAgent->cfg->player_speed_max;
	dOpponentStepDist = dBestOpponentBallDist / m_pAgent->cfg->player_speed_max;

	// Ak je blizsie super
	if(dOpponentStepDist < dStepDist && 
		dOpponentStepDist < MAX_BALL_CONTROL_DIST)
		return KPDL_FALSE;

	// Ak je blizsie hrac
	if(dOpponentStepDist > dStepDist && 
		dStepDist < MAX_BALL_CONTROL_DIST)
		return KPDL_TRUE;

	// nejaka funkcia vhodne ohodnocujuca zvysne pripady
	string sResult;
	sResult = (dOpponentStepDist / (dOpponentStepDist + dStepDist));
	return sResult;
	// ! mozno vhodne pridat este clen - cim dalej je lopta od hraca,
	// tym mensia kontrola
}

string CNearGoalTerm::Evaluate()
{
	string sResult;
	sResult = 
		((m_pAgent->GetPosition().GetX() - 
		  m_pAgent->field->GetGoal(GT_HOME).GetX()) / 
		 (m_pAgent->field->GetFieldLength()));
	return sResult;
}

string CFreePathToGoalTerm::Evaluate()
{
	C2DVector cUpperGoalFlag = m_pAgent->field->GetGoal(GT_HOME);
	C2DVector cLowerGoalFlag = m_pAgent->field->GetGoal(GT_HOME);
	CAgent *pOpponent = NULL;

	cUpperGoalFlag.SetY(- m_pAgent->field->GetGoalSize() / 2.0);
	cLowerGoalFlag.SetY(+ m_pAgent->field->GetGoalSize() / 2.0);

	return "";
}

string CHasFreeTeamMateTerm::Evaluate()
{
    double dOpponentDist;
	double dNormOpponentDist;
	double dWorstNormOpponentDist = 0.0;
	double dBestNormOpponentDist = 1.0;
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pOpponent = NULL;
	CAgent *pTeamMate = NULL;

	for(int i=0; i<iPlayerCount; i++)
	{
		pTeamMate = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		if(pTeamMate->GetPlayerSide() == S_HOME &&
			pTeamMate->GetLocalNumber() != m_pAgent->GetLocalNumber())
		{
			dBestNormOpponentDist = 1.0;
			for(int j=0; j<iPlayerCount; j++)
			{
				pOpponent = m_pAgent->GetWorldModel()->GetIdentifiedAgent(j);

				if(pOpponent->GetPlayerSide() == S_OPPONENT)
				{
					dOpponentDist = m_pAgent->bskl->RelVectorTo(
						pTeamMate, 
						pOpponent->GetPosition()
					).GetDistance();

					dNormOpponentDist = dOpponentDist / MAX_OPPONENT_DIST;

					if(dBestNormOpponentDist < dNormOpponentDist)
						dBestNormOpponentDist = dNormOpponentDist;
				}
			}

			if(dWorstNormOpponentDist < dBestNormOpponentDist)
				dWorstNormOpponentDist = dBestNormOpponentDist;
		}
	}

	string sResult;
	sResult = 1.0 - dWorstNormOpponentDist;
	return sResult;
}

string CCanIPassTerm::Evaluate()
{
    double dOpponentDist;
	double dNormOpponentDist;
	double dWorstNormOpponentDist = 0.0;
	double dBestNormOpponentDist = 1.0;
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pOpponent = NULL;
	CAgent *pTeamMate = NULL;

	for(int i=0; i<iPlayerCount; i++)
	{
		pTeamMate = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		if(pTeamMate->GetPlayerSide() == S_HOME &&
			pTeamMate->GetLocalNumber() != m_pAgent->GetLocalNumber())
		{
			dBestNormOpponentDist = 1.0;
			for(int j=0; j<iPlayerCount; j++)
			{
				pOpponent = m_pAgent->GetWorldModel()->GetIdentifiedAgent(j);

				if(pOpponent->GetPlayerSide() == S_OPPONENT)
				{
					dOpponentDist = m_pAgent->bskl->RelVectorTo(
						pTeamMate, 
						pOpponent->GetPosition()
					).GetDistance();

					dNormOpponentDist = dOpponentDist / MAX_OPPONENT_DIST;

					if(dBestNormOpponentDist < dNormOpponentDist)
						dBestNormOpponentDist = dNormOpponentDist;
				}
			}

			if(dWorstNormOpponentDist < dBestNormOpponentDist)
			{
				// ! Tu treba este skontrolovat, nakolko je volna draha lopty
				// k najdenemu spoluhracovi
				dWorstNormOpponentDist = dBestNormOpponentDist;
			}
		}
	}

	string sResult;
	sResult = 1.0 - dWorstNormOpponentDist;
	return sResult;
}

string CHasBetterLocatedTeamMateTerm::Evaluate()
{

	return "";
}

string CSeeBallTerm::Evaluate()
{
	double dPI2 = PI / 2;
	double dBallAngle = m_pAgent->GetWorldModel()->GetBall()->GetSpeed().GetAngle();

	// ! Dufam, ze last seen je pocet cyklov, kedy bola lopta naposledy videna, nie
	// je to uplne korektny test ...
	if(dBallAngle > -dPI2 && dBallAngle < dPI2 && 
		m_pAgent->GetWorldModel()->GetBall()->GetLastSeen() < MAX_LAST_SEEN)
		return KPDL_TRUE;

	return KPDL_FALSE;
}

string CSomeoneHasBallTerm::Evaluate()
{
	int i;
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pPlayer = NULL;

	double dBallSpeed = m_pAgent->GetWorldModel()->GetBall()->GetSpeed().GetDistance();
	double dBallDist;
	double dStepDist;
	double dBestStepDist = MAX_BALL_DIST; 
	
	for(i=0; i<iPlayerCount; i++)
	{
		pPlayer = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		dBallDist = m_pAgent->bskl->RelVectorTo(
			pPlayer, 
			m_pAgent->GetWorldModel()->GetBall()->GetPosition()
		).GetDistance();

		// ! Zakomponovat do tohto vypoctu aj rychlost lopty, resp
		// vypocet timu cislo 4, o aky cas dobehne hrac loptu ...
		dStepDist = dBallDist / m_pAgent->cfg->player_speed_max;

		if(dBallDist < m_pAgent->cfg->kickable_area ||
			dStepDist < MAX_BALL_CONTROL_DIST)
			return KPDL_TRUE;

		// vezmeme najlepsi vysledok
		if(dBestStepDist < dBallDist)
			dBestStepDist = dBallDist;
	}

	iPlayerCount = m_pAgent->GetWorldModel()->GetUnIdentifiedAgentsCount();
	for(i=0; i<iPlayerCount; i++)
	{
		pPlayer = m_pAgent->GetWorldModel()->GetUnIdentifiedAgent(i);

		dBallDist = m_pAgent->bskl->RelVectorTo(
			pPlayer, 
			m_pAgent->GetWorldModel()->GetBall()->GetPosition()
		).GetDistance();

		// ! Zakomponovat do tohto vypoctu aj rychlost lopty, resp
		// vypocet timu cislo 4, o aky cas dobehne hrac loptu ...
		dStepDist = dBallDist / m_pAgent->cfg->player_speed_max;

		if(dBallDist < m_pAgent->cfg->kickable_area ||
			dStepDist < MAX_BALL_CONTROL_DIST)
			return KPDL_TRUE;

		// vezmeme najlepsi vysledok
		if(dBestStepDist < dBallDist)
			dBestStepDist = dBallDist;
	}

	string sResult;
	sResult = (1 - dBestStepDist / MAX_BALL_DIST);
	return sResult;
}

string CTeamMateHasBallTerm::Evaluate()
{
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pTeamMate = NULL;
	CAgent *pOpponent = NULL;

	double dBallDist;
	double dOpponentBallDist;
	double dBestOpponentBallDist = MAX_BALL_DIST;
	double dStepDist;
	double dOpponentStepDist;
	double dResult;
	double dBestResult = 0.0; // pesimistic view

	for(int i=0; i<iPlayerCount; i++)
	{
		pTeamMate = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		dBallDist = m_pAgent->bskl->RelVectorTo(
			m_pAgent, 
			m_pAgent->GetWorldModel()->GetBall()->GetPosition()
		).GetDistance();
		
		if(dBallDist < m_pAgent->cfg->kickable_area)
			return KPDL_TRUE;

		// Dalej predpokladajme, ze hrac sa bude k lopte priblizovat priamo,
		// maximalnou rychlostou
		dStepDist = dBallDist / m_pAgent->cfg->player_speed_max;

		if(pTeamMate->GetPlayerSide() == S_HOME &&
			pTeamMate->GetLocalNumber() != m_pAgent->GetLocalNumber())
		{
			// Vzdialenost lopty od superov
			for(int i=0; i<iPlayerCount; i++)
			{
				pOpponent = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

				if(pOpponent->GetPlayerSide() == S_OPPONENT)
				{
					dOpponentBallDist = m_pAgent->bskl->RelVectorTo(
						pOpponent, 
						m_pAgent->GetWorldModel()->GetBall()->GetPosition()
					).GetDistance();

					if(dBestOpponentBallDist < dOpponentBallDist)
						dBestOpponentBallDist = dOpponentBallDist;
				}
			}

			dOpponentStepDist = dBestOpponentBallDist / m_pAgent->cfg->player_speed_max;

			// Ak je blizsie super
			if(dOpponentStepDist < dStepDist && 
				dOpponentStepDist < MAX_BALL_CONTROL_DIST)
			{
				dResult = 0.0;
				break;
			}

			// Ak je blizsie spoluhrac
			if(dOpponentStepDist > dStepDist && 
				dStepDist < MAX_BALL_CONTROL_DIST)
			{
				dResult = 1.0;
				break;
			}

			// nejaka funkcia vhodne ohodnocujuca zvysne pripady
			dResult = dOpponentStepDist / (dOpponentStepDist + dStepDist);
			// ! mozno vhodne pridat este clen - cim dalej je lopta od hraca,
			// tym mensia kontrola
		}

		// vezmeme najlepsi vysledok
		if(dBestResult < dResult)
			dBestResult = dResult;
	}

	string sResult;
	sResult = dBestResult;
	return sResult;
}

string CAmIFreeTerm::Evaluate()
{
    double dOpponentDist;
	double dNormOpponentDist;
	double dBestNormOpponentDist = 1.0;
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pOpponent = NULL;

	for(int i=0; i<iPlayerCount; i++)
	{
		pOpponent = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		if(pOpponent->GetPlayerSide() == S_OPPONENT)
		{
			dOpponentDist = m_pAgent->bskl->RelVectorTo(
				m_pAgent, 
				pOpponent->GetPosition()
			).GetDistance();

			dNormOpponentDist = dOpponentDist / MAX_OPPONENT_DIST;

			if(dBestNormOpponentDist < dNormOpponentDist)
				dBestNormOpponentDist = dNormOpponentDist;
		}
	}

	string sResult;
	sResult = (1.0 - dBestNormOpponentDist);
	return sResult;
}

string CCanIGetPassTerm::Evaluate()
{
    double dOpponentDist;
	double dNormOpponentDist;
	double dBestNormOpponentDist = 1.0;
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pOpponent = NULL;

	for(int i=0; i<iPlayerCount; i++)
	{
		pOpponent = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		if(pOpponent->GetPlayerSide() == S_OPPONENT)
		{
			dOpponentDist = m_pAgent->bskl->RelVectorTo(
				m_pAgent, 
				pOpponent->GetPosition()
			).GetDistance();

			dNormOpponentDist = dOpponentDist / MAX_OPPONENT_DIST;

			if(dBestNormOpponentDist < dNormOpponentDist)
				dBestNormOpponentDist = dNormOpponentDist;
		}
	}

	// ! ak som volny, treba este skontrolovat, ci medzi mnou a spoluhracom s loptou
	// je volna cesta
	string sResult;
	sResult = (1.0 - dBestNormOpponentDist);
	return sResult;
}

string CBallKickableTerm::Evaluate()
{
    double dBallDist = m_pAgent->bskl->RelVectorTo(
		m_pAgent, 
		m_pAgent->GetWorldModel()->GetBall()->GetPosition()
	).GetDistance();
    
	if (dBallDist < m_pAgent->cfg->kickable_area)
		return KPDL_TRUE;

    return KPDL_FALSE;
}

string CBallNearTerm::Evaluate()
{
    double dBallDist = m_pAgent->bskl->RelVectorTo(
		m_pAgent, 
		m_pAgent->GetWorldModel()->GetBall()->GetPosition()
	).GetDistance();

	double dNormBallDist = dBallDist / MAX_BALL_DIST;
	
	if(dNormBallDist < 1.0)
	{
		string sResult;
		sResult = (1.0 - dNormBallDist);
		return sResult;
	}
	
	// Ball is too far
	return "0.0";
}

string CTeamMateNearBallTerm::Evaluate()
{
    double dBallDist;
	double dNormBallDist;
	double dBestNormBallDist = 1.0;
	int iPlayerCount = m_pAgent->GetWorldModel()->GetIdentifiedAgentsCount();
	CAgent *pTeamMate = NULL;

	for(int i=0; i<iPlayerCount; i++)
	{
		pTeamMate = m_pAgent->GetWorldModel()->GetIdentifiedAgent(i);

		if(pTeamMate->GetPlayerSide() == S_HOME &&
			pTeamMate->GetLocalNumber() != m_pAgent->GetLocalNumber())
		{
			dBallDist = m_pAgent->bskl->RelVectorTo(
				pTeamMate, 
				m_pAgent->GetWorldModel()->GetBall()->GetPosition()
			).GetDistance();

			dNormBallDist = dBallDist / MAX_BALL_DIST;

			if(dBestNormBallDist > dNormBallDist)
				dBestNormBallDist = dNormBallDist;
		}
	}

	string sResult;
	sResult = (1.0 - dBestNormBallDist);
	return sResult;
}
