// ConditionQuery.h: interface for the CConditionQuery class
//
//////////////////////////////////////////////////////////////////////

#ifndef _CONDITIONQUERY_H_
#define _CONDITIONQUERY_H_

class CConditionQuery;

#include "ruleparser.h"
#include "worldobjects.h"

#define MAX_BALL_CONTROL_DIST  5      // krokov
#define MAX_BALL_DIST          50.0   // metrov
#define MAX_OPPONENT_DIST      10.0   // metrov
#define MAX_LAST_SEEN          2      // cykly
#define KPDL_TRUE              "1.0"
#define KPDL_FALSE             "0.0"

class CConditionQuery : public CQuery
{
public:
	CConditionQuery( CLocalAgent *pAgent = NULL );
};


class CHasBallTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CHasBallTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "HasBall";
	};

	string Evaluate();
};

class CNearGoalTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CNearGoalTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "NearGoal";
	};

	string Evaluate();
};

class CFreePathToGoalTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CFreePathToGoalTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "FreePathToGoal";
	};

	string Evaluate();
};

class CHasFreeTeamMateTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CHasFreeTeamMateTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "HasFreeTeamMate";
	};

	string Evaluate();
};

class CCanIPassTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CCanIPassTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "CanIPass";
	};

	string Evaluate();
};

class CHasBetterLocatedTeamMateTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CHasBetterLocatedTeamMateTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "HasBetterLocatedTeamMate";
	};

	string Evaluate();
};

class CSeeBallTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CSeeBallTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "SeeBall";
	};

	string Evaluate();
};

class CSomeoneHasBallTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CSomeoneHasBallTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "SomeoneHasBall";
	};

	string Evaluate();
};

class CTeamMateHasBallTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CTeamMateHasBallTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "TeamMateHasBall";
	};

	string Evaluate();
};

class CAmIFreeTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CAmIFreeTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "AmIFree";
	};

	string Evaluate();
};

class CCanIGetPassTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CCanIGetPassTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "CanIGetPass";
	};

	string Evaluate();
};

class CBallKickableTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CBallKickableTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "BallKickable";
	};

	string Evaluate();
};

class CBallNearTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CBallNearTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "BallNear";
	};

	string Evaluate();
};

class CTeamMateNearBallTerm : public CTerm
{
private:
	CLocalAgent *m_pAgent;

public:
	CTeamMateNearBallTerm( CLocalAgent *pAgent )
	:	m_pAgent(pAgent)
	{
		name = "TeamMateNearBall";
	};

	string Evaluate();
};

#endif // _CONDITIONQUERY_H_
