/* -*- Mode: C -*-
 *Header:
 *File: field.C
 *Author: Noda Itsuki
 *Date: 1996/02/14
 *EndHeader:
 */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *EndCopyright:
 */

#include <winsock.h>

#undef V_INT
#undef V_BOOL
#undef V_NONE
#undef INFINITE
#undef max
#undef min

#include <conio.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <strstream>
#include <errno.h>
using namespace std;

#include "param.h"
#include "types.h"
#include "utility.h"
#include "object.h"
#include "field.h"
#include "coach_lang.h"
#include "garbage.h"

CM_Message* g_pMess;
GarbageCollector g_GC;

static UINT uPeriod = TIMEDELTA; //presnost casovaca
/*
 *===================================================================
 *Part: Create Stadium Window
 *===================================================================
 */
void Stadium::init(int argc, char *argv[])
{
	WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 0), &wsaData))
    {
		cerr << "Error: Windows Sockets initialization failed." << endl;
		exit(EXIT_FAILURE);
    }

	TIMECAPS tc;
	timeGetDevCaps(&tc,sizeof(tc));
	if(uPeriod < tc.wPeriodMin)
	{
		uPeriod = tc.wPeriodMin;
	}
	else if(uPeriod > tc.wPeriodMax)
	{
		uPeriod = tc.wPeriodMax;
	}

	timeBeginPeriod(uPeriod);

	int i ;

    server_param = new ServerParam ( argc, argv );
    if ( server_param == NULL )
	{
	perror ( "Can not create server parameters" );
	exit ( EXIT_FAILURE );
	}

    player_param = new PlayerParam ( argc, argv );
    if ( player_param == NULL )
	{
	perror ( "Can not create player parameters" );
	exit ( EXIT_FAILURE );
	}

    player_types = new HetroPlayer* [ player_param->playerTypes () ];
    if ( player_types == NULL )
      {
	perror ( "Can not create hetro_player array" );
	exit ( EXIT_FAILURE );
      }
    
    player_types[0] = new HetroPlayer ( server_param, player_param, 0 );
    //cout << *(player_types[i]) << endl;

    for (i = 1; i < player_param->playerTypes (); i++ )
      {
	player_types[i] = new HetroPlayer ( server_param, player_param );
	//cout << *(player_types[i]) << endl;
      }
    
    if (server_param->logfile[0] != NULLCHAR) {
    if ((log_fp = fopen(server_param->logfile, "w")) == NULL)
	cerr << "can't open log_file " << server_param->logfile << endl ;
    }

    if (server_param->recfile[0] != NULLCHAR) 
	OpenForWrite() ;

    weather.init(this) ;
    assign() ;
    assignBall() ;
    set_ball(PVector(0.0,0.0),LEFT) ;
    //this->startp = FALSE ;
    this->startp = 0;
    caught_ball = FALSE ;
    player = new Player*[MAX_PLAYER * 2] ;
    olcoach = new OnlineCoach*[3] ;

    initPlayer() ;
    initCoach(this) ;
    initOnlineCoach(this) ;
    last_touch = player[0];
    time = 0;
    team_l->point = server_param->start_goal_l;
    team_r->point = server_param->start_goal_r;
    top.next = NULL ;

    change_play_mode(PM_BeforeKickOff) ;

    for ( i = 0 ; i <= MAX_PLAYER * 2; i++)
	socketfd[i] = -1 ;
    for ( i = 0 ; i <= MAX_PLAYER * 2; i++)
	dinfo.body.show.pos[i].enable = htons(FALSE) ;
    dinfo.mode = htons(SHOW_MODE) ;
    dinfo.body.show.team[0].name[0] = NULLCHAR ;
    dinfo.body.show.team[1].name[0] = NULLCHAR ;
    dinfo.body.show.time = htons((unsigned short)-1) ;
    minfo.mode = htons(MSG_MODE) ;

    for ( i = 0 ; i < MAX_PLAYER * 2; i++)
	dinfo2.body.show.pos[i].mode = htons(FALSE) ;
    dinfo2.mode = htons(SHOW_MODE) ;
    dinfo2.body.show.team[0].name[0] = NULLCHAR ;
    dinfo2.body.show.team[1].name[0] = NULLCHAR ;
    dinfo2.body.show.time = htons((unsigned short)-1) ;

    randomize() ;

    //[2000.01.02:I.Noda]  exit when udp_init failes.
    if(udp_init(0, server_param->portnum)) {
       cerr << "Error: udp_init():fail" << strerror(errno) << endl ;
       exit(1) ;
    }
}

void Stadium::clean ()
{
  if ( server_param != NULL )
    {
      delete server_param;
      server_param = NULL;
    }

  if ( player_types != NULL )
    {
      for ( int i = 0; i < player_param->playerTypes (); i++ )
	{
	  if ( player_types[i] != NULL )
	    {
	      delete player_types[i];
	      player_types[i] = NULL;
	    }
	}
      delete[] player_types;
      player_types = NULL;
    }

  timeEndPeriod(uPeriod);
  WSACleanup();

}

int gotsig = FALSE;
int timedelta = 0;
int lock_timedelta = FALSE;


void check(void)
{
	static int td_mult = 1;
	if (lock_timedelta) {
		td_mult += 1;
	}
	else {
		timedelta += td_mult * TIMEDELTA;
		td_mult = 1;
	}
	gotsig = TRUE;
}

void Stadium::reset_player_comm_flag()
{
	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2; i++){
		player[i]->command_done = FALSE ;
		player[i]->turn_neck_done = FALSE ;
	}
}
void CALLBACK TimeProc(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	check();
	SetEvent((HANDLE) dwUser);
}

void Stadium::MainLoop()
{
	// th 6.3.00
	unsigned long tp_new, tp_old;
	tp_new = GetTickCount();
	//
	
	int lcmt = 0, c_simt = 1, c_sent = 1, c_rect = 1, c_sbt = 1, c_svt = 1,
	  q_simt = server_param->lcm_st / server_param->sim_st, q_sent = server_param->lcm_st / server_param->send_st * 4,
	  q_rect = server_param->lcm_st / server_param->recv_st, q_sbt = server_param->lcm_st / server_param->sb_step, q_svt = server_param->lcm_st / server_param->sv_st;

	printf("Press any key to stop the simulation\n");
	HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	UINT wTimerID = timeSetEvent(TIMEDELTA, 0, TimeProc, (DWORD) hEvent, TIME_PERIODIC);


	while (!kbhit()) {
		if (!gotsig)
			WaitForSingleObject(hEvent, INFINITE);
		gotsig = false;
		
		lock_timedelta = TRUE;
		lcmt += timedelta;
		timedelta = 0;
		lock_timedelta = FALSE;
		
		if (lcmt >= server_param->sim_st * c_simt) {
		        c_simt = (int) (floor (lcmt / server_param->sim_st) );
  			lcmt = server_param->sim_st * c_simt;	 
				// the above lines are needed to prevent short "catch up" cycles if the current cycle was slow	
		}
		
		if (lcmt >= server_param->recv_st * c_rect) {
		        udp_recv_message() ;
			udp_recv_from_online_coach() ;
			if (server_param->CoachMode || server_param->CwRMode)
			        udp_recv_from_coach();
			c_rect = (int) (floor (lcmt / server_param->recv_st) );
			if (q_rect <= c_rect)
			  c_rect = 1;
			else
			  c_rect++;
		}
		
		if (lcmt >= server_param->sim_st * c_simt) {
		        // th 6.3.00
		        if (server_param->logfile[0] != NULLCHAR && server_param->log_times)
			{
			  tp_old = tp_new;  
			  tp_new = GetTickCount();
			  write_times(tp_new, tp_old);
			}
			//
		        reset_player_comm_flag() ;
		        step() ;
			if (q_simt <= c_simt)
			  c_simt = 1;
			else
			  c_simt++;
		}

		if (lcmt >= server_param->sb_step * c_sbt) {
		        int i ;
			for (i = 0 ; i < MAX_PLAYER * 2; i++)
			  if (player[i]->version >= 5.0) {
			    player[i]->sense_body() ;
			    if ((player[i]->team->side == LEFT &&
				 server_param->fullstate_l) ||
				(player[i]->team->side == RIGHT &&
				 server_param->fullstate_r)) {
			      player[i]->send_fullstate_information();
			    }
                          }
			c_sbt = (int) (floor (lcmt / server_param->sb_step ) );
			if (q_sbt <= c_sbt)
			  c_sbt = 1;
			else
			  c_sbt++;
		} 

		if (lcmt >= (server_param->send_st / 4) * c_sent) {
		        udp_send_message() ;
			c_sent = (int) (floor (lcmt / (server_param->send_st / 4)) );
			if (q_sent <= c_sent)
			  c_sent = 1;
			else
			  c_sent++;
		}
				
		if (lcmt >= server_param->sv_st * c_svt) {
		        int i ;
			if (coach->assignedp && coach->eye)
				coach->send_visual_info() ;
			for (i = 1 ; i < 3; i++)
				if (olcoach[i]->assignedp && olcoach[i]->eye)
					olcoach[i]->send_visual_info() ;
			c_svt = (int) (floor (lcmt / server_param->sv_st) );
			if (q_svt <= c_svt)
			  c_svt = 1;
			else
			  c_svt++;
		}
		
		if (lcmt >= server_param->lcm_st)
		  lcmt = 0;
	}

	timeKillEvent(wTimerID);
	CloseHandle(hEvent);
	getch();

	if (server_param->logfile[0] != NULLCHAR)
		fclose(log_fp) ;
	if (server_param->recfile[0] != NULLCHAR)
		fclose(rec_fp) ;
}

//pfr:SYNCH


void Stadium::assign(void)
{
	field.assign(this) ;

	votable.AssignObject(&(field.goal_l)) ;
	votable.AssignObject(&(field.goal_r)) ;

	/* pfr 06/07/200 added short name support */
	votable.AssignObject(new PObject(FLAG_C_NAME, FLAG_C_NAME_SHORT,
			PVector(0.0, 0.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_CT_NAME, FLAG_CT_NAME_SHORT,
			PVector(0.0, -PITCH_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_CB_NAME, FLAG_CB_NAME_SHORT,
			PVector(0.0, PITCH_WIDTH/2.0))) ;

	votable.AssignObject(new PObject(FLAG_LT_NAME, FLAG_LT_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0, -PITCH_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_LB_NAME, FLAG_LB_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0, PITCH_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_RT_NAME, FLAG_RT_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0, -PITCH_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_RB_NAME, FLAG_RB_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0, PITCH_WIDTH/2.0))) ;

	votable.AssignObject(new PObject(FLAG_PLT_NAME, FLAG_PLT_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0+PENALTY_AREA_LENGTH,
				-PENALTY_AREA_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_PLC_NAME, FLAG_PLC_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0+PENALTY_AREA_LENGTH, 0))) ;
	votable.AssignObject(new PObject(FLAG_PLB_NAME, FLAG_PLB_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0+PENALTY_AREA_LENGTH,
				PENALTY_AREA_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_PRT_NAME, FLAG_PRT_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0-PENALTY_AREA_LENGTH,
				-PENALTY_AREA_WIDTH/2.0))) ;
	votable.AssignObject(new PObject(FLAG_PRC_NAME, FLAG_PRC_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0-PENALTY_AREA_LENGTH, 0))) ;
	votable.AssignObject(new PObject(FLAG_PRB_NAME, FLAG_PRB_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0-PENALTY_AREA_LENGTH,
				PENALTY_AREA_WIDTH/2.0))) ;

	votable.AssignObject(new PObject(FLAG_GLT_NAME, FLAG_GLT_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0, -server_param->gwidth/2.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_GLB_NAME, FLAG_GLB_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0, server_param->gwidth/2.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_GRT_NAME, FLAG_GRT_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0, -server_param->gwidth/2.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_GRB_NAME, FLAG_GRB_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0, server_param->gwidth/2.0), 4.0)) ;

	votable.AssignObject(new PObject(FLAG_TL50_NAME, FLAG_TL50_NAME_SHORT,
			PVector(-50.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TL40_NAME, FLAG_TL40_NAME_SHORT,
			PVector(-40.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TL30_NAME, FLAG_TL30_NAME_SHORT,
			PVector(-30.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TL20_NAME, FLAG_TL20_NAME_SHORT,
			PVector(-20.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TL10_NAME, FLAG_TL10_NAME_SHORT,
			PVector(-10.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_T0_NAME, FLAG_T0_NAME_SHORT,
			PVector(0.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TR10_NAME, FLAG_TR10_NAME_SHORT,
			PVector(10.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TR20_NAME, FLAG_TR20_NAME_SHORT,
			PVector(20.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TR30_NAME, FLAG_TR30_NAME_SHORT,
			PVector(30.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TR40_NAME, FLAG_TR40_NAME_SHORT,
			PVector(40.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_TR50_NAME, FLAG_TR50_NAME_SHORT,
			PVector(50.0, -PITCH_WIDTH/2.0-5.0), 4.0)) ;

	votable.AssignObject(new PObject(FLAG_BL50_NAME, FLAG_BL50_NAME_SHORT,
			PVector(-50.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BL40_NAME, FLAG_BL40_NAME_SHORT,
			PVector(-40.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BL30_NAME, FLAG_BL30_NAME_SHORT,
			PVector(-30.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BL20_NAME, FLAG_BL20_NAME_SHORT,
			PVector(-20.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BL10_NAME, FLAG_BL10_NAME_SHORT,
			PVector(-10.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_B0_NAME, FLAG_B0_NAME_SHORT,
			PVector(0.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BR10_NAME, FLAG_BR10_NAME_SHORT,
			PVector(10.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BR20_NAME, FLAG_BR20_NAME_SHORT,
			PVector(20.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BR30_NAME, FLAG_BR30_NAME_SHORT,
			PVector(30.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BR40_NAME, FLAG_BR40_NAME_SHORT,
			PVector(40.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_BR50_NAME, FLAG_BR50_NAME_SHORT,
			PVector(50.0, PITCH_WIDTH/2.0+5.0), 4.0)) ;

	votable.AssignObject(new PObject(FLAG_LT30_NAME, FLAG_LT30_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, -30), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_LT20_NAME, FLAG_LT20_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, -20), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_LT10_NAME, FLAG_LT10_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, -10), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_L0_NAME, FLAG_L0_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, 0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_LB10_NAME, FLAG_LB10_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, 10), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_LB20_NAME, FLAG_LB20_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, 20), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_LB30_NAME, FLAG_LB30_NAME_SHORT,
			PVector(-PITCH_LENGTH/2.0-5.0, 30), 4.0)) ;

	votable.AssignObject(new PObject(FLAG_RT30_NAME, FLAG_RT30_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, -30), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_RT20_NAME, FLAG_RT20_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, -20), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_RT10_NAME, FLAG_RT10_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, -10), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_R0_NAME, FLAG_R0_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, 0), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_RB10_NAME, FLAG_RB10_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, 10), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_RB20_NAME, FLAG_RB20_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, 20), 4.0)) ;
	votable.AssignObject(new PObject(FLAG_RB30_NAME, FLAG_RB30_NAME_SHORT,
			PVector(PITCH_LENGTH/2.0+5.0, 30), 4.0)) ;
}

/* pfr 06/07/200 added short name support */
MPObject* Stadium::NewObject(
							const MPObjectType t,
							const Name name,
							const Name short_name,
							const PVector pos,
							const Value size,
							const Value angle,
							const Value decay,
							const Value rand,
							const Value weight,
							const Value speed,
							// th 6.3.00
							const Value accel)
                                                        //
{
	MPObject* obj = motable.NewObject(name, short_name) ;

	votable.AssignObject(obj) ;
	obj->stadium = this ;
	obj->weather = &weather ;
	obj->weight = weight ;
	obj->max_speed = speed ;
	// th 6.3.00
	obj->max_accel = accel ;
	//
	obj->Set(t,pos,size,angle,PVector(0.0,0.0),PVector(0.0,0.0), decay,rand) ;
	return obj ;
}

void Stadium::assignBall(void)
{
	ball = NewObject(
			MPO_Ball,
			BALL_NAME,
			BALL_NAME_SHORT,
			PVector(0.0,0.0),
			server_param->bsize,		/* ball size */
			0.0,		/* ball angle */
			server_param->bdecay,		/* ball decay */
			server_param->brand,		/* ball rand */ 
			server_param->bweight,	/* ball weight */ 
			server_param->bspeed_max,	/* ball speed max */
			// th 6.3.00
			server_param->baccel_max      /* ball acceleration max */
			//
			) ;
}

void Stadium::initPlayer(void)
{
	int i ;
	Player *p_l ;
	Player *p_r ;

	team_l = new Team(this, LEFT) ;
	team_r = new Team(this, RIGHT) ;

	for (i = 0; i < MAX_PLAYER; i++) {
		p_l = new Player(team_l, this, i+1) ;
		p_r = new Player(team_r, this, i+1) ;
		team_l->player[i] = p_l ;
		team_r->player[i] = p_r ;
		player[i] = p_l ;
		player[i+MAX_PLAYER] = p_r ;
		dest[i] = i ;
		dest[i+MAX_PLAYER] = i+MAX_PLAYER ;
	}
}

void Stadium::initCoach(Stadium *stad)
{
	coach = new Coach(stad) ;
	coach->eye = FALSE ;
}

void Stadium::initOnlineCoach(Stadium *stad)
{
	OnlineCoach *p_olcoach ;

	p_olcoach = new OnlineCoach(stad) ;
	team_l->olcoach = p_olcoach ;
	olcoach[1] = p_olcoach ;
	olcoach[1]->eye = FALSE ;
	olcoach[1]->freeform_messages_allowed = server_param->say_cnt_max;
	p_olcoach = new OnlineCoach(stad) ;
	team_r->olcoach = p_olcoach ;
	olcoach[2] = p_olcoach ;
	olcoach[2]->eye = FALSE ;
	olcoach[2]->freeform_messages_allowed = server_param->say_cnt_max;
}

Player* Stadium::newPlayer(const Name teamname, Value version, int goalie_flag,
							unsigned long host, int port)
{
	Team *tm ;

	if (team_l->name == NULLCHAR) {
		tm = team_l ;
		tm->name = teamname ;
		tm->enable = TRUE ;
	}
	else if (!strcmp(team_l->name, teamname))
		tm = team_l ;
	else if (team_r->name == NULLCHAR) {
		tm = team_r ;
		tm->name = teamname ;
		tm->enable = TRUE ;
	}
	else if (!strcmp(team_r->name, teamname))
		tm = team_r ;
	else {
		if (server_param->verbose)
			cerr << "Warning:Too many teams. [teamname = '"
					<< teamname << "']" << endl ;
		return NULL ; 
	}

	Player *p = tm->newPlayer(version, goalie_flag) ;

	if (p == NULL)
		return (Player*)NULL ;

	motable.AssignObject(p) ;
	votable.AssignObject(p) ;

	//[2000.01.02:I.Noda]  exit when udp_init failes.
	if(udp_init(((tm->side == LEFT) ? 0 : 1) * MAX_PLAYER + p->unum, 0)) {
	   cerr << "Error: udp_init():fail" << strerror(errno) << endl ;
	exit(1) ;
	}
	p->client.host = host ;
	p->client.port = port ;

	return p ;
}

Player* Stadium::get_client_player(unsigned long host, int port)
{
	int i ; 

	for (i = 0 ; i < MAX_PLAYER * 2 ; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		if (player[i]->client.host == host &&
			player[i]->client.port == port ) 
				return player[i] ;
	}
			
	return NULL ;
}

Player* Stadium::newClient(unsigned long host, int port, char* init_message)
{
	char buffer[128], *teamname ;
	char *ptr = init_message ;
	int goalie_flag = FALSE ;
	float version = 3.0 ;
	int count = 0 ;

    if (strncmp(ptr, "(init ", 6)) {
		if (server_param->verbose)
			cerr << "Warning:Illegal initialize message." << endl 
					<< "   message = " << init_message << endl ;
		return NULL ;
	}
    ptr += 6 ;

    while (*ptr != (char)NULL) {
        if (*ptr == '(') {
            if (!strncmp(ptr, "(version ", 9)) { /* noda */
                ptr += 9 ;
                break ;
			}
            else if (!strncmp(ptr, "(goalie)", 8)) {
				ptr += 8 ;
				goalie_flag = TRUE ;
				break ;
			}
			else { 
				if (server_param->verbose)
					cerr << "Warning:Illegal team name." << endl 
						<< "   message = " << init_message << endl ;
				return NULL ;
            }
        }
        else if (*ptr == ')' ) {
			buffer[count] = NULLCHAR ;
			sscanf(buffer, "%[-_a-zA-Z0-9]", buffer) ;
			if(count != strlen(buffer)) {
				if (server_param->verbose)
					cerr << "Warning:illegal_team_char."
						<< endl ;
			 	return NULL ;
			}
			teamname = strdup(buffer) ;
			return newPlayer(teamname, (Value)version, goalie_flag, host, port);
		}
        else {
            buffer[count++] = *ptr++ ;
			if (count > 32) {
				if (server_param->verbose)
					cerr << "Warning:Too long team name." << endl
						<< "   message = " << init_message << endl ;
				return NULL ;
			}
		}
    }
	
	if (isspace(buffer[count-1]))
		count-- ;
	buffer[count] = NULLCHAR ;
	sscanf(buffer, "%[-_a-zA-Z0-9]", buffer) ;
	if(count != strlen(buffer)) {
		if (server_param->verbose)
			cerr << "Warning:illegal_team_char." << endl ;
	 	return NULL ;
	}
	teamname = strdup(buffer) ;

    while (*ptr != (char)NULL) {
        if (isspace(*ptr))
            ptr++ ;
        else
            break ;
	}

	if (goalie_flag) {
		if (!strncmp(ptr, "(version ", 9))
			ptr += 9 ;
		else { 
			if (server_param->verbose)
				cerr << "Warning:Illegal team name." << endl 
					<< "   message = " << init_message << endl ;
			return NULL ;
		}
	}
		
    if (!sscanf(ptr, "%f", &version)) {
		if (server_param->verbose)
			cerr << "Warning:Illegal initialize message." << endl 
					<< "   message = " << init_message << endl ;
		return NULL ;
	}
    while (*ptr != (char)NULL) {
        if (isdigit(*ptr) || *ptr == '.' || isspace(*ptr))
            ptr++ ;
        else
            break ;
	}

    if (*ptr++ != ')') {
		if (server_param->verbose)
			cerr << "Warning:Illegal initialize message." << endl 
					<< "   message = " << init_message << endl ;
		return NULL ;
	}

    while (*ptr != (char)NULL) {
        if (isspace(*ptr))
            ptr++ ;
        else
            break ;
	}

    if (*ptr == ')') {
		return newPlayer(teamname, (Value)version, goalie_flag, host, port) ;
	}

    if (strncmp(ptr, "(goalie)", 8)) {
		if (server_param->verbose)
			cerr << "Warning:Illegal initialize message." << endl 
					<< "   message = " << init_message << endl ;
		return NULL ;
	}

    goalie_flag = TRUE ;
	return newPlayer(teamname, (Value)version, goalie_flag, host, port) ;
}

OnlineCoach* Stadium::newCoach(unsigned long host, int port, char* init_message)
{
    const int buflen = 1024;
	const Value default_olcoach_version = 5.0;

	char teamname[buflen] ;
	char *ptr = init_message ;
	double version = default_olcoach_version;

	// int goalie_flag = FALSE ;
	// int count = 0 ;
	
	int count_params = sscanf(init_message, "(init %[^ ()] (version %lf))", teamname, &version);

	switch (count_params) {
	case 0:			// was not able to read in even team name
		if (server_param->verbose)
			cerr << "Warning:Illegal initialize message." << endl 
					<< "   message = " << init_message << endl ;
		return NULL ;
		break;		// just for safety - 'return' code above might change
	case 1:
	        break;		// just use default version number
	case 2:
	        if (version < 5.00) {
	        	if (server_param->verbose)
	        		cerr << "Warning: No online coach for versions older than 5.00" << endl;
	        	return NULL;
	        }
	        break;		// ok, explicit version number read in and it is above 5.00
	default:
	  // we should *never* reach this branch
	        cerr << "Warning: Internal error at " << __FILE__ << ":" << __LINE__ << endl;
		return NULL;
	}

	if (strlen(teamname) > 32) {
	        if (server_param->verbose)
	                cerr << "Warning:Too long team name." << endl
	                     << "   message = " << init_message << endl ;
	        return NULL ;
	}

//      if (strncmp(ptr, "(init ", 6)) {
//  		if (server_param->verbose)
//  			cerr << "Warning:Illegal initialize message." << endl 
//  					<< "   message = " << init_message << endl ;
//  		return NULL ;
//  	}
//      ptr += 6 ;

        if (team_l->name != NULLCHAR && !strcmp(team_l->name, teamname)) {
        	if(team_l->olcoach->assignedp == TRUE) {
        		return NULL ;
        	}
        	team_l->olcoach->assignedp = TRUE ;
        	team_l->olcoach->client.host = host ;
        	team_l->olcoach->client.port = port ;
        	team_l->olcoach->side = LEFT ;
		team_l->olcoach->version = version;

		//[2000.01.02:I.Noda]  exit when udp_init failes.
		if(udp_init(-1, 0)) {
		   cerr << "Error: udp_init():fail" << strerror(errno) << endl ;
		   exit(1) ;
		}

		   
		if (team_l->olcoach->version >= 6.00) 
		        team_l->olcoach->udp_send("(init l ok)") ;
		else
		        team_l->olcoach->udp_send("(init ok)") ;

		if ( team_l->olcoach->version >= 7.00 ) {
                  send_parameter_info ( team_l->olcoach );
                  send_changed_players ( team_l->olcoach );
                }

		return team_l->olcoach ;
        }else if (team_r->name != NULLCHAR && !strcmp(team_r->name, teamname)) {
        	if(team_r->olcoach->assignedp == TRUE) {
        		return NULL ;
        	}
        	team_r->olcoach->assignedp = TRUE ;
        	team_r->olcoach->client.host = host ;
        	team_r->olcoach->client.port = port ;
        	team_r->olcoach->side = RIGHT ;
		team_r->olcoach->version = version;
		
		//[2000.01.02:I.Noda]  exit when udp_init failes.
		if(udp_init(-2, 0)) {
		   cerr << "Error: udp_init():fail" << strerror(errno) << endl ;
		   exit(1) ;
		}
		   
		if (team_r->olcoach->version >= 6.00) 
		        team_r->olcoach->udp_send("(init r ok)") ;
		else
		        team_r->olcoach->udp_send("(init ok)") ;

		if ( team_r->olcoach->version >= 7.00 ) {
                  send_parameter_info ( team_r->olcoach );
                  send_changed_players ( team_r->olcoach );
                }
		return team_r->olcoach ;
        }else {
        	return NULL ; 
        }    
	  


//      while (*ptr != (char)NULL) {
//      	if (*ptr == ')' ) {
//  		buffer[count] = NULLCHAR ;
//  		teamname = strdup(buffer) ;
//  		if (team_l->name != NULLCHAR && !strcmp(team_l->name, teamname)) {
//  			if(team_l->olcoach->assignedp == TRUE) {
//  				return NULL ;
//  			}
//  			team_l->olcoach->assignedp = TRUE ;
//  			team_l->olcoach->client.host = host ;
//  			team_l->olcoach->client.port = port ;
//  			team_l->olcoach->side = LEFT ;
//  			udp_init(-1, 0) ;
//  			team_l->olcoach->udp_send("(init l ok)") ;
//  			return team_l->olcoach ;
//  		}else if (team_r->name != NULLCHAR && !strcmp(team_r->name, teamname)) {
//  			if(team_r->olcoach->assignedp == TRUE) {
//  				return NULL ;
//  			}
//  			team_r->olcoach->assignedp = TRUE ;
//  			team_r->olcoach->client.host = host ;
//  			team_r->olcoach->client.port = port ;
//  			team_r->olcoach->side = RIGHT ;
//  			udp_init(-2, 0) ;
//  			team_r->olcoach->udp_send("(init r ok)") ;
//  			return team_r->olcoach ;
//  		}else {
//                 		return NULL ; 
//  		}
//  	}
//  	else if (!strncmp(ptr, "(version ", strlen("(version "))) {
	        
//  	}
//          else if (*ptr == '(') {	// opening brace, but no (version ...) command
//         	        if (server_param->verbose)
//  		        cerr << "Illegal command form" << endl;
//  		return NULL;
//  	}
//  	else {
//  		buffer[count++] = *ptr++ ;
//  		if (count > 32) {
//  		        if (server_param->verbose)
//  			        cerr << "Warning:Too long team name." << endl
//  				     << "   message = " << init_message << endl ;
//  			return NULL ;
//  		}
//  	}
//      }
}

Player* Stadium::reconnectClient(unsigned long host, int port,
										char* reconnect_message)
{
	char buffer[128] ;
	int r ;
	ID rnum ;

	r = sscanf(reconnect_message,"(reconnect %s %d)", buffer, &rnum) ;

	if(r < 2) {
		if (server_param->verbose)
			cerr << "Warning:Illegal initialize message." << endl 
					<< "   message = " << reconnect_message << endl ;
		return NULL ;
	}

	for(r = 0 ; r < MAX_PLAYER * 2; r++)
		if (player[r]->team->enable)
			if (!strcmp(player[r]->team->name,buffer)
				&& player[r]->unum == rnum)
					break ;

	/* It's same. */
	if (r < MAX_PLAYER * 2) {
 	       //[2000.01.02:I.Noda] exit when udp_init failes.
	       if(udp_init(((player[r]->team->side == LEFT) ? 0 : 1) 
			   * MAX_PLAYER + player[r]->unum, 
			   0)) {
		  cerr << "Error: udp_init():fail" << strerror(errno) << endl ;
		  exit(1) ;
	       }
		  
		player[r]->client.host = host ;
		player[r]->client.port = port ;
		player[r]->alive = STAND ;
		player[r]->enable = TRUE ;
		if (player[r]->goalie)
			player[r]->alive |= GOALIE ;
		return player[r] ;
	}
	else 
		return NULL ;
}

void Stadium::byeClient(Player* bye)
{
	if (bye->goalie == TRUE && bye->team->side == caught_ball)
		caught_ball = FALSE ;
	bye->enable = FALSE ;
	bye->client.host = 0 ;
	bye->client.port = -1 ;
	bye->alive = DISABLE ;
	bye->pos.x = -(bye->unum * 3 * bye->team->side) ;
	bye->pos.y = -PITCH_WIDTH/2.0 - 3.0 ;
	socketfd[((bye->team->side == LEFT) ? 0 : 1 )*MAX_PLAYER + bye->unum] = -1 ;
}

void Stadium::step() 
{
	int i ;

	for (i=1; i<=2; i++) {
	  if (olcoach[i]) {
	    olcoach[i]->check_message_queue(time);
	    olcoach[i]->update_messages_left(time);
	  }
	}
	
	if (mode == PM_BeforeKickOff) {
                motable._turn();
		if (server_param->kickoffoffside == TRUE)
		  check_players_move() ;
	}
	else if (mode == PM_AfterGoal_Right) {
		caught_ball = FALSE ;
		motable._inc ( mode ) ; 
		if (++after_g_time > AFTER_GOAL_WAIT) {
			set_ball(PVector(0.0,0.0),LEFT) ;
			set_players() ;
			change_play_mode(PM_KickOff_Left) ;
		}
	}
	else if (mode == PM_AfterGoal_Left) {
		caught_ball = FALSE ;
		motable._inc ( mode ) ; 
		if (++after_g_time > AFTER_GOAL_WAIT) {
			set_ball(PVector(0.0,0.0),RIGHT) ;
			set_players() ;
			change_play_mode(PM_KickOff_Right) ;
		}
	}
	else if (mode == PM_OffSide_Right) {
		caught_ball = FALSE ;
		motable._inc ( mode ) ; 
		check_player_after_offside() ;
		if (++after_offside_time > AFTER_OFFSIDE_WAIT) {
			change_play_mode(PM_FreeKick_Left) ;
		}
	}
	else if (mode == PM_OffSide_Left) {
		caught_ball = FALSE ;
		motable._inc ( mode ) ; 
		check_player_after_offside() ;
		if (++after_offside_time > AFTER_OFFSIDE_WAIT) {
			change_play_mode(PM_FreeKick_Right) ;
		}
	}
	else if (mode != PM_BeforeKickOff && mode != PM_TimeOver) {

	        if (mode == PM_KickOff_Right || mode == PM_KickIn_Right ||
                    mode == PM_FreeKick_Right || mode == PM_CornerKick_Right ||
                    mode == PM_GoalKick_Right || mode == PM_KickOff_Left ||
                    mode == PM_KickIn_Left || mode == PM_FreeKick_Left ||
                    mode == PM_CornerKick_Left || mode == PM_GoalKick_Left) {
                        if (drop_ball_time > -1)
                                drop_ball_time--;
                        if (drop_ball_time == 0)
                                drop_ball();
                }

		motable._inc ( mode ) ; 
		if (server_param->useoffside)
			offside_check() ;
		time++ ;
		referee_step() ;
		/* pfr - moved next line before if */
		prev_ball_pos = ball->pos ;
		if (caught_ball && mode != PM_BeforeKickOff) {
		   /* [2000.07.21:I.Noda] moved to field.h as inline func. 
			PVector temp(last_touch->relative_ball_ang.x,
					last_touch->relative_ball_ang.y);
			check_goalie_while_catch(last_touch) ;
			temp *= (last_touch->player_type->playerSize () + server_param->bsize);
			ball->pos = last_touch->pos + temp;
		   */
		   move_caught_ball() ;
		}

		/* pfr 7/26/2000 */
		//printf("time: %d\t2ht: %d\n", time, 2*server_param->half_time);
		if (time > 0 && time % (2*server_param->half_time) == 0) {
		  //printf("Giving coachs more messages\n");
		  //every game time cycles, give the online coach more messages
		  if (olcoach[1]->freeform_messages_allowed > 0)
		    olcoach[1]->freeform_messages_allowed += server_param->say_cnt_max;
		  if (olcoach[2]->freeform_messages_allowed > 0)
		    olcoach[2]->freeform_messages_allowed += server_param->say_cnt_max;
		}
		

		dinfo.body.show.time = htons(time) ;
		dinfo2.body.show.time = htons(time) ;
	}

	for (i = 0 ; i < MAX_PLAYER * 2; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		if (player[i]->stamina <= server_param->recover_dthr * server_param->stamina_max) {
			if (player[i]->recovery > server_param->recover_min)
				player[i]->recovery -= server_param->recover_dec ;
		}
		if (player[i]->stamina <= server_param->effort_dthr * server_param->stamina_max) {
			if (player[i]->effort > player[i]->player_type->effortMin () )
				player[i]->effort -= server_param->effort_dec ;
				if (player[i]->effort < player[i]->player_type->effortMin ())
					player[i]->effort = player[i]->player_type->effortMin () ;
		}
		if (player[i]->stamina >= server_param->effort_ithr * server_param->stamina_max) {
			if (player[i]->effort < player[i]->player_type->effortMax () ) { 
				player[i]->effort += server_param->effort_inc ;
				if (player[i]->effort > player[i]->player_type->effortMax ())
					player[i]->effort = player[i]->player_type->effortMax () ;
			}
		}
		player[i]->stamina += player[i]->recovery * player[i]->player_type->staminaIncMax () ;
		if (player[i]->stamina > server_param->stamina_max)
			player[i]->stamina = server_param->stamina_max ;

		player[i]->hear_capacity_from_teammate += server_param->hear_inc ;
		if (player[i]->hear_capacity_from_teammate > server_param->hear_max)
			player[i]->hear_capacity_from_teammate = server_param->hear_max ;

		player[i]->hear_capacity_from_opponent += server_param->hear_inc ;
		if (player[i]->hear_capacity_from_opponent > server_param->hear_max)
			player[i]->hear_capacity_from_opponent = server_param->hear_max ;
		if (player[i]->goalie_catch_ban)
			player[i]->goalie_catch_ban-- ;
	}

	/* send to displays */
	dinfo.body.show.pos[0].x = htons((short)rint((ball->pos.x * SHOWINFO_SCALE))) ;
	dinfo.body.show.pos[0].y = htons((short)rint((ball->pos.y * SHOWINFO_SCALE))) ;

	dinfo2.body.show.ball.x = htonl((long)rint((ball->pos.x * SHOWINFO_SCALE2))) ;
	dinfo2.body.show.ball.y = htonl((long)rint((ball->pos.y * SHOWINFO_SCALE2))) ;
	dinfo2.body.show.ball.deltax = htonl((long)rint((ball->vel.x * SHOWINFO_SCALE2))) ;
	dinfo2.body.show.ball.deltay = htonl((long)rint((ball->vel.y * SHOWINFO_SCALE2))) ;

	for (i = 0 ; i < MAX_PLAYER * 2 ; i++) {
		dinfo.body.show.pos[i+1].enable = htons(player[i]->alive) ; 		
		//player[i]->alive &= (STAND | GOALIE | DISCARD) ;
		dinfo.body.show.pos[i+1].x =
			htons((short)rint((player[i]->pos.x * SHOWINFO_SCALE))) ;
		dinfo.body.show.pos[i+1].y =
			htons((short)rint((player[i]->pos.y * SHOWINFO_SCALE))) ;
		dinfo.body.show.pos[i+1].angle = htons((short)Rad2IDegRound(player[i]->angle)) ;
		dinfo.body.show.pos[i+1].side = htons(player[i]->team->side) ;
		dinfo.body.show.pos[i+1].unum = htons(player[i]->unum) ;

		dinfo2.body.show.pos[i].mode = htons(player[i]->alive) ;
		dinfo2.body.show.pos[i].type = htons(player[i]->player_type_id);

		//player[i]->alive &= (STAND | GOALIE | DISCARD) ;
		dinfo2.body.show.pos[i].x = htonl((long)rint((player[i]->pos.x * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].y = htonl((long)rint((player[i]->pos.y * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].deltax = htonl((long)rint((player[i]->vel.x * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].deltay = htonl((long)rint((player[i]->vel.y * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].body_angle = htonl((long)rint((player[i]->angle * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].head_angle = htonl((long)rint((player[i]->angle_neck_committed * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].view_width = htonl((long)rint((player[i]->vis_angle * SHOWINFO_SCALE2)) );
		dinfo2.body.show.pos[i].view_quality = htons((short)rint((player[i]->highquality))) ;
		
		dinfo2.body.show.pos[i].stamina = htonl((long)rint((player[i]->stamina * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].effort = htonl((long)rint((player[i]->effort * SHOWINFO_SCALE2))) ;
		dinfo2.body.show.pos[i].recovery = htonl((long)rint((player[i]->recovery * SHOWINFO_SCALE2))) ;

		dinfo2.body.show.pos[i].kick_count = htons(player[i]->count_kick) ;
		dinfo2.body.show.pos[i].dash_count = htons(player[i]->count_dash) ;
		dinfo2.body.show.pos[i].turn_count = htons(player[i]->count_turn) ;
		dinfo2.body.show.pos[i].say_count = htons(player[i]->count_say) ;
		dinfo2.body.show.pos[i].tneck_count = htons(player[i]->count_turn_neck) ;
		dinfo2.body.show.pos[i].catch_count = htons(player[i]->count_catch) ;
		dinfo2.body.show.pos[i].move_count = htons(player[i]->count_move) ;
		dinfo2.body.show.pos[i].chg_view_count = htons(player[i]->count_change_view) ;

		player[i]->alive &= (STAND | GOALIE | DISCARD) ;
	}
	if (team_l->name != NULLCHAR) {
		strcpy(dinfo.body.show.team[0].name, team_l->name) ;
		dinfo.body.show.team[0].score = htons(team_l->point) ;

		strcpy(dinfo2.body.show.team[0].name, team_l->name) ;		
		dinfo2.body.show.team[0].score = htons(team_l->point) ;
	}
	else
	    {
		strcpy(dinfo2.body.show.team[0].name, "") ;
		dinfo2.body.show.team[0].score = htons(0) ;
	    }
	if (team_r->name != NULLCHAR) {
		strcpy(dinfo.body.show.team[1].name, team_r->name) ;
		dinfo.body.show.team[1].score = htons(team_r->point) ;

		strcpy(dinfo2.body.show.team[1].name, team_r->name) ;
		dinfo2.body.show.team[1].score = htons(team_r->point) ;
	}
	else
	    {
		strcpy(dinfo2.body.show.team[1].name, "") ;
		dinfo2.body.show.team[1].score = htons(0) ;
	    }

	displist_t *p ;
	for (p = top.next ; p != NULL ; p = p->next)
	    {
	    if ( p->version == 1 )
		udp_send(&dinfo,p->shost,p->sport,0) ;
	    else
		{
		udp_send(&dinfo2,p->shost,p->sport,0) ;
		}
	    }
	
	/* TH - 2-NOV-2000 */
	static bool wrote_final_cycle = false;
	
	if (server_param->recfile[0] != NULLCHAR)
	  {
	    if (mode != PM_BeforeKickOff && mode != PM_TimeOver) 
		{
		  if (server_param->rec_ver == REC_OLD_VERSION || server_param->rec_ver == REC_VERSION_2)
		    WriteLog(&dinfo) ;
		  else
		    WriteLog(&dinfo2) ;
		}
	    else if (mode == PM_TimeOver && !wrote_final_cycle)
		{
		  if (server_param->rec_ver == REC_OLD_VERSION || server_param->rec_ver == REC_VERSION_2)
		    WriteLog(&dinfo) ;
		  else
		    WriteLog(&dinfo2) ;
		  wrote_final_cycle = true;
		}
	  }
}


/*
 *===================================================================
 *Part: Field
 *===================================================================
 */
void Field::assign(Stadium *stad)
{
	goal_l.name = GOAL_L_NAME ;
	goal_r.name = GOAL_R_NAME ;
	line_l.name = LINE_L_NAME ;
	line_r.name = LINE_R_NAME ;
	line_t.name = LINE_T_NAME ;
	line_b.name = LINE_B_NAME ;

	goal_l.short_name = GOAL_L_NAME_SHORT ;
	goal_r.short_name = GOAL_R_NAME_SHORT ;
	line_l.short_name = LINE_L_NAME_SHORT ;
	line_r.short_name = LINE_R_NAME_SHORT ;
	line_t.short_name = LINE_T_NAME_SHORT ;
	line_b.short_name = LINE_B_NAME_SHORT ;

	goal_l.Set(PVector(-PITCH_LENGTH/2.0,0.0),stad->server_param->gwidth) ;
	goal_r.Set(PVector(PITCH_LENGTH/2.0,0.0),stad->server_param->gwidth) ;
	line_l.Set(PVector(-PITCH_LENGTH/2.0,0.0),1.0) ;
	line_r.Set(PVector(PITCH_LENGTH/2.0,0.0),1.0) ;
	line_t.Set(PVector(-PITCH_WIDTH/2.0,0.0),1.0) ;
	line_b.Set(PVector(PITCH_WIDTH/2.0,0.0),1.0) ;
}

/*
 *===================================================================
 *Part: Referee
 *===================================================================
 */
void Stadium::referee_step()
{
	/* if a value of half_time is negative, then ignore time. */
	if(server_param->half_time > 0) {
		/* check half_time and time_over. */
	          if(time >= server_param->half_time * 2) {
			if (!team_l || !team_r) {
				say("time_up_without_a_team") ;
				change_play_mode(PM_TimeOver) ;
				return ;
			}
			else if (team_l->point != team_r->point) {
				say("time_up") ;
				change_play_mode(PM_TimeOver) ;
				return ;
			}
			else if (time >= server_param->half_time * ( startp + 1 ) )
			  {
			    startp++;
			    caught_ball = FALSE;
			    say("time_extended") ;
			    if ( startp % 2 == 0 )
				set_ball(PVector(0.0,0.0),LEFT) ;
			    else
				set_ball(PVector(0.0,0.0),RIGHT) ;      
			    set_players() ;
			    change_play_mode(PM_BeforeKickOff) ;
			    time = server_param->half_time * startp ;
			    weather.wind_vector.x *= -1 ;
			    weather.wind_vector.y *= -1 ;
			    //startp = FALSE ;
			    return ;
			  }
		}
		else if(time >= server_param->half_time && startp == 0) {
			caught_ball = FALSE ;
			say("half_time") ;
			set_ball(PVector(0.0,0.0),RIGHT) ;
			set_players() ;
			change_play_mode(PM_BeforeKickOff) ;
			time = server_param->half_time ;
			startp++;
			weather.wind_vector.x *= -1 ;
			weather.wind_vector.y *= -1 ;
			return ;
		}
	}

	if(!server_param->CoachMode) {
		check_mode() ;
		check_goal() ;
		check_ball_in_pitch() ;
		check_player_in_field() ;
		return ;
	}
	if(!coach->assignedp) {
		check_mode() ;
		check_goal() ;
		check_ball_in_pitch() ;
	} else
		if(check_ball_pos() != BPI_InField)
			ball->vel = PVector(0.0,0.0) ;

	check_player_in_field() ;
}

void Stadium::check_mode()
{
	if (mode == PM_GoalKick_Right || mode == PM_GoalKick_Left) {
	        check_field_before_goalkick() ;
		check_player_in_field() ;
		}
	else if (mode != PM_PlayOn)
		if(ball->vel.x != 0.0 || ball->vel.y != 0.0){
			check_player_in_field() ;
			change_play_mode(PM_PlayOn) ;
		}
}

void Stadium::set_ball(PVector p,Side s, PVector vel)
{
    ball->acc = PVector(0.0,0.0) ;
    ball->vel = vel ;
    ball->pos = p ;
    kick_off_side = s ;
}

void Stadium::set_ball(PVector p,Side s,Angle ang, PVector vel)
{
    ball->angle = ang ;
    set_ball(p, s, vel) ;
}

void Stadium::check_players_move()
{
	static RArea fld_l(PVector(-PITCH_LENGTH/4,0.0),
					PVector(PITCH_LENGTH/2, PITCH_WIDTH)) ;
	static RArea fld_r(PVector(PITCH_LENGTH/4,0.0),
					PVector(PITCH_LENGTH/2, PITCH_WIDTH)) ;
	static CArea c(PVector(0.0,0.0),KICK_OFF_CLEAR_DISTANCE) ;

	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2 ; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		switch (player[i]->team->side) {
		case LEFT:
			if(player[i]->pos.x > 0)
				player[i]->pos = fld_l.randomize() ;
			break ;
		case RIGHT:
			if(player[i]->pos.x < 0)
				player[i]->pos = fld_r.randomize() ;
			break ;
		}
		if(player[i]->team->side != kick_off_side)
			if(c.inArea(player[i]->pos))
				player[i]->pos = c.nearestEdge(player[i]->pos) ;
	}
}

void Stadium::check_player_after_offside()
{

    CArea c(offside_pos, server_param->offside_area) ;

	int offsideside ;
	PVector ce, si ;

    if (mode == PM_OffSide_Right) {
        ce += PVector(PITCH_LENGTH/4 + offside_pos.x/2,0.0) ;
		si += PVector(PITCH_LENGTH/2 - offside_pos.x, PITCH_WIDTH) ;
		offsideside = RIGHT ;
	}
	else {
        ce += PVector(-PITCH_LENGTH/4 + offside_pos.x/2,0.0) ;
		si += PVector(PITCH_LENGTH/2 + offside_pos.x, PITCH_WIDTH) ;
		offsideside = LEFT ;
    }
    RArea fld(ce, si) ;

	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2 ; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		if (player[i]->team->side == offsideside) {
			if (c.inArea(player[i]->pos))
				player[i]->pos = c.nearestEdge(player[i]->pos) ;
			if (!fld.inArea(player[i]->pos))
				if(mode == PM_OffSide_Right)
					player[i]->pos = fld.nearestVEdge(player[i]->pos)
						+ PVector(server_param->offside_kick_margin,0) ;
				else
					player[i]->pos = fld.nearestVEdge(player[i]->pos)
						- PVector(server_param->offside_kick_margin,0) ;
		}
	}
}

void Stadium::check_goalie_while_catch(Player *goalie)
{
	RArea pta(PVector((-PITCH_LENGTH/2+PENALTY_AREA_LENGTH/2.0) * goalie->team->side, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	if (!(pta.inArea(goalie->pos)))
		goalie->pos = pta.nearestEdge(goalie->pos) ;
}

void Stadium::check_field_before_goalkick()
{
	static RArea p_l(PVector(-PITCH_LENGTH/2+PENALTY_AREA_LENGTH/2.0, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	static RArea p_r(PVector(+PITCH_LENGTH/2-PENALTY_AREA_LENGTH/2.0, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	RArea* fld ;

	int oppside ;

	if (mode == PM_GoalKick_Left) {
                oppside = RIGHT ;
		fld = &p_l ;
	} else {
                oppside = LEFT ;
		fld = &p_r ;
        }

	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2 ; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		if (player[i]->team->side == oppside) {
			if ((fld->inArea(player[i]->pos)))
				player[i]->pos = fld->nearestEdge(player[i]->pos) ;
			if ( player[i]->pos.x * oppside >= PITCH_LENGTH/2)
				player[i]->pos.x = (PITCH_LENGTH/2 - PENALTY_AREA_LENGTH) * oppside ;
		}
	}
	if(!(fld->inArea(ball->pos))){
		check_player_in_field() ;
		change_play_mode(PM_PlayOn) ;
        }
}

void Stadium::set_players()
{
	static RArea fld_l(PVector(-PITCH_LENGTH/4,0.0),
					PVector(PITCH_LENGTH/2, PITCH_WIDTH)) ;
	static RArea fld_r(PVector(PITCH_LENGTH/4,0.0),
					PVector(PITCH_LENGTH/2, PITCH_WIDTH)) ;
	static CArea c(PVector(0.0,0.0),KICK_OFF_CLEAR_DISTANCE) ;

	int i ;

	c.center = ball->pos ;

	for (i = 0 ; i < MAX_PLAYER * 2; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		switch (player[i]->team->side) {
		case LEFT:
			if(!(fld_l.inArea(player[i]->pos)))
				player[i]->pos = fld_l.randomize() ;
			break ;
		case RIGHT:
			if(!(fld_r.inArea(player[i]->pos)))
				player[i]->pos = fld_r.randomize() ;
			break ;
		}
		if(player[i]->team->side != kick_off_side)
			if(c.inArea(player[i]->pos))
				player[i]->pos = c.nearestEdge(player[i]->pos) ;
	}
}

void Stadium::recovery_players()
{
	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		player[i]->stamina = server_param->stamina_max ;
		player[i]->recovery = 1.0 ;
		player[i]->effort = player[i]->player_type->effortMax () ;
		player[i]->hear_capacity_from_teammate = server_param->hear_max ;
		player[i]->hear_capacity_from_opponent = server_param->hear_max ;
	}
}

void Stadium::announce_goal(Team& tm)
{
	static char buf[MaxStringSize] ;

	sprintf(buf,"goal_%s_%d",SideStr(tm.side),tm.point) ;
	say(buf) ;
}

BallPosInfo Stadium::check_ball_pos()
{
	static RArea g_l(PVector((-PITCH_LENGTH-GOAL_DEPTH)/2.0,0.0),
			PVector(GOAL_DEPTH,server_param->gwidth)) ;
	static RArea g_r(PVector((+PITCH_LENGTH+GOAL_DEPTH)/2.0,0.0),
			PVector(GOAL_DEPTH,server_param->gwidth)) ;

	static RArea pt(PVector(0.0,0.0),
			PVector(PITCH_LENGTH,PITCH_WIDTH)) ;

	if(g_l.inArea(ball->pos)) return BPI_GoalL ;
	if(g_r.inArea(ball->pos)) return BPI_GoalR ;
	if(!(pt.inArea(ball->pos))) return BPI_OutOfField ;
	return BPI_InField ;
}

Logical Stadium::cross_gline(Side side)
{
	static double tmp ;

	if (prev_ball_pos.x == ball->pos.x)
		return FALSE ;

	if (fabs(ball->pos.x) < PITCH_LENGTH/2.0)
		return FALSE ;

	if ((side * ball->pos.x) >= 0)
		return FALSE ;

	tmp = ((prev_ball_pos.y - ball->pos.y) * (PITCH_LENGTH/2.0) * (-side)
			+ prev_ball_pos.x * ball->pos.y - ball->pos.x * prev_ball_pos.y)
				/ (prev_ball_pos.x - ball->pos.x) ;

	if (tmp >= -server_param->gwidth/2 && tmp <= server_param->gwidth/2)
		return TRUE ;
	else
		return FALSE ;

}

void Stadium::check_goal()
{
	static RArea pt(PVector(0.0,0.0),
				PVector(PITCH_LENGTH,PITCH_WIDTH)) ;

	if ((pt.inArea(ball->pos)))
		return ;

	if (cross_gline(LEFT)) {
		if (team_r) {
			team_r->point++ ;
			announce_goal(*team_r) ;
			after_g_time = 0 ;
			set_ball(ball->pos, LEFT) ;
			if (server_param->half_time >= 0 && time >= server_param->half_time * 2)
				change_play_mode(PM_TimeOver) ;
			else {
				mode = PM_AfterGoal_Right ;
				dinfo.body.show.pmode = (char)PM_AfterGoal_Right ;
				dinfo2.body.show.pmode = (char)PM_AfterGoal_Right ;
								
			}
		}
	}
	else if (cross_gline(RIGHT)) {
		if (team_l) {
			team_l->point++ ;
			announce_goal(*team_l) ;
			after_g_time = 0 ;
			set_ball(ball->pos, RIGHT) ;
			if (server_param->half_time >= 0 && time >= server_param->half_time * 2)
				change_play_mode(PM_TimeOver) ;
			else {
				mode = PM_AfterGoal_Left ;
				dinfo.body.show.pmode = (char)PM_AfterGoal_Left ;
				dinfo2.body.show.pmode = (char)PM_AfterGoal_Left ;
			}
		}
	}
}

void Stadium::check_ball_in_pitch()
{
	static RArea pt(PVector(0.0,0.0),
				PVector(PITCH_LENGTH,PITCH_WIDTH)) ;
	PVector pos ;

	if (mode == PM_AfterGoal_Left ||
				mode == PM_AfterGoal_Right || mode == PM_TimeOver)
		return ;

	if (pt.left > ball->pos.x) {
		Side side = last_touch->team->side ;
		if(side == LEFT) {
			pos = pt.nearestHEdge(ball->pos) ;
			pos.x += server_param->ckmargin ;
			if ( pos.y > 0 )
				pos.y -= server_param->ckmargin ;
			else
				pos.y += server_param->ckmargin ;
				set_ball(pos, -side) ;
			caught_ball = FALSE; /* pfr 1/7/00 */
			change_play_mode(PM_CornerKick_Right) ;
		}
		else {
			PVector r(-PITCH_LENGTH/2.0+GOAL_AREA_LENGTH,
						((ball->pos.y > 0) ? 1 : -1) * GOAL_AREA_WIDTH/2.0) ;
			set_ball(r,-side) ;
			caught_ball = FALSE; /* pfr 1/7/00 */
			change_play_mode(PM_GoalKick_Left) ;
		}
	}
	else if (pt.right < ball->pos.x) {
		Side side = last_touch->team->side ;
		if(side == RIGHT) {
			pos = pt.nearestHEdge(ball->pos) ;
			pos.x -= server_param->ckmargin ;
			if ( pos.y > 0 )
				pos.y -= server_param->ckmargin ;
			else
				pos.y += server_param->ckmargin ;
			set_ball(pos, -side) ;
			caught_ball = FALSE; /* pfr 1/7/00 */
			change_play_mode(PM_CornerKick_Left) ;
		}
		else {
			PVector r(PITCH_LENGTH/2.0-GOAL_AREA_LENGTH,
						((ball->pos.y > 0) ? 1 : -1) * GOAL_AREA_WIDTH/2.0) ;
			set_ball(r,-side) ;
			caught_ball = FALSE; /* pfr 1/7/00 */
			change_play_mode(PM_GoalKick_Right) ;
		}
	}
	else if (!(pt.inArea(ball->pos))) {
		Side side = last_touch->team->side ;
		set_ball(pt.nearestHEdge(ball->pos), -side) ; 
		/* this one is probably superfluous - pfr */
		caught_ball = FALSE; /* pfr 1/7/00 */
		if (side == LEFT)
			change_play_mode(PM_KickIn_Right) ;
		else
			change_play_mode(PM_KickIn_Left) ;
	} 
}

void Stadium::check_player_in_field()
{
	static RArea fld(PVector(0.0,0.0),
					PVector(PITCH_LENGTH+PITCH_MARGIN*2.0,
					PITCH_WIDTH+PITCH_MARGIN*2.0)) ;
	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2; i++ ) {
		if (player[i]->alive == DISABLE)
			continue ;
		if(!(fld.inArea(player[i]->pos)))
			player[i]->pos = fld.nearestEdge(player[i]->pos) ;
	}

	if (mode != PM_PlayOn && mode != PM_GoalKick_Left && mode != PM_GoalKick_Right) {
		static CArea c(PVector(0.0,0.0), KICK_OFF_CLEAR_DISTANCE) ;
			c.center = ball->pos ;

		for(i = 0 ; i < MAX_PLAYER * 2; i++ ) {
			if (player[i]->alive == DISABLE)
				continue ;
			if(player[i]->team->side != kick_off_side)
				if(c.inArea(player[i]->pos))
					player[i]->pos = c.nearestEdge(player[i]->pos) ;
		}
	}

	if (server_param->kickoffoffside == TRUE) { 
		if (mode == PM_KickOff_Left) {
			for(i = 0 ; i < MAX_PLAYER * 2 ; i++ ) {
				if (player[i]->alive == DISABLE)
					continue ;
				if(player[i]->team->side == kick_off_side) {
					if(player[i]->pos.x > 0)
						player[i]->pos.x = 0 ;
				}
				else {
					if(player[i]->pos.x < 0)
						player[i]->pos.x = 0 ;
				}
			}
		}
		if(mode == PM_KickOff_Right) {
			for(i = 0 ; i < MAX_PLAYER * 2 ; i++ ) {
				if (player[i]->alive == DISABLE)
					continue ;
				if(player[i]->team->side == kick_off_side) {
					if(player[i]->pos.x < 0)
						player[i]->pos.x = 0 ;
				}
				else {
					if(player[i]->pos.x > 0)
						player[i]->pos.x = 0 ;
				}
			}
		}
	}
}

void Stadium::change_play_mode(PlayMode pm)
{
	static char *PlayModeString[] = PLAYMODE_STRINGS ;
	int i ;
	mode = pm ;
	dinfo.body.show.pmode = (char)pm ;
	dinfo2.body.show.pmode = (char)pm ;

	if (pm == PM_KickOff_Left ||
	    pm == PM_KickIn_Left ||
	    pm == PM_FreeKick_Left ||
	    pm == PM_CornerKick_Left ||
	    pm == PM_GoalKick_Left) {
	  kick_off_side = LEFT ;
	  //since we decrement first, if this is <=0, we will never drop
	  drop_ball_time = server_param->drop_time;
	} else if (pm == PM_KickOff_Right ||
		   pm == PM_KickIn_Right ||
		   pm == PM_FreeKick_Right ||
		   pm == PM_CornerKick_Right ||
		   pm == PM_GoalKick_Right) {
	  kick_off_side = RIGHT ;
	  //since we decrement first, if this is <=0, we will never drop
	  drop_ball_time = server_param->drop_time;
	} else if (pm == PM_Drop_Ball)
	  kick_off_side = NEUTRAL ;
	if (pm != PM_PlayOn)
	  for (i = 0; i < MAX_PLAYER * 2; i++)
	    player[i]->offside_mark = FALSE ;
	
	say(PlayModeString[pm]) ;
}

void
Stadium::drop_ball()
{
        static RArea lga(PVector(-PITCH_LENGTH/2 + PENALTY_AREA_LENGTH/2, 0),
                         PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH));
        static RArea rga(PVector(+PITCH_LENGTH/2 - PENALTY_AREA_LENGTH/2, 0),
                         PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH));

        if (lga.inArea(ball->pos)) {
                ball->pos.x = -PITCH_LENGTH/2 + PENALTY_AREA_LENGTH;
                if (ball->pos.y > 0) {
                        ball->pos.y = +PENALTY_AREA_WIDTH/2;
                } else {
                        ball->pos.y = -PENALTY_AREA_WIDTH/2;
                }
        }
        else if (rga.inArea(ball->pos)) {
                ball->pos.x = +PITCH_LENGTH/2 - PENALTY_AREA_LENGTH;
                if (ball->pos.y > 0) {
                        ball->pos.y = +PENALTY_AREA_WIDTH/2;
                } else {
                        ball->pos.y = -PENALTY_AREA_WIDTH/2;
                }                
        }
        referee_drop_ball(ball->pos.x, ball->pos.y, NEUTRAL);
}

void Stadium::referee_get_foul(double x, double y, int side)
{
	char msg[8] ;

	sprintf(msg,"foul_%s", SideStr(-side)) ;
	say(msg) ;
	caught_ball = FALSE ;
	(side == LEFT) ? 
		change_play_mode(PM_FreeKick_Left) : change_play_mode(PM_FreeKick_Right) ;
	set_ball(PVector(x,y),side) ;
	check_player_in_field() ;
}

void Stadium::referee_drop_ball(double x, double y, int side)
{
	change_play_mode(PM_Drop_Ball) ;
	set_ball(PVector(x,y),side) ;
	check_player_in_field() ;
	change_play_mode(PM_PlayOn) ;
	caught_ball = FALSE ;
}

void Stadium::discard_player(int side, int unum)
{
	int i ;

	for (i = 0; i < MAX_PLAYER * 2; i++) {
		if (player[i]->team->side == side && player[i]->unum == unum)
			break ;
	}

	if (i < MAX_PLAYER * 2) {
		if (player[i]->alive & STAND) {
			if (!(player[i]->alive & DISCARD))
				player[i]->alive |= DISCARD ;
			else
				player[i]->alive &= ~DISCARD ;
		}
	}
}

void Stadium::goalie_catch(Player *goalie)
{
	char msg[32] ;

	sprintf(msg,"goalie_catch_ball_%s", SideStr(goalie->team->side)) ;
	say(msg) ;
	(goalie->team->side == LEFT) ? 
		change_play_mode(PM_FreeKick_Left) : change_play_mode(PM_FreeKick_Right) ;
	set_ball(ball->pos, goalie->team->side) ;
	caught_ball = goalie->team->side ;
	last_touch = goalie ;
	check_player_in_field() ;
}

int Stadium::mark_offside_mark(Player *kicker)
{
	double fast = 0 ;
	double second = 0 ;
	double offside_line ;
	double tmp ;
	int i ;

	if (kicker->offside_mark) {
		offside_pos = kicker->offside_pos ;
		offside() ;
		return TRUE ;
	}

	for (i = 0; i < MAX_PLAYER * 2; i++)
		player[i]->offside_mark = FALSE ;

	switch (kicker->team->side) {
	  case LEFT:
		for (i = 0; i < MAX_PLAYER * 2; i++) {
			if (player[i]->alive == DISABLE)
				continue ;
			if (player[i]->team->side == RIGHT)
				if (player[i]->pos.x > second) {
					second = player[i]->pos.x ;
					if (second > fast) {
						tmp = fast ;
						fast = second ;
						second = tmp ;
					}
				}
		}

		if (second > ball->pos.x)
			offside_line = second ;
		else
			offside_line = ball->pos.x ;

		for (i = 0; i < MAX_PLAYER * 2; i++) {
			if (player[i]->alive == DISABLE)
				continue ;
			if (player[i]->team->side == LEFT &&
				player[i]->pos.x > offside_line &&
				player[i]->unum != kicker->unum) {
					player[i]->offside_mark = TRUE ;
					player[i]->offside_pos.x = offside_line ;
					player[i]->offside_pos.y = player[i]->pos.y ;
				}
		}
		break ;

	  case RIGHT:	  
		for (i = 0; i < MAX_PLAYER * 2; i++) {
			if (player[i]->alive == DISABLE)
				continue ;
			if (player[i]->team->side == LEFT)
				if (player[i]->pos.x < second) {
					second = player[i]->pos.x ;
					if (second < fast) {
						tmp = fast ;
						fast = second ;
						second = tmp ;
					}
				}
		}

		if (second < ball->pos.x)
			offside_line = second ;
		else
			offside_line = ball->pos.x ;

		for (i = 0; i < MAX_PLAYER * 2; i++) {
			if (player[i]->alive == DISABLE)
				continue ;
			if (player[i]->team->side == RIGHT &&
				player[i]->pos.x < offside_line &&
				player[i]->unum != kicker->unum) {
					player[i]->offside_mark = TRUE ;
					player[i]->offside_pos.x = offside_line ;
					player[i]->offside_pos.y = player[i]->pos.y ;
				}
		}
		break ;
	}
	return FALSE ;
}

void Stadium::offside_check(void)
{
	int i ;
	double dist = server_param->offside_area ;
	double tmp ;

	for (i = 0; i < MAX_PLAYER * 2; i++)
		if (player[i]->offside_mark)
			if ((tmp = player[i]->pos.distance(ball->pos)) < dist) {
				dist = tmp ;
				offside_pos = player[i]->offside_pos ;
			}

	if (dist != server_param->offside_area)
		offside() ;
}

void Stadium::offside(void)
{
	PVector pos ;

	static RArea pt(PVector(0.0,0.0),
			PVector(PITCH_LENGTH - 2.0*server_param->ckmargin , PITCH_WIDTH - 2.0*server_param->ckmargin)) ;
	static RArea g_l(PVector(-PITCH_LENGTH/2.0+GOAL_AREA_LENGTH/2.0, 0.0),
							PVector(GOAL_AREA_LENGTH, GOAL_AREA_WIDTH)) ;
	static RArea g_r(PVector(+PITCH_LENGTH/2.0-GOAL_AREA_LENGTH/2.0, 0.0),
							PVector(GOAL_AREA_LENGTH, GOAL_AREA_WIDTH)) ;
	(last_kicker->team->side == LEFT) ? 
		change_play_mode(PM_OffSide_Left) : change_play_mode(PM_OffSide_Right) ;
	/* this is probably superfluous, but let's be safe */
	caught_ball = FALSE; /* pfr 1/7/00 */

	after_offside_time = 0 ;

	if (offside_pos.x > PITCH_LENGTH/2.0 || g_r.inArea(offside_pos)) {
		pos.x = +PITCH_LENGTH/2.0-GOAL_AREA_LENGTH ;
		pos.y = ((offside_pos.y > 0) ? 1 : -1) * GOAL_AREA_WIDTH/2.0 ;
	}
	else if (offside_pos.x < -PITCH_LENGTH/2.0 || g_l.inArea(offside_pos)) {
		pos.x = -PITCH_LENGTH/2.0+GOAL_AREA_LENGTH ;
		pos.y = ((offside_pos.y > 0) ? 1 : -1) * GOAL_AREA_WIDTH/2.0 ;
	}
	else if (!(pt.inArea(offside_pos)))
		pos = pt.nearestEdge(offside_pos) ;
	else
		pos = offside_pos ;

	set_ball(pos, -last_kicker->team->side) ; 
	check_player_in_field() ;
}

/*
 *===================================================================
 *Part: Weather
 *===================================================================
 */
void Weather::init(Stadium *stad)
{
	if (stad->server_param->win_no) {
		wind_vector = PVector(0.0, 0.0) ;
		wind_rand = 0.0 ;
	}
	else if (stad->server_param->win_random) {
		wind_vector = Polar2PVector(drand(0, 100), Deg2Rad(drand(-180, 180))) ;
		wind_rand = drand(0, 0.5) ;
	}
	else {
		wind_vector = Polar2PVector(stad->server_param->winforce, Deg2Rad(stad->server_param->windir)) ;
		wind_rand = stad->server_param->winrand ;
	}

#if 1
	printf("wind factor: rand: %f, vector: (%f, %f)\n",
						wind_rand, wind_vector.x, wind_vector.y) ;
#endif

}

/*
 *===================================================================
 *Part: Coach
 *===================================================================
 */
Player* Stadium::get_player_by_name(char *name)
{
	int i ;

	for (i = 0 ; i < MAX_PLAYER * 2; i++) {
		if (player[i]->alive == DISABLE)
			continue ;
		if(!strcmp(player[i]->name,name))
			return player[i] ;
	}

	return NULL ;
} 

void Stadium::move_object(char *obj_name, PVector pos, PVector vel) 
{
    Player *p ;
    if ((p = get_player_by_name(obj_name)) == NULL) {
        if (server_param->verbose)
            cerr << "unknown name" << endl ;
        return ;
    }

    p->pos = pos ;
    p->vel = vel ;
}

void Stadium::move_object(char *obj_name, PVector pos, Angle ang, PVector vel) 
{
    Player *p ;
    if ((p = get_player_by_name(obj_name)) == NULL) {
        if (server_param->verbose)
            cerr << "unknown name" << endl ;
        return ;
    }

    p->pos = pos ;
    p->vel = vel ;
    p->rotate_to(ang) ;
}

void Stadium::check_collision(char *obj_name) 
{
	int i ;
	MPObject *obj = NULL;

	for(i = 0; i < motable.n; i++)
		if (!strcmp(obj_name, motable.object[i]->name))
			obj = motable.object[i] ;
	
	if (obj)
	    check_collision(obj) ;
}

void Stadium::check_collision(MPObject *obj) 
{
	int i ;
	MPObject *ob ;

	for(i = 0; i < motable.n; i++) {
		ob = motable.object[i] ;
		if (obj->collisionP(*ob))
			obj->pos.x += ob->size + obj->size ;
	}
}

void _Start(Stadium& stad)
{
  if ( stad.time % stad.server_param->half_time == 0 )
    {
	if( ( stad.time / stad.server_param->half_time ) % 2 == 0 ) {
		stad.recovery_players() ;
		stad.set_ball(PVector(0.0,0.0),LEFT) ;
		stad.change_play_mode(PM_KickOff_Left) ;
		cout << "Kick_off_left" << endl;
	}
	else if( ( stad.time / stad.server_param->half_time ) % 2 == 1 ) {
		stad.recovery_players() ;
		stad.set_ball(PVector(0.0,0.0),RIGHT) ;
		stad.change_play_mode(PM_KickOff_Right) ;
		cout << "Kick_off_right" << endl;		
	}
    }
}

void Stadium::OpenForWrite(void)
{
	char	buf[5] ;

	// open the file
	if ((rec_fp = fopen(server_param->recfile, "w")) == NULL){
		cerr << "can't open record_file " << server_param->recfile << endl ;
	}

	if (server_param->rec_ver != REC_OLD_VERSION) {
		// write version information
		buf[0] = 'U' ;
		buf[1] = 'L' ;
		buf[2] = 'G' ;
		buf[3] = server_param->rec_ver ;
		fwrite(buf, 4, 1, rec_fp) ;
		
		if ( server_param->rec_ver >= REC_VERSION_3 )
		    {
		    short mode = htons(PARAM_MODE);
		    server_params_t stmp = server_param->convertToStruct ();
		    fwrite((char *)(&mode), sizeof(mode), 1, rec_fp) ;
		    fwrite((char *)(&stmp), sizeof(stmp), 1, rec_fp) ;      

		    mode = htons(PPARAM_MODE);
		    player_params_t ptmp = player_param->convertToStruct ();
		    fwrite((char *)(&mode), sizeof(mode), 1, rec_fp) ;
		    fwrite((char *)(&ptmp), sizeof(ptmp), 1, rec_fp) ;      

		    mode = htons(PT_MODE);
		    for ( int i = 0; i < player_param->playerTypes (); i++ )
		      {
			player_type_t pt_tmp = player_types[i]->convertToStruct ( i );
			fwrite((char *)(&mode), sizeof(mode), 1, rec_fp) ;
			fwrite((char *)(&pt_tmp), sizeof(pt_tmp), 1, rec_fp) ;      
		      }
		    }
		
		fflush(rec_fp) ;
	}
}

void Stadium::WriteLog(dispinfo_t* msg)
{
    short	len = 1 ;

	if (server_param->rec_ver != REC_OLD_VERSION) {
		// write the type of next msg
		if (msg->mode != MSG_MODE || server_param->rec_msg_mode) {
			fwrite((char *)(&msg->mode), sizeof(msg->mode), 1, rec_fp) ;

			switch (ntohs(msg->mode)) {
			case SHOW_MODE:
			       // write a show struct
				fwrite((char *)(&msg->body.show), 
				       sizeof(msg->body.show), 1, rec_fp) ;
				break ;
			case MSG_MODE:

				// it is a message write the board info
				fwrite((char *)(&msg->body.msg.board), 
						sizeof(msg->body.msg.board), 1, rec_fp);

				// calculate the string length and write it
				while ((msg->body.msg.message[len-1] != '\0') && (len < 2048))
					len++;
			        /* pfr 1/7/00 Need to put length in network byte order */
			        short nlen = htons(len);
				fwrite((char *)(&nlen), sizeof(nlen), 1, rec_fp);

				// write the message
				fwrite(msg->body.msg.message, len, 1, rec_fp);
				break ;
			}
		}
	}
	else
		fwrite((char *)&msg, sizeof(dispinfo_t), 1, rec_fp) ;
	fflush(rec_fp) ;
}

void Stadium::WriteLog(dispinfo_t2* msg)
{
  short mode;
  static char pm = (char)0;
  static team_t teams[2] = { { "", 0 },
			     { "", 0 } };


  // if playmode has changed wirte playmode
  if ( pm != msg->body.show.pmode )
    {
      pm = msg->body.show.pmode;


      mode = htons(PM_MODE);
      fwrite((char *)(&mode), sizeof(mode), 1, rec_fp) ;
      fwrite((char *)(&pm), sizeof(pm), 1, rec_fp) ;      
  
    }

  // if teams or score has changed, write teams and score
  if ( strlen ( teams [ 0 ].name ) != strlen ( msg->body.show.team [ 0 ].name ) ||
       strcmp ( teams [ 0 ].name, msg->body.show.team [ 0 ].name ) ||
       teams [ 0 ].score != msg->body.show.team [ 0 ].score ||
       strlen ( teams [ 1 ].name ) != strlen ( msg->body.show.team [ 1 ].name ) ||
       strcmp ( teams [ 1 ].name, msg->body.show.team [ 1 ].name ) ||
       teams [ 1 ].score != msg->body.show.team [ 1 ].score
       )
    {
      strcpy ( teams [ 0 ].name, msg->body.show.team [ 0 ].name );
      teams [ 0 ].score = msg->body.show.team [ 0 ].score;
      strcpy ( teams [ 1 ].name, msg->body.show.team [ 1 ].name );
      teams [ 1 ].score = msg->body.show.team [ 1 ].score;

      mode = htons(TEAM_MODE);
      fwrite((char *)(&mode), sizeof(mode), 1, rec_fp) ;
      fwrite((char *)(teams), sizeof(team_t), 2, rec_fp) ;      

    }

  // write positional data
  short_showinfo_t2 pos_data;
  pos_data.ball = msg->body.show.ball;
  for ( int i = 0; i < MAX_PLAYER * 2; i++ )
    pos_data.pos [ i ] = msg->body.show.pos [ i ];
  pos_data.time = msg->body.show.time;

  mode = htons(SHOW_MODE);
  fwrite((char *)(&mode), sizeof(mode), 1, rec_fp) ;
  fwrite((char *)(&pos_data), sizeof(pos_data), 1, rec_fp) ;

  
  fflush(rec_fp) ;
}

// th 6.3.00
void Stadium::write_times(unsigned long tp_new, unsigned long tp_old) 
{
	if (server_param->log_times) {
	        unsigned long diff = (tp_new - tp_old) ;
	
		if (server_param->logfile[0] != NULLCHAR) {
			fprintf(log_fp, "%d\tCYCLE_TIMES: %d\n", time,  diff);
			fflush(log_fp) ;
		}
	}
}
//
