/*
 *Header:
 *File: field.h (for C++)
 *Author: Noda Itsuki
 *Date: 1996/02/16
 *EndHeader:
 */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */


#include <stdio.h>
#include <time.h>
#include <deque>
using namespace std;

#include "object.h"
#include "serverparam.h"
#include "hetroplayer.h"
#include "playerparam.h"
#include "garbage.h"
#include "coach_lang.h"

/* these global variables are used to interface with the lex/bison
   coach language parser */
extern CM_Message* g_pMess;
extern GarbageCollector g_GC;

/*
 *===================================================================
 *Part: Stadium Class
 *===================================================================
 */
class Coach {
 public:
	Logical	assignedp ;
	Client	client ;
	Stadium *stadium ;
	int	socketfd ;
	int 	eye ;
	int	hear ;
        Value version;
	int side ;

	int done_received; //pfr:SYNCH

	Coach(Stadium *stad) { stadium = stad; assignedp = False ; 
			      hear = FALSE ; version = 0.0; side = NEUTRAL; } ;
	int parse_init ( Stadium& stad, char *command, 
			 unsigned long host, int port);
	void parse_command(Stadium& stad, char *command,
			unsigned long host,
			int port) ;
	void look(Stadium& stad) ;
        void team_names(Stadium& stad) ;
        void recover(Stadium& stad) ;
	void send_visual_info(void) ;
	void check_ball(Stadium& stad) ;
	void udp_send(const char *msg, Logical logp = false) ;
} ;

class OnlineCoach : public Coach {
 public:
	Angle angle ;

	OnlineCoach(Stadium *stad);
	void parse_command(Stadium& stad, char *command,
			unsigned long host,
			int port) ;
	void say(char *message, bool standard = false) ;
	/* this deques and says any messages which should come out
	   returns the number of messages se nt out */
	int  check_message_queue(TheNumber time);
	
	int freeform_messages_said;
	int freeform_messages_allowed;
	char message_buf[MaxMesg];
	deque<CM_Message*> message_queue;

	int define_messages_left;
	int advice_messages_left;
	int info_messages_left;
	int meta_messages_left;

	TheNumber msg_left_update_time;

	//returns whether it updated anything
	bool update_messages_left(TheNumber time);
} ;

class Field {
 public:
	PObject	goal_l ;
	PObject	goal_r ;
	PObject	line_l ;
	PObject	line_r ;
	PObject	line_t ;
	PObject	line_b ;

	void assign(Stadium *stad) ;
} ;

class Weather {
 public:
	PVector	wind_vector ;
	Value	wind_rand ;

	void	init(Stadium *stad) ;
} ;

class Stadium {
 public:
	Field	field ;
	Weather	weather ;
	dispinfo_t dinfo ;
	dispinfo_t minfo ;
	displist_t top ;
	dispinfo_t2 dinfo2;
	
	MPObjectTable motable ;	/* Movable objects */
	PObjectTable votable ;	/* Visible objects */
	MPObject *ball ;
	PVector prev_ball_pos ;
	PVector offside_pos ;
	Player **player ;
//	TheNumber n_player ;
	TheNumber dest[MAX_PLAYER*2] ;
	Team *team_l ;
	Team *team_r ;
	TheNumber time ;
	TheNumber after_g_time ;
	TheNumber after_offside_time ;
        TheNumber drop_ball_time;

	Player *last_touch ;
	Player *last_kicker ;
	Side kick_off_side ;
	PlayMode mode ;
	int caught_ball ;	/* whether goalie has catch ball or not */

	TheNumber startp ;	/* flag for physical simulator */

	/* server[1 socket] + players[MAX_PLAYER * 2] */
	int socketfd[MAX_PLAYER*2 + 1] ;
	int olcoachfd[3] ;

	 /* for commandline/configure file option */
	ServerParam *server_param;
	PlayerParam *player_param;
	
	HetroPlayer **player_types; 
	
	
	FILE *log_fp ;					/* file pointer for server log */
	FILE *rec_fp ;					/* file pointer for recording */

	Coach *coach ;
	OnlineCoach **olcoach ;

	void init(int argc, char **argv) ;
	void clean ();
	void MainLoop() ;
	void check_io(int) ;
	void assign(void) ;
	MPObject* NewObject(
			const MPObjectType type,
			const Name name,
			const Name short_name,
			const PVector pos,
			const Value size,
			const Value angle,
			const Value decay,
			const Value rand,
			const Value weight,
			const Value max_speed,
			// th 6.3.00
			const Value max_accel) ;
	                //
	void assignBall(void) ;

	Team *assocTeam(Name name) ;
	void initPlayer(void) ; 
	void initCoach(Stadium *stad) ;
	void initOnlineCoach(Stadium *stad) ;
	Player* newPlayer(const Name teamname, Value version, int goalie_flag, 
						unsigned long host, int port) ;
	OnlineCoach* newCoach(long unsigned int, int, char *) ;
	Player* get_client_player(unsigned long host, int port) ;
	Player* newClient(unsigned long host, int port, char* init_message) ;
	Player* reconnectClient(unsigned long host, int port, char* init_message) ;
	void byeClient(Player *bye_player) ;

	void referee_get_foul(double x, double y, int side) ;
	void referee_drop_ball(double x, double y, int side) ;
	void discard_player(int side, int unum) ;
	void goalie_catch(Player *goalie) ;
	int mark_offside_mark(Player *kicker) ;
	void offside_check(void) ;
	void offside(void) ;

	void step() ;
	char* now() ;

	void referee_step() ;
	void check_mode() ;
	void check_goal() ;
	void check_ball_in_pitch() ;
	void set_ball(PVector p,Side s, PVector vel=PVector(0.0,0.0)) ;
	void set_ball(PVector p,Side s,Angle ang, PVector vel=PVector(0.0,0.0)) ;
	void check_players_move() ;
	void check_player_after_offside() ;
	void check_goalie_while_catch(Player *goalie) ;
	void check_field_before_goalkick() ;
	void set_players() ;
	void recovery_players() ;
	void check_player_in_field() ;
	void reset_player_comm_flag() ;
	void announce_goal(Team& tm) ;

	void change_play_mode(PlayMode pm) ;
        void drop_ball();

	int udp_init(int fd, int portnum) ;
	void udp_recv_message() ;
	void udp_send_message() ;
	void udp_send(const char *msg, Player &player,
			const Logical logp = TRUE) ;	
	void udp_send(const char *msg, unsigned long host, int port, int fd) ;
	void udp_send(const char *msg, unsigned long host, int port, int fd, int online) ;
	void udp_send(dispinfo_t *msg, unsigned long host, int port, int fd) ;
	void udp_send(dispinfo_t2 *msg, unsigned long host, int port,int fd) ;
	void send_init_info(Player& player) ;
	void send_reconnect_info(Player& player) ;
	void send_audio_info(char *info, Player *id, int olcoach_side, bool standard = false) ;
	void say(char *message, bool standard = true) ;

	Logical coachp() { return coach->assignedp ; } ;
	void udp_recv_from_coach() ;
	void udp_recv_from_online_coach() ;
	void move_object(char *obj_name, PVector pos,
		PVector vel=PVector(0.0,0.0)) ;
	void move_object(char *obj_name, PVector pos,Angle ang,
		PVector vel=PVector(0.0,0.0));
	void check_collision(char *obj_name) ;
	void check_collision(MPObject *obj) ;
	Player* get_client_coach(unsigned long host, int port) ;
	Player* get_player_by_name(char *name) ;
	BallPosInfo check_ball_pos() ;
	Logical cross_gline(Side side) ;

	void OpenForWrite() ;
	void WriteLog(dispinfo_t *) ;
	void WriteLog(dispinfo_t2* msg);

	void write_times(unsigned long, unsigned long);

	void send_parameter_info ( Player* );
	void send_parameter_info ( Coach* );
	void send_changed_players ( Player* );
	void send_changed_players ( Coach* );
	void broadcastSubstitution ( const int& side, const int& unum, const int& player_type, bool tell_ally_coach = false );

	inline double slowdownOnTop ( const PVector& pos, const Side& s )
	  {
	    if ( pos.y < 0.0 )
	      {
		if ( s == LEFT )
		  return server_param->slowness_on_top_for_left_team;
		else
		  return server_param->slowness_on_top_for_right_team;
	      }
	    else
	      return 1.0;
	  }
private:
	void ParsePlayerCommand(char* message, int iPlayer, unsigned long host, int port);
	bool ParseMonitorCommand(char* message, unsigned long host, int port);

	// [2000.07.21: I.Noda]
 public:
	inline void move_caught_ball() {
	   PVector temp(last_touch->relative_ball_ang.x,
			last_touch->relative_ball_ang.y);
	   check_goalie_while_catch(last_touch) ;
	   temp *= (last_touch->player_type->playerSize () +server_param->bsize);
	   ball->pos = last_touch->pos + temp;
	} ;
	inline void collisions () { motable.collisions ( mode ); }
} ;

/*
 *===================================================================
 *Part: ProtoTypes
 *===================================================================
 */

extern void _Start(Stadium& stad) ;

/*
 *===================================================================
 *Part: for Netif.C
 *===================================================================
 */

extern void start_udp(Stadium *stad) ;
