/* -*- Mode: C -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */
#include <winsock.h>

#undef V_INT
#undef V_BOOL
#undef V_NONE
#undef INFINITE
#undef max
#undef min

#include "hetroplayer.h"

bool HetroPlayer::seeded = false;

void HetroPlayer::seed (const PlayerParam* pp)
{
    if ( !HetroPlayer::seeded )
	{
          if (pp->randomSeed() >= 0) {
            cout << "Using given Hetero Player Seed: " << pp->randomSeed() << endl;
            srand ( pp->randomSeed() );
          } else {
            time_t now = time(NULL);
            
            cout << "Hetero Player Seed: " << now << endl;

            srand ( now );

          }
	HetroPlayer::seeded = true;
	}
}

Value HetroPlayer::delta ( Value min, Value max )
{
    return drand ( min, max );
}

HetroPlayer::HetroPlayer ( const ServerParam* sp, const PlayerParam* pp )
{
    HetroPlayer::seed (pp);

    Value tmp_delta = HetroPlayer::delta ( pp->playerSpeedMaxDeltaMin (), pp->playerSpeedMaxDeltaMax () );
    player_speed_max = sp->pspeed_max + tmp_delta;
    stamina_inc_max = sp->stamina_inc + tmp_delta * pp->staminaIncMaxDeltaFactor ();

    tmp_delta = HetroPlayer::delta ( pp->playerDecayDeltaMin (), pp->playerDecayDeltaMax () );
    player_decay = sp->pdecay + tmp_delta;
    inertia_moment = sp->inertia_moment + tmp_delta * pp->inertiaMomentDeltaFactor ();

    tmp_delta = HetroPlayer::delta ( pp->dashPowerRateDeltaMin (), pp->dashPowerRateDeltaMax () );
    dash_power_rate = sp->dprate + tmp_delta;
    player_size = sp->psize + tmp_delta * pp->playerSizeDeltaFactor ();

    tmp_delta = HetroPlayer::delta ( pp->kickableMarginDeltaMin (), pp->kickableMarginDeltaMax () );
    kickable_margin = sp->kmargin + tmp_delta;
    kick_rand = sp->kick_rand + tmp_delta * pp->kickRandDeltaFactor ();

    tmp_delta = HetroPlayer::delta ( pp->extraStaminaDeltaMin (), pp->extraStaminaDeltaMax () );
    extra_stamina = tmp_delta;
    effort_max = 1.0 + tmp_delta * pp->effortMaxDeltaFactor ();
    effort_min = sp->effort_min + tmp_delta * pp->effortMinDeltaFactor ();
}

HetroPlayer::HetroPlayer ( const ServerParam* sp, const PlayerParam* pp, int )
{
    player_speed_max = sp->pspeed_max;
    stamina_inc_max = sp->stamina_inc;

    player_decay = sp->pdecay;
    inertia_moment = sp->inertia_moment;

    dash_power_rate = sp->dprate;
    player_size = sp->psize;

    kickable_margin = sp->kmargin;
    kick_rand = sp->kick_rand;

    extra_stamina = pp->extraStaminaDeltaMin ();
    effort_max = 1.0;
    effort_min = sp->effort_min;
}

HetroPlayer::~HetroPlayer ()
{
}

ostream& operator<< ( ostream& o, const HetroPlayer& hp )
{
  o << "Hetro Player Type:" << endl;
  o << "\tPlayer speed max = " << hp.playerSpeedMax () << endl;
  o << "\tStamina inc = " << hp.staminaIncMax () << endl;
  o << "\tPlayer decay = " << hp.playerDecay () << endl;
  o << "\tInertia moment = " << hp.inertiaMoment () << endl;
  o << "\tDash power rate = " <<  hp.dashPowerRate () << endl;
  o << "\tPlayer size = " << hp.playerSize () << endl;
  o << "\tKickable margin = " << hp.kickableMargin () << endl;
  o << "\tKick rand = " << hp.kickRand () << endl;
  o << "\tExtra stamina = " << hp.extraStamina () << endl;
  o << "\tEffort max = " << hp.effortMax () << endl;
  o << "\tEffort min = " << hp.effortMin () << endl;
  return o;
}

void HetroPlayer::getSettingsString ( int id, char* buf )
{
    ostrstream ost(buf, MaxStringSize) ;

    ost << "(player_type " << id << " "
	<< player_speed_max << " "
	<< stamina_inc_max << " "
	<< player_decay << " "
	<< inertia_moment << " "
	<< dash_power_rate << " "
	<< player_size << " "
	<< kickable_margin << " "
	<< kick_rand << " " 
	<< extra_stamina << " "
	<< effort_max << " "
	<< effort_min
	<< ")" << endl << ends;
}

player_type_t HetroPlayer::convertToStruct ( int id )
{
    player_type_t tmp;

    tmp.id = htons ( id );
    tmp.player_speed_max = htonl ( (long)rint((player_speed_max * SHOWINFO_SCALE2) ));
    tmp.stamina_inc_max = htonl ( (long)rint((stamina_inc_max * SHOWINFO_SCALE2) ));
    tmp.player_decay = htonl ( (long)rint((player_decay * SHOWINFO_SCALE2) ));
    tmp.inertia_moment = htonl ( (long)rint((inertia_moment * SHOWINFO_SCALE2) ));
    tmp.dash_power_rate = htonl ( (long)rint((dash_power_rate * SHOWINFO_SCALE2) ));
    tmp.player_size = htonl ( (long)rint((player_size * SHOWINFO_SCALE2) ));
    tmp.kickable_margin = htonl ( (long)rint((kickable_margin * SHOWINFO_SCALE2) ));
    tmp.kick_rand = htonl ( (long)rint((kick_rand * SHOWINFO_SCALE2) ));
    tmp.extra_stamina = htonl ( (long)rint((extra_stamina * SHOWINFO_SCALE2) ));
    tmp.effort_max = htonl ( (long)rint((effort_max * SHOWINFO_SCALE2) ));
    tmp.effort_min = htonl ( (long)rint((effort_min * SHOWINFO_SCALE2) ));

    return tmp;
}
