/* -*- Mode: C++ -*-
 *Header:
 *File: netif.C (for C++ & cc)
 *Author: Noda Itsuki
 *Date: 1996/02/14
 *EndHeader:
 */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */


// [2000/01/14:I.Noda]
//  escape prototype of recvfrom 
//  for escape conversion of signed-unsigned ints 
//
#include <winsock.h>

#undef V_INT
#undef V_BOOL
#undef V_NONE
#undef INFINITE
#undef max
#undef min

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>
#include <fcntl.h>

#include <iostream>
#include <strstream>
using namespace std;

#include "param.h"
#include "types.h"
#include "utility.h"
#include "object.h"
#include "field.h"
#include "netif.h"
extern "C" {
  int parse_coach_lang_string(char*); //defined in coach_lang.l
}



void write_log(Stadium& stad, Player& p, const char *message, int flag) ;
void write_log(Stadium& stad, Coach& p, const char *message, int flag) ;
void write_log(Stadium& stad, OnlineCoach& p, const char *message, int flag) ;

Value distance(PVector *pos1, PVector *pos2) ;

static char *PlayModeString[] = PLAYMODE_STRINGS ;


/*
 *===================================================================
 *Part: Initialize 
 *===================================================================
 */

int Stadium::udp_init(int fd, int portnum)
{
        struct sockaddr_in serv_addr, olcoach_addr, coach_addr ;
        unsigned long one ;

        if (fd >= 0) {
                /* Open UDP socket. */
                if ((socketfd[fd] = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't open socket." << endl ;
                   return errno ;
                }

                /* Set non-blocking. */
                one = 1 ;
                if(ioctlsocket(socketfd[fd], FIONBIO, &one) < 0 ){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't ioctl on socket" << endl ;
                   return errno ;
                   
                }

                /* Bind local address */
                //bzero((char *) &serv_addr, sizeof(serv_addr)) ;
                memset((char *) &serv_addr, 0, sizeof(serv_addr)) ;
                serv_addr.sin_family            = AF_INET ;
                serv_addr.sin_addr.s_addr       = htonl(INADDR_ANY) ;
                serv_addr.sin_port                      = htons(portnum) ;

                if(bind(socketfd[fd], (struct sockaddr *) &serv_addr,
                        sizeof(serv_addr)) < 0){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't bind local address:" << portnum << endl ;
                   return errno ;
                }

                if( fd != 0) return 0 ;

                /* Open olcoach socket. */
                if ((olcoachfd[0] = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't open socket." << endl ;
                   return errno ;
                }
        
                /* Set non-blocking. */
                one = 1 ;
                if(ioctlsocket(olcoachfd[0], FIONBIO, &one) < 0 ){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't ioctl on socket" << endl ;
                   return errno ;
                }
        
                /* Bind local address */
                //bzero((char *) &olcoach_addr, sizeof(serv_addr)) ;
                memset((char *) &olcoach_addr, 0, sizeof(serv_addr)) ;
                olcoach_addr.sin_family         = AF_INET ;
                olcoach_addr.sin_addr.s_addr    = htonl(INADDR_ANY) ;
                olcoach_addr.sin_port           = htons(server_param->olcoach_pnum) ;
        
                if(bind(olcoachfd[0], (struct sockaddr *) &olcoach_addr,
                        sizeof(serv_addr)) < 0){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't bind local address:" 
                        << server_param->olcoach_pnum
                        << endl ;
                   return errno ;
                }

                if(server_param->CoachMode || server_param->CwRMode) {
                        /* open coach socket */
                        if((coach->socketfd = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
                           // [2001.01.02:I.Noda]
                           // make to show error message always
                           //if (server_param->verbose)
                           cerr << "Can't open socket." << endl ;
                           return errno ;
                        }
        
                        one = 1 ;
                        if(ioctlsocket(coach->socketfd, FIONBIO, &one) < 0) {
                           // [2001.01.02:I.Noda]
                           // make to show error message always
                           //if (server_param->verbose)
                           cerr << "Can't ioctl on socket." << endl ;
                           return errno ;
                        }

                        //bzero((char *) &coach_addr, sizeof(coach_addr)) ;
                        memset((char *) &coach_addr, 0, sizeof(coach_addr)) ;
                        coach_addr.sin_family           = AF_INET ;
                        coach_addr.sin_addr.s_addr      = htonl(INADDR_ANY) ;
                        coach_addr.sin_port                     = htons(server_param->coach_pnum) ;

                        if(bind(coach->socketfd, (struct sockaddr *) &coach_addr,
                                                sizeof(coach_addr)) < 0){
                           // [2001.01.02:I.Noda]
                           // make to show error message always
                           cerr << "Can't bind local address: " 
                                << server_param->coach_pnum << endl ;
                           return errno ;
                        }
                }

                return 0 ;
        }else { /* online coach (fd < 0) */
                fd *= -1 ;
                /* Open UDP socket. */
                if ((olcoachfd[fd] = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't open socket." << endl ;
                   return errno ;
                }

                /* Set non-blocking. */
                one = 1 ;
                if(ioctlsocket(olcoachfd[fd], FIONBIO, &one) < 0 ){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't ioctl on socket" << endl ;
                   return errno ;
                }

                /* Bind local address */
                //bzero((char *) &olcoach_addr, sizeof(olcoach_addr)) ;
                memset((char *) &olcoach_addr, 0, sizeof(olcoach_addr)) ;
                olcoach_addr.sin_family         = AF_INET ;
                olcoach_addr.sin_addr.s_addr    = htonl(INADDR_ANY) ;
                olcoach_addr.sin_port           = htons(portnum) ;

                if(bind(olcoachfd[fd], (struct sockaddr *) &olcoach_addr,
                        sizeof(olcoach_addr)) < 0){
                   // [2001.01.02:I.Noda]
                   // make to show error message always
                   //if (server_param->verbose)
                   cerr << "Can't bind local address:" << portnum << endl ;
                   return errno ;
                }

                olcoach[fd]->socketfd = olcoachfd[fd] ;
                        
                return 0 ;
        }

}

/*
 *===================================================================
 *Part: Recv
 *===================================================================
 */
void Stadium::udp_recv_message()
{
        static char message[MaxMesg] ;

        int i, num ;
        int len;
        int iMsgLength;
        int buf, sw ;

        /* randomize order */
        for(i = 0 ; i < MAX_PLAYER * 2; i++) 
        {
                sw = irand(MAX_PLAYER * 2) ;
                buf = dest[i] ;
                dest[i] = dest[sw] ;
                dest[sw] = buf ;
        }

        /* receive from socket for server(i=0) and players(i=1...) */
        for ( i = 0 ; i <= MAX_PLAYER * 2; i++) 
        {
                if (i != 0) 
                {
                        num = dest[i-1] + 1 ;
                        if (socketfd[num] == -1)
                                continue ;
                }

                struct sockaddr_in cli_addr ;
                len = sizeof(cli_addr) ;

                /* read all messages of a player */ 
                iMsgLength = 1 ; 
                while (iMsgLength >= 0) 
                {
                        if (i == 0)             /* server socket */
                                iMsgLength = recvfrom(socketfd[0], message, MaxMesg, 0,
                                                      (struct sockaddr *) &cli_addr, &len) ;
                        else                    /* client socket */
                                iMsgLength = recvfrom(socketfd[num], message, MaxMesg, 0,
                                                     (struct sockaddr *) &cli_addr, &len) ;

                        /* treat errors */
                        if (iMsgLength == -1 && errno == WSAEWOULDBLOCK) 
                                // no more messages of this player
                                continue;
                        else if(iMsgLength == -1 && errno != WSAENOTSOCK)
                        {
                                if (server_param->verbose)
                                        cerr << "recvfrom error" << endl ;
                                continue;
                        }
                        else if(iMsgLength == 0)
                                continue ;

                        unsigned long host = ntohl(cli_addr.sin_addr.s_addr) ;
                        int port = ntohs(cli_addr.sin_port) ;

                        /* chop newline */
                        if(message[iMsgLength - 1] == '\n') 
                                iMsgLength--;
                        message[iMsgLength] = NULLCHAR ;

                        /* soccermonitor commands */
                        if (ParseMonitorCommand(message, host, port))
                                continue;

                        /* player commands */
                        ParsePlayerCommand(message, i, host, port);
                }
        }
}


void Stadium::ParsePlayerCommand(char *message, int iPlayer, unsigned long host, int port)
{
        Player *p ;
        static char tmp[MaxStringSize] ;

        p = get_client_player(host,port) ;
        if(p == (Player *)NULL && iPlayer == 0) 
        {
                /* receive to server's port */
                sscanf(message,"(%s",tmp);
                if (!strcmp(tmp,"reconnect")) 
                {
                        // a player reconnects to the server
                        if (mode == PM_PlayOn)
                                udp_send("(can't reconnect)", host, port, 0) ;
                        else 
                        {
                                p = reconnectClient(host,port,message) ;
                                if(p == (Player *)NULL)
                                        udp_send("(error reconnect)", host,port,0) ;
                                else
                                        send_reconnect_info(*p) ;
                        }
                }
                else 
                {
                        // a new player connects to the server
                        p = newClient(host,port,message) ;
                        if(p == (Player *)NULL) 
                                udp_send("(error no_more_team_or_player_or_goalie)", host,port,0) ;
                        else
                                send_init_info(*p) ;
                }
                write_log(*this, *p, message, RECV) ;
        }
        else if (p != (Player *)NULL && iPlayer != 0) 
        {
                /* receive to player's port */
                p->parse_command(message) ;
                write_log(*this, *p, message, RECV) ;
        }
        else 
        {
                if (server_param->verbose)
                        cerr << "command error" << endl ;
        }
}

bool Stadium::ParseMonitorCommand(char *message, unsigned long host, int port)
{
    int n, ver = 1;

	if (!strcmp(message,"(dispinit)")) 
        {
        cerr << "a new (v1) monitor connected\n";
                displist_t *p ;

                for (p = &top ; p->next != NULL ; p = p->next)
                        /* NULL statement */ ;

                p->next = (displist_t *)malloc(sizeof(displist_t)) ;
                if (p->next == NULL) {
                        perror("malloc") ;
                        exit(1) ;
                }

                p->next->shost = host ;
                p->next->sport = port ;
                p->next->version = 1;
                p->next->next = NULL ;
        }
        else if (sscanf(message,"(dispinit version %d)",&ver) == 1)
            {
            cerr << "a new (v" << ver << ") monitor connected\n";
                // a new monitor connected
                displist_t *p ;

                for (p = &top ; p->next != NULL ; p = p->next)
                        /* NULL statement */ ;

                p->next = (displist_t *)malloc(sizeof(displist_t)) ;
                if (p->next == NULL) {
                        perror("malloc") ;
                        exit(1) ;
                }

                p->next->shost = host ;
                p->next->sport = port ;
                p->next->version = ver;
                p->next->next = NULL ;

                // send server parameter information to monitor
                if ( ver >= 2 )
                    {
                    dispinfo_t2 di;

                    di.mode = htons ( PARAM_MODE );
                    di.body.sparams = server_param->convertToStruct ();
                    udp_send(&di, host, port,0) ;

                    di.mode = htons ( PPARAM_MODE );
                    di.body.pparams = player_param->convertToStruct ();
                    udp_send(&di, host, port,0) ;

                    di.mode = htons ( PT_MODE );
                    for ( int i = 0; i < player_param->playerTypes (); i++ )
                      {
                        di.body.ptinfo = player_types [ i ]->convertToStruct ( i );
                        udp_send(&di, host, port,0) ;
                      }
                    }
            

            }
        else if (!strcmp(message,"(dispstart)")) 
        {
                // kick off message
                _Start(*this) ;
        }
        else if (!strncmp(message,"(dispplayer", 11)) 
        {
                // a player is given new position by the monitor
                int side, unum ;
                int x, y, a;
                double real_x, real_y, angle;
                n = sscanf(message,"(dispplayer %d %d %d %d %d)", &side, &unum, &x, &y, &a) ;
                if (n == 5) {
		  real_x= x / SHOWINFO_SCALE;
		  real_y= y / SHOWINFO_SCALE;
		  angle= Deg2Rad(a);
		  
		  //D.N.
		  if(unum < 1 || unum > MAX_PLAYER)
		  {
			  cerr << "Monitor poslal nespravny prikaz! Oficialny server by teraz spadol!\n"
				<< message << endl;
			  return false;
		  }

		  Player* p = (side == LEFT ? team_l : team_r)->player[unum-1];
		  p->pos = PVector(real_x, real_y);
		  p->vel = 0;
		  p->angle= angle;
		  //check_collision(p);
		  collisions ();
		}
	}
        else if (!strncmp(message,"(dispdiscard", 12)) 
        {
                // a player is discarded by the monitor
                int side, unum ;
                n = sscanf(message,"(dispdiscard %d %d)", &side, &unum) ;
                if (n == 2)
                        discard_player(side, unum) ;
        }
        else 
        {
                // foul or drop_ball
                int x, y, side ;
                double real_x, real_y ;
                n = sscanf(message,"(dispfoul %d %d %d)", &x, &y, &side) ;
                        
                if (n == 3) {
                        real_x = x / SHOWINFO_SCALE ;
                        real_y = y / SHOWINFO_SCALE ;
                        if ( side == NEUTRAL )
                                referee_drop_ball(real_x, real_y, side) ;
                        else
                                referee_get_foul(real_x, real_y, side) ;
                }
                else
                        // no message sent by a soccermonitor
                        return false;
        }
        return true;
}

void Stadium::udp_recv_from_coach()
{
        char message[MaxMesg] ;
        struct sockaddr_in cli_addr ;
	int n;
        int len ;

        len = sizeof(cli_addr) ;

	/* read all the coach messages */
	int iMsgLength = 1 ; 
	while (iMsgLength >= 0) {

	    iMsgLength = recvfrom(coach->socketfd, message, MaxMesg, 0,
				  (struct sockaddr *)&cli_addr, &len) ;
	    
	    if(iMsgLength < 0) {
                if(errno != WSAEWOULDBLOCK && errno != WSAENOTSOCK)
		    if (server_param->verbose)
			cerr << "recvfrom error" << endl ;
                return ;
	    }
	    /* chop newline */
	    if(iMsgLength != 0 && message[iMsgLength-1] == '\n')
		iMsgLength-- ;
	    message[iMsgLength] = NULLCHAR ;

	    unsigned long host = ntohl(cli_addr.sin_addr.s_addr) ;
	    int port = ntohs(cli_addr.sin_port) ;

	    if ( coach->assignedp == True )
		{
		    coach->parse_command(*this,message,host,port) ;	    
		}
	    else
		{
		    n = coach->parse_init(*this,message,host,port) ;
		    if ( n == -1 )
			{
			    coach->assignedp = True ;
			    _Start(*this) ;  // need to remove this line if we dont want the server to start when the coach connects

			    coach->parse_command(*this,message,host,port) ;
			}
		    else if ( n == 1 )
			{
			    coach->assignedp = True ;
			}
		}
	
	    write_log(*this, *coach, message, RECV) ;
	}
}

void Stadium::udp_recv_from_online_coach()
{
    static char message[MaxMesg] ;
    static char tmp[MaxStringSize] ;
    unsigned long host ;
    int i, n, port, iMsgLength;
    int len;
    OnlineCoach* p ;

    /* receive from socket for server(i=0) and online coachs(i=1, 2) */
    for ( i = 0 ; i < 3; i++) {
	if (i != 0) {
	    if (olcoach[i]->assignedp == FALSE)
		continue ;
	}

	struct sockaddr_in cli_addr ;
	len = sizeof(cli_addr) ;

	n = 1 ; 
	iMsgLength = 1 ; 

	while (iMsgLength >= 0) {
	    if (i == 0)             /* server socket */
		iMsgLength = recvfrom(olcoachfd[0], message, MaxMesg, 0,
				      (struct sockaddr *) &cli_addr, &len) ;
	    else                    /* client socket */
		iMsgLength = recvfrom(olcoachfd[i], message, MaxMesg, 0,
				      (struct sockaddr *) &cli_addr, &len) ;

	    host = ntohl(cli_addr.sin_addr.s_addr) ;
	    port = ntohs(cli_addr.sin_port) ;

	    if (iMsgLength > 0) {
		if(iMsgLength == 0) continue ;
                                /* chop newline */
		if(message[iMsgLength-1] == '\n') iMsgLength--;
		message[iMsgLength] = NULLCHAR ;
			    
		if(i == 0) {
				/* receive to server's port */
		    n = sscanf(message,"(%s",tmp);
				//                                      if (!strcmp(tmp,"reconnect")) {
				//                                      }
		    p = newCoach(host,port,message) ;
		    if(p == (OnlineCoach *)NULL)
			udp_send("(error no_such_team_or_already_have_coach)",
				 host,port,0, TRUE) ;
		    else
			write_log(*this, *p, message, RECV) ;
				
		}
		else {
				/* receive to coach's port */
		    olcoach[i]->parse_command(*this,message,
					      host,port) ;
		    write_log(*this, *olcoach[i], message, RECV) ;
				
		}
	    } else if(errno != WSAENOTSOCK && errno == WSAEWOULDBLOCK)
		if (server_param->verbose)
		    cerr << "recvfrom error" << endl ;
	}
    }
    
}

void chop_last_parenthesis(char *str, int max_size)
{
        int l = strlen(str) ;

         if(l > max_size) {
                str[max_size] = NULLCHAR ;
         }
         else {
                l-- ;
                if(str[l] == ')') str[l] = NULLCHAR ;
         }
}



void Player::parse_command(char *command)
{
        char com[MaxStringSize] ;
        int n ;
        double effective_dash_power ;

        for ( ; ; ) {
                if ((n = sscanf(command,"(%[^ )]",com)) < 1)
                        break ;

                if(alive & DISCARD)
                        return ;

                if(!strcmp(com,"dash")) {
                        double power, power_need ;
                        if ((n = sscanf(command,"(%s %lf)",com,&power)) < 2)
                                break ;
                        if (!command_done) {
                                // culculate stamina
//                              power = NormalizeDashPower(power) ;
                                power_need = (power < 0) ? NormalizeDashPower(power) * -2 
                                                                : NormalizeDashPower(power) ;
                                if (power_need > stamina + player_type->extraStamina () )
                                        power_need = stamina + player_type->extraStamina ();
                                stamina -= power_need ;
                                if ( stamina < 0.0 )
                                  stamina = 0.0;
                                // 
                                power = (power < 0) ? power_need / -2 : power_need ;
                                effective_dash_power = effort * power ;
                                dash(ReduceDashPower(effective_dash_power)) ;
                                count_dash++ ;
                                command_done = TRUE ;
                        }
                }
                else if (!strcmp(com,"turn")) {
                        double moment ;
                        if ((n = sscanf(command,"(%s %lf)",com,&moment)) < 2)
                                break ;
                        if (!command_done) {
                                turn(NormalizeMoment(moment)) ;
                                count_turn++ ;
                                command_done = TRUE ;
                        }
                }
                else if (!strcmp(com,"turn_neck")) {
                        double moment ;
                        if ((n = sscanf(command,"(%s %lf)",com,&moment)) < 2)
                                break ;
                        if (!turn_neck_done) {
                                turn_neck(NormalizeNeckMoment(moment)) ;
                                count_turn_neck++ ;
                                turn_neck_done = TRUE ;
                        }
                }
                else if (!strcmp(com,"kick")) {
                        double power,dir ;
                        if ((n = sscanf(command,"(%s %lf %lf)",com,&power,&dir)) < 3)
                                break ;
                        if (!command_done) {
                                kick(NormalizeKickPower(power),NormalizeMoment(dir)) ;
                                count_kick++ ;
                                command_done = TRUE ;
                        }
                }
                else if (!strcmp(com,"catch")) {
                        double dir ;
                        if ((n = sscanf(command, "(%s %lf)",com,&dir)) < 2)
                                break ;
                        if (!command_done) {
                                goalie_catch(NormalizeMoment(dir)) ;
                                count_catch++;
                                command_done = TRUE ;
                        }
                }
                else if (!strcmp(com,"say")) {
                        char msg[MaxStringSize] ;
                        if ((n = sscanf(command,SAY_MESSAGE_SCAN_FORMAT,com,msg)) < 2)
                                break ;
                        chop_last_parenthesis(msg, stadium->server_param->SayMsgSize) ;
                        say(msg) ;
                        count_say++ ;
                }
                else if (!strcmp(com,"sense_body")) {
                        sense_body() ;
                }
		else if (!strcmp(com,"score"))
		    {
		    send_score ();
		    }
                else if (!strcmp(com,"move")) {
                        double x, y ;
                        /* (move X Y) */
                        n = sscanf(command,"(%s %lf %lf)",com,&x,&y) ;
                        if(n<3)
                                /* (move (X Y)) */
                                n = sscanf(command,"(%s (%lf %lf))",com,&x,&y) ;
                        if(n<3)
                                break ;
                        if (!command_done) {
                                if( stadium->mode == PM_BeforeKickOff ||
                                        stadium->mode == PM_AfterGoal_Right ||
#ifndef _FREEKICK_MOVE_ALL_
                                        stadium->mode == PM_AfterGoal_Left) {
#else
                                        stadium->mode == PM_AfterGoal_Left ||
                                        stadium->mode == PM_FreeKick_Left ||
                                        stadium->mode == PM_FreeKick_Right ){
#endif
                                        pos.x = x * team->side ;
                                        pos.y = y * team->side ;
                                        //stadium->check_collision(this) ;
					stadium->collisions ();
                                }
#ifndef _FREEKICK_MOVE_ALL_
                                else if ((stadium->mode == PM_FreeKick_Left ||
                                          stadium->mode == PM_FreeKick_Right) &&
                                         stadium->caught_ball == team->side &&
                                         goalie) {
                                  if (stadium->server_param->goalie_max_moves < 0 ||
                                      goalie_moves_since_catch <
                                      stadium->server_param->goalie_max_moves) {
                                           pos.x = x * team->side ;
                                           pos.y = y * team->side ;
                                           goalie_moves_since_catch++;
                                           stadium->move_caught_ball() ; // [2000.07.21:I.Noda] 
                                           //stadium->check_collision(this) ;
					   stadium->collisions ();
                                  } else
                                    udp_send("(error too_many_moves)") ;
                                }
#endif
                                else
                                        break ;
                                command_done = TRUE ;
                                count_move++;
                        }
                }
                else if (!strcmp(com, "change_view")) {
                        char awd[16], qly[16] ;
                        if ((n = sscanf(command, "(%s %s %[^ )]", com, awd, qly)) < 3)
                                break ;

                        if (!strcmp(awd, "narrow")) {
                                vis_angle = defangle / 2 ;      
                                vis_send = 2 ;
                        }
                        else if (!strcmp(awd, "normal")) {
                                vis_angle = defangle ;
                                vis_send = 4 ;
                        }
                        else if (!strcmp(awd, "wide")) {
                                vis_angle = defangle * 2 ;
                                vis_send = 8 ;
                        }
                        else
                                break ;

                        if (!strcmp(qly, "high"))
                                highquality = TRUE ;
                        else if (!strcmp(qly, "low")) {
                                vis_send /= 2 ;
                                highquality = FALSE ;
                        }
                        else
                                break ;

                        count_change_view++ ;
                }
                else if (!strcmp(com, "bye")) {
                        stadium->byeClient(this) ;
                        command_done = TRUE ;
                        return ;
                } 
		//pfr:SYNCH
                else if (!strcmp(com, "done")) {
		        done_received = TRUE;
                        return ;
                } 
                else {
                        udp_send("(error unknown_command)") ;
                        return ;
                }

                n = 0 ;
                for(;;) {
                        if (*command == '(')
                                n++ ;
                        else if (*command == ')') {
                                n-- ;
                                if (n == 0) {
                                        command++ ;
                                        if (*command == NULLCHAR)
                                                return ;
                                        else
                                                break ;
                                }
                        }
                        else if (*command == NULLCHAR) {
                                return ;
                        }
                        command++ ;
                }
        }
        udp_send("(error illegal_command_form)") ;
        return ;
}

void Stadium::say(char *message, bool standard)
{
        displist_t *p ;

        send_audio_info(message, NULL, FALSE, standard) ;

        if (server_param->logfile[0] != NULLCHAR || server_param->rec_log || server_param->send_log) {
                minfo.body.msg.board = htons(MSG_BOARD) ;
                sprintf(minfo.body.msg.message, "(%s %s)\n", REFEREE_NAME, message) ;

                if (server_param->logfile[0] != NULLCHAR) {
                        static char tmp[MaxStringSize] ;

                        sprintf(tmp, "%d\t%s", time, minfo.body.msg.message ) ;
                        fputs(tmp, log_fp) ;
                        fflush(log_fp) ;
                }

                if (server_param->send_log)
                        for (p = top.next ; p != NULL ; p = p->next)
                            {
                            if ( p->version == 1 )
                                udp_send(&minfo,p->shost,p->sport,0) ;
                            else
                                {
                                dispinfo_t2 minfo2;
                                minfo2.mode = minfo.mode;
                                minfo2.body.msg = minfo.body.msg;
                                udp_send(&minfo2,p->shost,p->sport,0) ;
                                }
                            }
                if (server_param->recfile[0] != NULLCHAR && server_param->rec_log == TRUE)
                        if (mode != PM_BeforeKickOff && mode != PM_TimeOver)
                                WriteLog(&minfo) ;
        }
}

void Player::say(char *message)
{
        Stadium &stad = *(this->stadium) ;
        displist_t *p ;

        stad.send_audio_info(message, this, FALSE) ;

        if (stad.server_param->rec_log || stad.server_param->send_log) {
                stad.minfo.body.msg.board = htons(MSG_BOARD) ;
                sprintf(stad.minfo.body.msg.message, "(%s %s)\n", name, message) ;

                if (stad.server_param->send_log)
                        for (p = stad.top.next ; p != NULL ; p = p->next)
                            {
                            if ( p->version == 1 )
                                stad.udp_send(&(stad.minfo),p->shost,p->sport,0) ;
                            else
                                {
                                dispinfo_t2 minfo2;
                                minfo2.mode = stad.minfo.mode;
                                minfo2.body.msg = stad.minfo.body.msg;
                                stad.udp_send(&minfo2,p->shost,p->sport,0) ;
                                }
                            }

                if (stad.server_param->recfile[0] != NULLCHAR && stad.server_param->rec_log == TRUE)
                        if (stad.mode != PM_BeforeKickOff && stad.mode != PM_TimeOver)
                                stad.WriteLog(&stad.minfo) ;
        }
}

int OnlineCoach::check_message_queue(TheNumber time)
{
  if (message_queue.size() <= 0)
    return 0;
  int messages_sent = 0;
  for (int i=stadium->server_param->clang_mess_per_cycle;
       i > 0 && message_queue.size() > 0 ;
       i++) {
    CM_Message* pMess = *(message_queue.begin());
    if (stadium->mode != PM_PlayOn ||
        time - pMess->getTimeRecv() >= stadium->server_param->clang_mess_delay) {
      message_queue.pop_front();
      //NOTE: if you change the way that coach messages are printed, you will need
      // to modify param.h: MaxCoachMesg and MaxCoachMesgWTime
      ostrstream coach_mess(message_buf, MaxMesg);
      coach_mess << pMess->getTimeRecv() << " ";
      pMess->Print(coach_mess);
      coach_mess << ends; //null-terminate
      if (coach_mess.pcount() > MaxCoachMesgWTime) {
	char buf[100];
	sprintf(buf, "(error say_message_too_long %d)", pMess->getTimeRecv());
	udp_send(buf);
      }	else {
	say(coach_mess.str(), true);
	//cout << "Coach says: " << coach_mess.str() << endl;
      }
      delete pMess;
      ++messages_sent;
    } else {
      break;
    }
  }
  return messages_sent;
}
void OnlineCoach::say(char *message, bool standard)
{
        Stadium &stad = *(this->stadium) ;
        displist_t *p ;

        stad.send_audio_info(message, NULL, this->side, standard) ;

        if (stad.server_param->rec_log || stad.server_param->send_log) {
                stad.minfo.body.msg.board = htons(MSG_BOARD) ;
		char format[40];
		sprintf(format, "(%%s %%.%ds)\n",
			max_message_length_for_display - (MaxMesg - MaxCoachMesg));
                sprintf(stad.minfo.body.msg.message, format,
                        (side == RIGHT) ? OLCOACH_NAME_R : OLCOACH_NAME_L, message) ;

                if (stad.server_param->send_log)
                        for (p = stad.top.next ; p != NULL ; p = p->next)
                            {
                            if ( p->version == 1 )
                                stad.udp_send(&(stad.minfo),p->shost,p->sport,0) ;
                            else
                                {
                                dispinfo_t2 minfo2;
                                minfo2.mode = stad.minfo.mode;
                                minfo2.body.msg = stad.minfo.body.msg;
                                stad.udp_send(&minfo2,p->shost,p->sport,0) ;
                                }
                            }

                if (stad.server_param->recfile[0] != NULLCHAR && stad.server_param->rec_log == TRUE)
                        if (stad.mode != PM_BeforeKickOff && stad.mode != PM_TimeOver)
                                stad.WriteLog(&stad.minfo) ;
        }
}

void Stadium::send_audio_info(char *info, 
                              Player *id, 
                              int olcoach_side,
                              bool standard)
{
        int i,sw,buf ;
        char tmp[MaxStringSize] ;
        char tmp2[MaxStringSize] ;
        char tmp_name[MaxStringSize] ;
        char tmp_name_new[MaxStringSize] ;
        ostrstream ost(tmp, MaxStringSize) ;
        ostrstream ost2(tmp2, MaxStringSize) ;
        PVector *source_pos ;

        /* randomize order */
        for(i = 0 ; i < MAX_PLAYER * 2; i++) {
                sw = irand(MAX_PLAYER * 2) ;
                buf = dest[i] ;
                dest[i] = dest[sw] ;
                dest[sw] = buf ;
        }

        if ( id == NULL) {
                if ( olcoach_side == FALSE && standard ) {
                        ost << "(hear " << time << " "
                                << REFEREE_NAME << " "
                                << info << ")" << endl << ends ;
                        for(i = 0 ; i < MAX_PLAYER * 2; i++) {
                                if (player[dest[i]]->alive == DISABLE)
                                    continue ;
                                player[dest[i]]->send_audio_info(tmp) ;
                        }
                }
		// non standard messeges with no side are from the coach
		else if ( olcoach_side == FALSE )
		{
			ost << "(hear " << time << " "
 			    << REFEREE_NAME << " "
                            << info << ")" << endl << ends ;
			ost2 << "(hear " << time << " "
			    << COACH_NAME << " \""
			    << info << "\")" << endl << ends ;		
			
			for(i = 0 ; i < MAX_PLAYER * 2; i++) 
			{
			        if (player[dest[i]]->alive == DISABLE)
                			continue ;

			        if ( player[dest[i]]->version >= 7.0 )
          				player[dest[i]]->send_audio_info(tmp2) ;
        			else
          				player[dest[i]]->send_audio_info(tmp) ;
			}	
	
		}
                else {
                        ost << "(hear " << time << " "
			    << ((olcoach_side == RIGHT) ? OLCOACH_NAME_R : OLCOACH_NAME_L )<< " "
			    << " " << info << ")" << endl << ends ;

			/* th: only quote the coach message if it is not standard.
			       (i.e. Pre v7.0 coach) */ 
                        ost2 << "(hear " << time << " "
			     << ((olcoach_side == RIGHT) ? OLCOACH_NAME_R : OLCOACH_NAME_L )
			     << " ";
			if ( !standard )
				ost2 << "\""; 
	
                        ost2 << info;
			
			if ( !standard )
				ost2 << "\"";

			ost2 << ")" << endl << ends ;

                        for(i = 0 ; i < MAX_PLAYER * 2; i++) {
                                if (player[dest[i]]->alive == DISABLE)
                                        continue ;
                                if (player[dest[i]]->team->side != olcoach_side
				    && standard)
                                  continue; //pfr: for v7.00: only send to
                               	// th: if it's a standard message

 
				if ( player[dest[i]]->version >= 7.0 )
                                  player[dest[i]]->send_audio_info(tmp2) ;
				else	
				  player[dest[i]]->send_audio_info(tmp) ;
                        }
                }
        }
        else {
                for(i = 0 ; player[dest[i]] != id ; i++ ) ;
                source_pos = &(player[dest[i]]->pos) ;

                for(i = 0 ; i < MAX_PLAYER * 2; i++ ) {
                        if (player[dest[i]]->alive == DISABLE)
                                continue ;
                        if (distance(source_pos, &(player[dest[i]]->pos)) <= server_param->audio_dist) {
                                if (player[dest[i]] == id) {
                                if (player[dest[i]]->version >= 7.0 )
                                    {
                                        sprintf(tmp, "(hear %d self \"", time) ;
                                        strcat(tmp, info) ;
                                        strcat(tmp, "\")\n") ;
                                    }
                                else
                                    {
                                    sprintf(tmp, "(hear %d self ", time) ;
                                    strcat(tmp, info) ;
                                    strcat(tmp, ")\n") ;
                                    }
                                        player[dest[i]]->send_audio_info(tmp) ;
                                }
                                else {
                                    if (player[dest[i]]->version >= 7.0 )
                                        {                       
                                        sprintf(tmp, "(hear %d %d \"", time,
                                                ((player[dest[i]]->version >= 7.0 ) ?
						 Rad2IDegRound(player[dest[i]]->vangle(*id)) :						 Rad2IDeg(player[dest[i]]->vangle(*id)))
						 ) ;
                                        strcat(tmp, info) ;
                                        strcat(tmp, "\")\n") ;
                                        }
                                    else
                                        {
                                        sprintf(tmp, "(hear %d %d ", time,
                                                ((player[dest[i]]->version >= 7.0 ) ?
						 Rad2IDegRound(player[dest[i]]->vangle(*id)) :
						 Rad2IDeg(player[dest[i]]->vangle(*id)))
						) ;
                                        strcat(tmp, info) ;
                                        strcat(tmp, ")\n") ;
                                        }
                                        if (player[dest[i]]->team == id->team) {
                                                if (player[dest[i]]->hear_capacity_from_teammate
                                                                                                                        >= server_param->hear_decay) {
                                                        player[dest[i]]->send_audio_info(tmp) ;
                                                        player[dest[i]]->hear_capacity_from_teammate
                                                                                                                        -= server_param->hear_decay ;
                                                }
                                        }
                                        else {
                                                if (player[dest[i]]->hear_capacity_from_opponent
                                                                                                                        >= server_param->hear_decay) {
                                                        player[dest[i]]->send_audio_info(tmp) ;
                                                        player[dest[i]]->hear_capacity_from_opponent
                                                                                                                        -= server_param->hear_decay ;
                                                }
                                        }
                                }
                        }
                }
        }
	bool ref = false;
	
    if ( id == NULL ){
                switch( olcoach_side ){
                        case RIGHT:
                                strcpy(tmp_name,OLCOACH_NAME_R);
                                break;
                        case LEFT:
                                strcpy(tmp_name,OLCOACH_NAME_L);
                                break;
		        default:
			        ref = true;
                                strcpy(tmp_name,REFEREE_NAME);
                }
		strcpy(tmp_name_new, tmp_name);
        }else{
                strcpy(tmp_name,((Player *)id)->name);
                strcpy(tmp_name_new,((Player *)id)->short_name);
        }

        if ((server_param->CoachMode || server_param->CwRMode) && coach->hear) {
                if (server_param->old_hear)     /* old format */
                        sprintf(tmp, "(hear %s %d %s)\n", tmp_name, time, info) ;
                else                    /* new format (same as player's) */
                    {
		      if ( coach->version >= 7.0 && !ref ) {
			sprintf(tmp, "(hear %d %s \"%s\")\n",
				time, tmp_name_new, info) ;
		      } else {
			sprintf(tmp, "(hear %d %s %s)\n", time,
				(coach->version >= 7.0) ? tmp_name_new : tmp_name,
				info) ;
		      }
                    }
                coach->udp_send(tmp) ;
        }
        

        for (i = 1; i < 3; i++) {
	  if (olcoach_side == RIGHT || olcoach_side == LEFT)
	    //online coaches don't need to hear themselves and shouldn't hear each other
	    continue; 

	  if (olcoach[i]->assignedp)
	    {
	      if ( olcoach[i]->version >= 7.0 && !ref ) {
		sprintf(tmp, "(hear %d %s \"%s\")\n", time,
			tmp_name_new, info) ;
	      } else {
		sprintf(tmp, "(hear %d %s %s)\n", time,
			(coach->version >= 7.0) ? tmp_name_new : tmp_name,
			info) ;		  
	      }
	      olcoach[i]->udp_send(tmp) ;
	    }
        }
        
}

void Player::send_audio_info(char *message)
{
        udp_send(message,FALSE) ;
}


/*
 *===================================================================
 *  Visual Information Loop
 *===================================================================
 */
void Stadium::udp_send_message() 
{
        int i, sw, tmp ;
        
        /* send to clients */
        for(i = 0 ; i < MAX_PLAYER * 2; i++) {
                sw = irand(MAX_PLAYER * 2) ;
                tmp = dest[i] ;
                dest[i] = dest[sw] ;
                dest[sw] = tmp ;
        }

        for(i = 0 ; i < MAX_PLAYER * 2; i++ ) {
                if (player[dest[i]]->alive == DISABLE)
                        continue ;
                player[dest[i]]->send_visual_info() ;
        }
}

void Player::send_visual_info(void) 
{
        int i, n ;
        static char tmp[MaxStringSize] ;
        ostrstream ost(tmp,MaxStringSize) ;

        sendcnt++ ;

        if (sendcnt >= vis_send)
                sendcnt = 0 ;
        else
                return ;

        ost << "(see " << stadium->time; 

        n = scanField() ;
        for( i = 0 ; i < n ; i++ ) {
                if (vi[i].object->obj_ver <= version)
                        ost << " " << vi[i] ;
        }

        ost << ")" << endl << ends ;

        udp_send(tmp,FALSE) ;
                
}

/* contributed by Artur Merke */
void Player::send_fullstate_information(void) { 
  char tmp[MaxStringSize] ;
  char quality[8] ;
  char width[8] ;
  const float quantize_step = .001f;
  
  strcpy(quality, (highquality) ? "high" : "low") ;

  if (vis_angle == defangle/2)
    strcpy(width, "narrow") ;
  else if (vis_angle == defangle*2)
    strcpy(width, "wide") ;
  else
    strcpy(width, "normal") ;

  ostrstream ostr(tmp,MaxStringSize) ;
  ostr.precision(3);
  ostr << "(fullstate " << stadium->time;

  if ( stadium->mode == PM_FreeKick_Left 
       && stadium->caught_ball == LEFT)
    ostr  << " (pmode goalie_catch_ball_l)" ;
  else if ( stadium->mode == PM_FreeKick_Right 
	    && stadium->caught_ball == RIGHT)
    ostr  << " (pmode goalie_catch_ball_r)" ;
  else
    ostr  << " (pmode " << PlayModeString[stadium->mode] << ")" ;


  ostr     << " (vmode " << quality << " " << width << ")"
       << " (score " << stadium->team_l->point << " " << stadium->team_r->point << ")";
  //print ball information
  ostr << " (ball"
       << " " << Quantize(stadium->ball->pos.x, quantize_step)
       << " " << Quantize(stadium->ball->pos.y, quantize_step)
       << " " << Quantize(stadium->ball->vel.x, quantize_step)
       << " " << Quantize(stadium->ball->vel.y, quantize_step)
       << ")";
  //print players information
  for (int k=0; k< 2*MAX_PLAYER; k++) {
    Player* p = stadium->player[k];
    char side = (p->team->side == LEFT) ? 'l' : 'r';

    if (p->alive == DISABLE)
      continue;

    ostr << " (" << side << "_" << p->unum 
	 << " " << Quantize(p->pos.x, quantize_step) //pos_x
	 << " " << Quantize(p->pos.y, quantize_step) //pos_y
	 << " " << Quantize(p->vel.x, quantize_step) //vel_x
	 << " " << Quantize(p->vel.y, quantize_step) //vel_y
	 << " " << Quantize(Rad2Deg(p->angle_body), quantize_step)
	 << " " << Quantize(Rad2Deg(p->angle_neck_committed), quantize_step) //neck_angle
	 << " " << int(p->stamina) //stamina
	 << " " << Quantize(p->effort, .0001)
	 << " " << Quantize(p->recovery, .0001)
	 << ")";
  }

  ostr << ")";
  ostr << ends;

  udp_send(tmp,FALSE) ;
}
/*
 *===================================================================
 *  Reply sense_body
 *===================================================================
 */
void Player::sense_body(void)
{
        static char tmp[MaxStringSize] ;
        static char quality[8] ;
        static char width[8] ;
        ostrstream ost(tmp,MaxStringSize) ;

        strcpy(quality, (highquality) ? "high" : "low") ;
        
        if (vis_angle == defangle/2)
                strcpy(width, "narrow") ;
        else if (vis_angle == defangle*2)
                strcpy(width, "wide") ;
        else
                strcpy(width, "normal") ;
        /* th 15.01.01 */
        if ( version >= 7.0 )
            {
            ost << "(sense_body " << stadium->time
		<< " (view_mode " << quality << " " << width
		<< ") (stamina " << stamina << " " << effort
		<< ") (speed " << Quantize(vel.r(), 0.01) << " " 
		<< Rad2IDegRound(normalize_angle(vel.th() - angle_body - angle_neck_committed)) 
		<< ") (head_angle " << Rad2IDegRound(angle_neck_committed) 
		<< ") (kick " << count_kick
		<< ") (dash " << count_dash
		<< ") (turn " << count_turn
		<< ") (say " << count_say
		<< ") (turn_neck " << count_turn_neck
		<< ") (catch " << count_catch
		<< ") (move " << count_move
		<< ") (change_view " << count_change_view
		<< "))"
		<< endl << ends ;
            }
        /* th 19.05.00 */
        else
            if( version >= 6.0 ){
           ost << "(sense_body " << stadium->time
               << " (view_mode " << quality << " " << width
               << ") (stamina " << stamina << " " << effort
               << ") (speed " << Quantize(vel.r(), 0.01) << " " 
               << Rad2IDeg(normalize_angle(vel.th() - angle_body - angle_neck_committed))
               << ") (head_angle " <<   Rad2IDeg(angle_neck_committed)
               << ") (kick " << count_kick
               << ") (dash " << count_dash
               << ") (turn " << count_turn
               << ") (say " << count_say
               << ") (turn_neck " << count_turn_neck
               << "))"
               << endl << ends ;
        }
        else if( version >= 5.0 ){
                    ost << "(sense_body " << stadium->time
                        << " (view_mode " << quality << " " << width
                        << ") (stamina " << stamina << " " << effort
                        << ") (speed " << Quantize(vel.r(), 0.01)
                        << ") (head_angle " <<  Rad2IDeg(angle_neck_committed)
                        << ") (kick " << count_kick
                        << ") (dash " << count_dash
                        << ") (turn " << count_turn
                        << ") (say " << count_say
                        << ") (turn_neck " << count_turn_neck
                        << "))"
                        << endl << ends ;                           
        }else{
                ost << "(sense_body " << stadium->time
                        << " (view_mode " << quality << " " << width
                        << ") (stamina " << stamina << " " << effort
                        << ") (speed " << Quantize(vel.r(), 0.01)
                        << ") (kick " << count_kick
                        << ") (dash " << count_dash
                        << ") (turn " << count_turn
                        << ") (say " << count_say
                        << "))"
                        << endl << ends ;
        }
        udp_send(tmp,FALSE) ;
}

void Player::send_score(void)
{
        static char tmp[MaxStringSize] ;
        static char quality[8] ;
        static char width[8] ;
        ostrstream ost(tmp,MaxStringSize) ;

	ost << "(score " << stadium->time
	    << " " << team->point;
	if ( team == stadium->team_l )
	  {
	    if ( stadium->team_r == NULL )
	      ost << " " << 0;
	    else
	      ost << " " << stadium->team_r->point;
	  }
	else
	  {
	    if ( stadium->team_l == NULL )
	      ost << " " << 0;
	    else
	      ost << " " << stadium->team_l->point;	    
	  }
	ost << ")" << endl << ends;
	udp_send ( tmp, FALSE );
}

/*
 *===================================================================
 *  UDP Send
 *===================================================================
 */
void Stadium::udp_send(const char *msg, Player &player,const Logical logp)
{
        if (player.client.port > 0) {
                udp_send(msg,player.client.host,player.client.port,
                        ((((player.team)->side) == LEFT) ? 0 : 1) * MAX_PLAYER + player.unum) ;
                if(logp)
                        write_log(*this,player,msg,SEND) ;
        }
}

void Stadium::udp_send(const char *msg, unsigned long host, int port,int fd)
{
        struct sockaddr_in cli_addr ;
        int n ;

        //bzero((char *) &cli_addr, sizeof(cli_addr)) ;
        memset((char *) &cli_addr, 0, sizeof(cli_addr)) ;
        cli_addr.sin_family                     = AF_INET ;
        cli_addr.sin_addr.s_addr        = htonl(host) ;
        cli_addr.sin_port                       = htons(port) ;

        n = strlen(msg) ;
        if (socketfd[fd] != -1)
                if(sendto(socketfd[fd], (char *)msg, n, 0,
                                        (struct sockaddr *)&cli_addr, sizeof(cli_addr)) != n){
#ifdef DEBUG
                        if (server_param->verbose)
                                cerr << "sendto error on socket." << endl ;
#endif
                }
}

void Stadium::udp_send(const char *msg, unsigned long host, int port,int fd, int online)
{
        struct sockaddr_in cli_addr ;
        int n ;

        //bzero((char *) &cli_addr, sizeof(cli_addr)) ;
        memset((char *) &cli_addr, 0, sizeof(cli_addr)) ;
        cli_addr.sin_family             = AF_INET ;
        cli_addr.sin_addr.s_addr        = htonl(host) ;
        cli_addr.sin_port               = htons(port) ;

        n = strlen(msg) ;
        if (olcoachfd[fd] != -1)
                if(sendto(olcoachfd[fd], (char *)msg, n, 0,
                                        (struct sockaddr *)&cli_addr, sizeof(cli_addr)) != n){
#ifdef DEBUG
                        if (server_param->verbose)
                                cerr << "sendto error on socket." << endl ;
#endif
                }
}

void Stadium::udp_send(dispinfo_t *msg, unsigned long host, int port,int fd)
{
        struct sockaddr_in cli_addr ;
        int n ;

        //bzero((char *) &cli_addr, sizeof(cli_addr)) ;
        memset((char *) &cli_addr, 0, sizeof(cli_addr)) ;
        cli_addr.sin_family                     = AF_INET ;
        cli_addr.sin_addr.s_addr        = htonl(host) ;
        cli_addr.sin_port                       = htons(port) ;

        n = sizeof(dispinfo_t) ;
        if (socketfd[fd] != -1)
                if(sendto(socketfd[fd], (char *)msg, n, 0,
                                        (struct sockaddr *) &cli_addr, sizeof(cli_addr)) != n){
#ifdef DEBUG
                        if (server_param->verbose)
                                cerr << "sendto error on socket." << endl ;
#endif
                }
}

void Stadium::udp_send(dispinfo_t2 *msg, unsigned long host, int port,int fd)
{
        struct sockaddr_in cli_addr ;
        int n ;

        //bzero((char *) &cli_addr, sizeof(cli_addr)) ;
        memset((char *) &cli_addr, 0, sizeof(cli_addr)) ;
        cli_addr.sin_family                     = AF_INET ;
        cli_addr.sin_addr.s_addr        = htonl(host) ;
        cli_addr.sin_port                       = htons(port) ;

        n = sizeof(dispinfo_t2) ;
        if (socketfd[fd] != -1)
                if(sendto(socketfd[fd], (char *)msg, n, 0,
                                        (struct sockaddr *) &cli_addr, sizeof(cli_addr)) != n){
#ifdef DEBUG
                        if (server_param->verbose)
                                cerr << "sendto error on socket." << endl ;
#endif
                }
}

PlayMode PlayModeID(char *mode)
{
        int n ;
        for(n = 0 ; n < PM_MAX ; n++) {
                if(!strcmp(PlayModeString[n],mode))
                        return (PlayMode)n ;
        }
        return PM_Null ;
}

void Stadium::send_init_info(Player& player) 
{
        char tmp[MaxStringSize] ;
        ostrstream ost(tmp,MaxStringSize) ;

        ost << "(init " << SideStr(player.team->side) << " " 
                << player.unum << " "
                << PlayModeString[mode]
                << ")" << endl << ends ;
        udp_send(tmp,player) ;

        send_parameter_info ( &player );
        
        send_changed_players ( &player );
}

void Stadium::send_reconnect_info(Player& player) 
{
        char tmp[MaxStringSize] ;
        ostrstream ost(tmp,MaxStringSize) ;

        
        ost << "(reconnect " << SideStr(player.team->side) << " " 
            /* start - TH 21-JUNE-2000 */
            << PlayModeString[mode];
        
//      if( player.version >= 6.0 )
//        {
//          if (player.team == this->team_l)
//            ost << " " << team_l->point << " " << team_r->point;
//          else
//            ost << " " << team_r->point << " " << team_l->point;
//        }
        
        ost << ")" << endl << ends ;
        /* end - TH 21-JUNE-2000 */
        udp_send(tmp,player) ;

        send_parameter_info (&player);
        send_changed_players ( &player );
}

void Player::udp_send(const char *message,const Logical logp)
{
        stadium->udp_send(message,*this,logp) ;
}

void Stadium::send_parameter_info ( Player* player )
{
    if ( player == NULL || player->version < 7.0 )
        return;
    
    char tmp[MaxStringSize] ;

    // send server parameters
    server_param->getSettingsString ( tmp, player->version );
    player->udp_send ( tmp );

    
    // send player parameters
    player_param->getSettingsString ( tmp );
    player->udp_send ( tmp );

    // send player types
    for ( int i = 0; i < player_param->playerTypes (); i++ )
        {
        if ( player_types [ i ] != NULL )
            {
            player_types [ i ]->getSettingsString ( i, tmp );
            player->udp_send ( tmp );
            }
        }
}

void Stadium::send_parameter_info ( Coach* coach )
{
    if ( coach == NULL || coach->version < 7.0 )
        return;
    
    char tmp[MaxStringSize] ;

    // send server parameters
    server_param->getSettingsString ( tmp );
    coach->udp_send ( tmp );

    // send player parameters
    player_param->getSettingsString ( tmp );
    coach->udp_send ( tmp );
    
    // send player types
    for ( int i = 0; i < player_param->playerTypes (); i++ )
        {
        if ( player_types [ i ] != NULL )
            {
            player_types [ i ]->getSettingsString ( i, tmp );
            coach->udp_send ( tmp );
            }
        }
}

void Stadium::send_changed_players ( Coach* coach )
{
    if ( coach == NULL || coach->version < 7.0 )
        return;

    char tmp[MaxStringSize] ;
    if ( team_l != NULL )
        for ( int i = 0; i < team_l->n; i++ )
            {
            if ( team_l->player[i]->player_type_id != 0 )
                {
                if ( coach->side == LEFT )
                    sprintf ( tmp, "(change_player_type %d %d)\n", team_l->player[i]->unum, team_l->player[i]->player_type_id );
                else if ( coach->side == RIGHT )
                    sprintf ( tmp, "(change_player_type %d)\n", team_l->player[i]->unum );
                else
                    sprintf ( tmp, "(change_player_type %s %d %d)\n", team_l->name, team_l->player[i]->unum, team_l->player[i]->player_type_id );

                coach->udp_send ( tmp );
                }
            }

    if ( team_r != NULL )
        for ( int i = 0; i < team_r->n; i++ )
            {
            if ( team_r->player[i]->player_type_id != 0 )
                {
                if ( coach->side == RIGHT )
                    sprintf ( tmp, "(change_player_type %d %d)\n", team_r->player[i]->unum, team_r->player[i]->player_type_id );
                else if ( coach->side == LEFT )
                    sprintf ( tmp, "(change_player_type %d)\n", team_r->player[i]->unum );
                else
                    sprintf ( tmp, "(change_player_type %s %d %d)\n", team_r->name, team_r->player[i]->unum, team_r->player[i]->player_type_id );

                coach->udp_send ( tmp );
                }
            }
}

void Stadium::send_changed_players ( Player* player )
{
    if ( player == NULL || player->version < 7.0 )
        return;

    char tmp[MaxStringSize] ;
    if ( team_l != NULL )
        for ( int i = 0; i < team_l->n; i++ )
            {
            if ( team_l->player[i]->player_type_id != 0 )
                {
                if ( player->team == team_l )
                    sprintf ( tmp, "(change_player_type %d %d)\n", team_l->player[i]->unum, team_l->player[i]->player_type_id );
                else if ( player->team == team_r )
                    sprintf ( tmp, "(change_player_type %d)\n", team_l->player[i]->unum );
                else
                    return;

                player->udp_send ( tmp );
                }
            }

    if ( team_r != NULL )
        for ( int i = 0; i < team_r->n; i++ )
            {
            if ( team_r->player[i]->player_type_id != 0 )
                {
                if ( player->team == team_r )
                    sprintf ( tmp, "(change_player_type %d %d)\n", team_r->player[i]->unum, team_r->player[i]->player_type_id );
                else if ( player->team == team_l )
                    sprintf ( tmp, "(change_player_type %d)\n", team_r->player[i]->unum );
                else
                    return;

                player->udp_send ( tmp );
                }
            }

}

/*
 *===================================================================
 *  Coach
 *===================================================================
 */
int Coach::parse_init ( Stadium& stad, char *command, 
			 unsigned long host, int port)
{
    client.host = host ;
        client.port = port ;

        char com[MaxStringSize] ;
        int n ;
        n = sscanf(command, "(%[-0-9a-zA-Z.+*/?<>_]",com) ;
        if(n<1) {udp_send("(error illegal_command_form)", TRUE); return 0; }

	if (!strcmp ( com, "init" ) )
	    {
	    float ver = 0.0;
	    n = sscanf ( command, "(init (version %f))", &ver );
	    if ( n < 1 )
		{
		udp_send("(error illegal_command_form)");
		return false;
		}
	    version = ver;
	    udp_send("(init ok)") ;
	    stad.send_parameter_info ( this );
	    stad.send_changed_players ( this );
	    return 1;
	    }
	
	return -1;
}

void Coach::parse_command(Stadium& stad, char *command, 
			  unsigned long host, int port)
{
        client.host = host ;
        client.port = port ;

        char com[MaxStringSize] ;
        int n ;

        n = sscanf(command, "(%[-0-9a-zA-Z.+*/?<>_]",com) ;
        if(n<1) {udp_send("(error illegal_command_form)", TRUE); return ; }

	if ( !strcmp ( com, "start" ) )
	  {
	    _Start ( stad );
	    udp_send("(ok start)");
	  }
	
        else if(!strcmp(com,"change_mode")) {
                char new_mode[MaxStringSize] ;
                PlayMode mode_id ;
                n = sscanf(command,"(change_mode %[-0-9a-zA-Z.+*/?<>_])", new_mode) ;
                if(n<1) {udp_send("(error illegal_command_form)"); return ;}
                mode_id = PlayModeID(new_mode) ;

                if(mode_id == PM_Null) { udp_send("(error illegal_mode)") ; return ; }
                stad.change_play_mode(mode_id) ; 
                udp_send("(ok change_mode)") ;
        }
        else if (!strcmp(com,"move")) {
                char obj[MaxStringSize] ;
                double x,y,ang, velx,vely ;

                command++ ;
                while(*command != '(') {
                        if (*command == NULLCHAR) {
                                udp_send("(error illegal_object_form)") ;
                                return ;
                        }
                        command++ ;
                }
                int i = 0 ;
                while(*command != ')') {
                        if (*command == NULLCHAR) {
                                udp_send("(error illegal_object_form)") ;
                                return ;
                        }
                        *(obj+i) = *command ;
                        i++ ; command++ ;
                }
                *(obj+i) = *command ;
                i++ ; command++ ;
                *(obj+i) = NULLCHAR ;

                if (!strcmp(obj, BALL_NAME))
                        stad.caught_ball = FALSE;

                n = sscanf(command," %lf %lf %lf %lf %lf)",&x,&y,&ang,&velx,&vely) ;
                if (n==2) {
                        if (!strcmp(obj, BALL_NAME))
                                stad.set_ball(PVector(x,y), LEFT) ;
                        else
                                stad.move_object(obj,PVector(x,y)) ;
                        //stad.check_collision(obj) ;
			stad.collisions ();
                }
                else if (n==3) {
                        if (!strcmp(obj, BALL_NAME))
                                stad.set_ball(PVector(x,y), LEFT, ang) ;
                        else
                                stad.move_object(obj,PVector(x,y),
                                        Deg2Rad(max(min((ang),stad.server_param->maxm),stad.server_param->minm))) ;
                        //stad.check_collision(obj) ;
			stad.collisions ();
                }
                else if (n==5) {
                        if (!strcmp(obj, BALL_NAME))
                                stad.set_ball(PVector(x,y), LEFT, ang, PVector(velx, vely)) ;
                        else
                                stad.move_object(obj,PVector(x,y),Deg2Rad(max(min((ang),stad.server_param->maxm),stad.server_param->minm)),
                                        PVector(velx, vely));
                        //stad.check_collision(obj) ;
			stad.collisions ();
                }
                else {
                        udp_send("(error illegal_command_form)");
                        return ;
                }
                udp_send("(ok move)") ;
        }
        else if (!strcmp(com,"look")) {
                look(stad) ;
        }
        else if (!strcmp(com,"team_names")) {
                team_names(stad) ;
        }
        else if (!strcmp(com,"recover")) {
                recover(stad);
        }
        else if (!strcmp(com,"check_ball")) {
                check_ball(stad) ;
        }
        else if (!strcmp(com,"say")) {
                static char msg[MaxStringSize] ;
                n = sscanf(command,SAY_MESSAGE_SCAN_FORMAT,com,msg) ;
                if(n<2) {udp_send("(error illegal_command_form)"); return ;}
                chop_last_parenthesis(msg, stad.server_param->SayMsgSize) ;
                stad.say(msg, false) ;
                udp_send("(ok say)") ;
        }
        else if (!strcmp(com,"ear")) {
                char mode[16] ;
                n = sscanf(command,"(%s %s)", com, mode) ;
                if(n<2) {udp_send("(error illegal_command_form)"); return ;}
                if (!strcmp(mode,"on)")) {
                        hear = TRUE ;
                        udp_send("(ok ear on)") ;
                }
                else if (!strcmp(mode, "off)")) {
                        hear = FALSE ;
                        udp_send("(ok ear off)") ;
                }
        }
        else if (!strcmp(com,"eye")) {
                char mode[16] ;
                n = sscanf(command,"(%s %s)", com, mode) ;
                if(n<2) {udp_send("(error illegal_command_form)"); return ;}
                if (!strcmp(mode,"on)")) {
                        eye = TRUE ;
                        udp_send("(ok eye on)") ;
                }
                else if (!strcmp(mode, "off)")) {
                        eye = FALSE ;
                        udp_send("(ok eye off)") ;
                }
        }
        else if ( !strcmp ( com, "change_player_type" ) )
            {
            char name[128];
            int unum, player_type;
            if ( sscanf ( command, "(%s %s %d %d)", com, name, &unum, &player_type ) < 4 )
                {
                udp_send("(error illegal_command_form)");
                return ;
                }

            Team* team = NULL;
            if ( stad.team_l != NULL && stad.team_l->name != NULL && 
                 !strcmp ( name, stad.team_l->name ) ) {
              team = stad.team_l;
            }
            if ( stad.team_r != NULL && stad.team_r->name != NULL && 
                 !strcmp ( name, stad.team_r->name ) ) {
              team = stad.team_r;
            }

            if ( team == NULL )
                {
                udp_send("(warning no_team_found)");
                return ;
                }

            if ( player_type < 0 || player_type >= stad.player_param->playerTypes() )
              {
                udp_send("(error out_of_range_player_type)");
                return;
              }
            
            Player* player = NULL; 
            for ( int i = 0; i < team->n; i++ )
                {
                if ( team->player[ i ] != NULL && team->player[ i ]->unum == unum )
                    {
                    player = team->player[ i ];
                    break;
                    }
                }

            if ( player == NULL )
                {
                udp_send("(warning no_such_player)");
                return ;
                }

            player->substitute ( player_type );

            if ( team == stad.team_l )
                stad.broadcastSubstitution ( LEFT, unum , player_type, true );
            else
                stad.broadcastSubstitution ( RIGHT, unum , player_type, true );

            char buf[64];
            sprintf ( buf, "(ok change_player_type %s %d %d)", name, unum, player_type );
            
            udp_send( buf );
            
            return ;
            }   
	//pfr:SYNCH
	else if (!strcmp(com, "done")) {
	  //cerr << "Recv trainer done" << endl;
	    done_received = TRUE;
	    return ;
	} 
        else {
                udp_send("(error unknown_command)") ;
                return ;
        }
}

void Coach::look(Stadium& stad) 
{
    static char tmp[MaxStringSize] ;
    ostrstream ost(tmp,MaxStringSize) ;

    ost << "(ok look " << stad.time ;
    if ( version >= 7.0 )
        ost << " (" << GOAL_L_NAME_SHORT;
    else
        ost << " (" << GOAL_L_NAME;
            
    ost << " " << -(PITCH_LENGTH/2.0)
        << " " << 0.0
        << ")" ;

    if ( version >= 7.0 )
        ost << " (" << GOAL_R_NAME_SHORT; 
    else
        ost << " (" << GOAL_R_NAME;
        
    ost << " " << (PITCH_LENGTH/2.0)
        << " " << 0.0
        << ")" ;
    if ( version >= 7.0 )
        ost << " (" << BALL_NAME_SHORT;
    else
        ost << " (" << BALL_NAME;
        
    ost << " " << stad.ball->pos.x
        << " " << stad.ball->pos.y
        << " " << stad.ball->vel.x
        << " " << stad.ball->vel.y
        << ")" ;

    TheNumber i ;
    for(i = 0 ; i < MAX_PLAYER * 2; i++ ) {
    if (stad.player[i]->alive == DISABLE)
        continue ;
    if ( version >= 7.0 )               
        ost << " (" << (stad.player[i])->short_name
	    << " " << (stad.player[i])->pos.x
	    << " " << (stad.player[i])->pos.y
	    << " " << (stad.player[i])->vel.x
	    << " " << (stad.player[i])->vel.y
	    << " " << Rad2IDegRound((stad.player[i])->angle)
	    << " " << Rad2IDegRound((stad.player[i])->angle_neck_committed)
	    << ")" ;    
    else
        ost << " (" << (stad.player[i])->name
	    << " " << (stad.player[i])->pos.x
	    << " " << (stad.player[i])->pos.y
	    << " " << Rad2IDeg((stad.player[i])->angle)
	    << " " << Rad2IDeg((stad.player[i])->angle_neck_committed)
	    << " " << (stad.player[i])->vel.x
	    << " " << (stad.player[i])->vel.y
	    << ")" ;
    }
    ost << ")" << endl << ends;

    udp_send(tmp) ;
}

void Coach::team_names(Stadium& stad)
{
        static char tmp[MaxStringSize] ;
        ostrstream ost(tmp,MaxStringSize) ;

        ost << "(ok team_names";
        
        if (stad.team_l->name) {
                ost << " (team l " << stad.team_l->name << ")";
        }
        if (stad.team_r->name) {
                ost << " (team r " << stad.team_r->name << ")";
        }

        ost << ")" << endl << ends;
        udp_send(tmp) ;
}

void Coach::recover(Stadium& stad)
{
        static char tmp[MaxStringSize] ;
        ostrstream ost(tmp,MaxStringSize) ;

        stad.recovery_players();

        ost << "(ok recover)" << endl << ends;
        udp_send(tmp);
}



void Coach::send_visual_info(void) 
{
    int i;
    static char tmp[MaxStringSize] ;
    ostrstream ost(tmp,MaxStringSize) ;

    if ( version >= 7.0 )    
      ost << "(see_global " << stadium->time; 
    else
      ost << "(see " << stadium->time; 
      
    if ( version >= 7.0 )
        ost << " (" << GOAL_L_NAME_SHORT;
    else
        ost << " (" << GOAL_L_NAME;
    ost << " " << -(PITCH_LENGTH/2.0)
        << " " << 0.0
        << ")" ;
    
    if ( version >= 7.0 )
        ost << " (" << GOAL_R_NAME_SHORT;
    else
        ost << " (" << GOAL_R_NAME;
    ost << " " << (PITCH_LENGTH/2.0)
        << " " << 0.0
        << ")" ;
    
    if ( version >= 7.0 )
        ost << " (" << BALL_NAME_SHORT;
    else
        ost << " (" << BALL_NAME;    
    ost << " " << stadium->ball->pos.x
        << " " << stadium->ball->pos.y
        << " " << stadium->ball->vel.x
        << " " << stadium->ball->vel.y
        << ")" ;

    for(i = 0 ; i < MAX_PLAYER * 2; i++ ) {
    if (stadium->player[i]->alive == DISABLE)
        continue ;
    
    if ( version >= 7.0 )
        ost << " (" << (stadium->player[i])->short_name
	    << " " << (stadium->player[i])->pos.x
	    << " " << (stadium->player[i])->pos.y
	    << " " << (stadium->player[i])->vel.x
	    << " " << (stadium->player[i])->vel.y
	    << " " << Rad2IDegRound((stadium->player[i])->angle)
	    << " " << Rad2IDegRound((stadium->player[i])->angle_neck_committed)
	    << ")" ;
    else
        ost << " (" << (stadium->player[i])->name
	    << " " << (stadium->player[i])->pos.x
	    << " " << (stadium->player[i])->pos.y
	    << " " << Rad2IDeg((stadium->player[i])->angle)
	    << " " << Rad2IDeg((stadium->player[i])->angle_neck_committed)
	    << " " << (stadium->player[i])->vel.x
	    << " " << (stadium->player[i])->vel.y
	    << ")" ;
    }
    ost << ")" << endl << ends;

    udp_send(tmp) ;
}

void Coach::check_ball(Stadium& stad)
{
        static char tmp[MaxStringSize] ;
        ostrstream ost(tmp,MaxStringSize) ;

        static char* BallPosInfoStr[] = BALL_POS_INFO_STRINGS ;
        BallPosInfo info = stad.check_ball_pos() ;

        ost << "(ok check_ball " << stad.time << " " ;

        ost << BallPosInfoStr[info] << ")" << ends ;

        udp_send(tmp) ;
}

void Coach::udp_send(const char *msg, Logical logp)
{
        struct sockaddr_in cli_addr ;
        int n ;

        //bzero((char *) &cli_addr, sizeof(cli_addr)) ;
        memset((char *) &cli_addr, 0, sizeof(cli_addr)) ;
        cli_addr.sin_family             = AF_INET ;
        cli_addr.sin_addr.s_addr        = htonl(client.host) ;
        cli_addr.sin_port               = htons(client.port) ;

        n = strlen(msg) ;
        if(sendto(socketfd, (char *)msg, n, 0,
                        (struct sockaddr *) &cli_addr, sizeof(cli_addr)) != n){
#ifdef DEBUG
                if (server_param->verbose)
                        cerr << "sendto error on socket." << endl ;
#endif
        }
        else if (logp)
            write_log(*stadium,*this,msg,SEND) ;
}

void OnlineCoach::parse_command(Stadium& stad, char *command, 
                                unsigned long host, int port)
{
  char com[MaxStringSize] ;
  int n;

  n = sscanf(command, "(%[-0-9a-zA-Z.+*/?<>_]",com) ;
  if(n<1) {udp_send("(error illegal_command_form)"); return ; }


  if (!strcmp(com,"check_ball")) {
    check_ball(stad) ;
  }
  else if (!strcmp(com,"look")) {
    look(stad) ;
  }
  else if (!strcmp(com,"team_names")) {
    team_names(stad) ;
  }
  else if (!strcmp(com,"say")) {
    if (version >= 7.0) {
      /* parsing for new coach language */
      //command still has it's trialing parentheses, but it shouldn't matter
      chop_last_parenthesis(command, MaxMesg) ;
      int ret = parse_coach_lang_string(command + 4); //skip '(say'
      if (ret == 0 && g_pMess != NULL) {
        //succeful parse
        bool should_queue = false;

        g_GC.remove(g_pMess, GCE_pMess);
        g_pMess->setTimeRecv(stadium->time);

        switch (g_pMess->getType()) {
        case CMT_Freeform:
          if(stad.mode != PM_PlayOn) {
            if (freeform_messages_said < freeform_messages_allowed
                || freeform_messages_allowed < 0) {
              should_queue = true;
	      freeform_messages_said++;
            } else {
              udp_send("(error said_too_many_freeform_messages)");
            }
          } else {
            udp_send("(error cannot_say_freeform_while_playon)");
          }
          break;
        case CMT_Meta:
          if(stad.mode != PM_PlayOn) {
            should_queue = true;
          } else if (meta_messages_left != 0) {
            //remember, negative implies we can send as many as we want
            --meta_messages_left;
            should_queue = true;
          } else {
            udp_send("(error said_too_many_meta_messages)");
          }
          break;
        case CMT_Advice:
          if(stad.mode != PM_PlayOn) {
            should_queue = true;
          } else if (advice_messages_left != 0) {
            //remember, negative implies we can send as many as we want
            --advice_messages_left;
            should_queue = true;
          } else {
            udp_send("(error said_too_many_advice_messages)");
          }
          break;
        case CMT_Info:
          if(stad.mode != PM_PlayOn) {
            should_queue = true;
          } else if (info_messages_left != 0) {
            //remember, negative implies we can send as many as we want
            --info_messages_left;
            should_queue = true;
          } else {
            udp_send("(error said_too_many_info_messages)");
          }
          break;
        case CMT_Define:
          if(stad.mode != PM_PlayOn) {
            should_queue = true;
          } else if (define_messages_left != 0) {
            //remember, negative implies we can send as many as we want
            --define_messages_left;
            should_queue = true;
          } else {
            udp_send("(error said_too_many_define_messages)");
          }
          break;
        default:
          cerr << "g_pMess->getType has strange value: " << g_pMess->getType() << endl;
          break;
        }

        if (should_queue) {
          message_queue.push_back(g_pMess);
          udp_send("(ok say)") ;
        } else {
          delete g_pMess;
        }
      } else {
	udp_send("(error could_not_parse_say)");
      }

      g_GC.deleteAll(); //reclaims any memory

      g_pMess = NULL;
    } else {
      //pre v7.0
      if(stad.mode != PM_PlayOn) {
        if (freeform_messages_said < freeform_messages_allowed || freeform_messages_allowed < 0) {
          static char msg[MaxStringSize] ;
          n = sscanf(command,SAY_MESSAGE_SCAN_FORMAT,com,msg) ;
          if(n<2) {
            udp_send("(error illegal_command_form)");
            return ;
          }
          chop_last_parenthesis(msg, stad.server_param->SayCoachMsgSize) ;
          say(msg) ;
          udp_send("(ok say)") ;
        } else {
          udp_send("(error said_too_many_messages)");
          return;
        }
      }
      else {
        udp_send("(warning cannot_say_while_playon)");
        return ;
      }
    }
          
  }
  else if (!strcmp(com, "bye")) {
    assignedp = FALSE;
    client.host = 0 ;
    client.port = -1 ;
    socketfd = -1 ;
    return ;
  } 
  else if (!strcmp(com,"eye")) {
    char mode[16] ;
    n = sscanf(command,"(%s %s)", com, mode) ;
    if(n<2) {udp_send("(error illegal_command_form)"); return ;}
    if (!strcmp(mode,"on)")) {
      eye = TRUE ;
      udp_send("(ok eye on)") ;
    }
    else if (!strcmp(mode, "off)")) {
      eye = FALSE ;
      udp_send("(ok eye off)") ;
    }
  }
  else if ( !strcmp ( com, "change_player_type" ) ) {
    if ( stad.mode == PM_PlayOn)
      {
        udp_send("(warning cannot_sub_while_playon)");
        return ;
      }

    Team* team = NULL;
    if ( side == LEFT )
      team = stad.team_l;
    if ( side == RIGHT )
      team = stad.team_r;

    if ( team == NULL )
      {
        udp_send("(warning no_team_found)");
        return ;
      }
                
    if ( team->subs == stad.player_param->subsMax () )
      {
        udp_send("(warning no_subs_left)");
        return ;
      }
                
                
    int unum, player_type;
    if ( sscanf ( command, "(%s %d %d)", com, &unum, &player_type ) < 3 )
      {
        udp_send("(error illegal_command_form)");
        return ;
      }

    if ( player_type < 0 || player_type >= stad.player_param->playerTypes() )
      {
        udp_send("(error out_of_range_player_type)");
        return;
      }
    
    
    if ( player_type != 0 && team->ptype_count [ player_type ] == stad.player_param->ptMax () )
      {
        udp_send("(warning max_of_that_type_on_field)");
        return ;
      }

        
    Player* player = NULL; 
    for ( int i = 0; i < team->n; i++ )
      {
        if ( team->player[ i ]->unum == unum )
          {
            player = team->player[ i ];
            break;
          }
      }

    if ( player == NULL )
      {
        udp_send("(warning no_such_player)");
        return ;
      }

    if ( player->goalie && player_type != 0 )
      {
        udp_send("(warning cannot_change_goalie)");
        return ;
      }
            
    team->ptype_count [ player->player_type_id ] = team->ptype_count [ player->player_type_id ] - 1;

    player->substitute ( player_type );

    if ( stad.mode != PM_BeforeKickOff )
      {
        team->subs++;
      }
            
    team->ptype_count [ player_type ] = team->ptype_count [ player_type ] + 1;
            
    stad.broadcastSubstitution ( side, unum , player_type );

    char buf[64];
    sprintf ( buf, "(ok change_player_type %d %d)\n", unum, player_type );
            
    udp_send( buf );
            
    return ;
    //pfr:SYNCH
  } else if (!strcmp(com, "done")) {
      //cerr << "Recv olc done" << endl;
      done_received = TRUE;
      return ;
  } else {
    udp_send("(error unknown_command)") ;
    return ;
  }
}

void Stadium::broadcastSubstitution ( const int& side, const int& unum, const int& player_type, bool tell_ally_coach )
{
  char allies[64];
  char enemies[64];
  sprintf ( allies, "(change_player_type %d %d)\n", unum, player_type );
  sprintf ( enemies, "(change_player_type %d)\n", unum );

  // tell players
  for (int i = 0 ; i < MAX_PLAYER * 2; i++)
    {
      if (player[i]->alive == DISABLE || player[i]->version < 7.0)
        continue ;
      if ( ( side == LEFT && player [ i ]->team == team_l ) || ( side == RIGHT && player [ i ]->team == team_r )  )
        player[i]->udp_send ( allies );
      else
        player[i]->udp_send ( enemies );        
    }

  // tell coaches
  if ( team_r != NULL && team_r->olcoach != NULL && 
       team_r->olcoach->version >= 7.0 ) {
    if( side == RIGHT && tell_ally_coach ) {
      team_r->olcoach->udp_send ( allies );
    }
    if( side == LEFT ) {
      team_r->olcoach->udp_send ( enemies );
    }
  }
  if ( team_l != NULL && team_l->olcoach != NULL &&
       team_l->olcoach->version >= 7.0 ) {
    if( side == LEFT && tell_ally_coach ) {
      team_l->olcoach->udp_send ( allies );
    }
    if( side == RIGHT ) {
      team_l->olcoach->udp_send ( enemies );
    }
  }
}

/*
 *===================================================================
 *  Misc
 *===================================================================
 */
void write_log(Stadium& stad, Player& p, const char *message, int flag) 
{
        displist_t *dp ;
        static char tmp[MaxStringSize] ;

        if (stad.server_param->rec_log || stad.server_param->send_log) {
                stad.minfo.body.msg.board = htons(LOG_BOARD) ;

                if (&p == NULL)
                        sprintf(tmp, "%s", "Illegal\n") ;
                else
                        sprintf(tmp, "%s %s_%d: %s\n", (flag == SEND) ? "Send" : "Recv", 
                                                        p.team->name, p.unum, message) ;

                strncpy(stad.minfo.body.msg.message, tmp, 128) ;

                if (stad.server_param->logfile[0] != NULLCHAR) {
                        static char tmp2[MaxStringSize] ;

                        sprintf(tmp2, "%d\t%s", stad.time, tmp ) ;
                        fputs(tmp2, stad.log_fp) ;
                        fflush(stad.log_fp) ;
                }

                if (stad.server_param->send_log)
                        for (dp = stad.top.next ; dp != NULL ; dp = dp->next)
                            {
                            if ( dp->version == 1 )
                                stad.udp_send(&(stad.minfo),dp->shost,dp->sport,0) ;
                            else
                                {
                                dispinfo_t2 minfo2;
                                minfo2.mode = stad.minfo.mode;
                                minfo2.body.msg = stad.minfo.body.msg;
                                stad.udp_send(&minfo2,dp->shost,dp->sport,0) ;
                                }
                            }

                if (stad.server_param->recfile[0] != NULLCHAR && stad.server_param->rec_log == TRUE)
                        if (stad.mode != PM_TimeOver)
                                stad.WriteLog(&stad.minfo) ;
        }
}

void write_log(Stadium& stad, Coach& p, const char *message, int flag)
{
        displist_t *dp ;
        static char tmp[MaxStringSize] ;

        if (stad.server_param->rec_log || stad.server_param->send_log) {
                stad.minfo.body.msg.board = htons(LOG_BOARD) ;

                if (&p == NULL)
                        sprintf(tmp, "%s", "Illegal\n") ;
                else
                        sprintf(tmp, "%s Coach: %s\n", (flag == SEND) ? "Send" : "Recv", message) ;

                strncpy(stad.minfo.body.msg.message, tmp, max_message_length_for_display) ;

                if (stad.server_param->logfile[0] != NULLCHAR) {
                        static char tmp2[MaxStringSize] ;

                        sprintf(tmp2, "%d\t%s", stad.time, tmp ) ;
                        fputs(tmp2, stad.log_fp) ;
                        fflush(stad.log_fp) ;
                }

                if (stad.server_param->send_log)
                        for (dp = stad.top.next ; dp != NULL ; dp = dp->next)
                            {
                            if ( dp->version == 1 )
                                stad.udp_send(&(stad.minfo),dp->shost,dp->sport,0) ;
                            else
                                {
                                dispinfo_t2 minfo2;
                                minfo2.mode = stad.minfo.mode;
                                minfo2.body.msg = stad.minfo.body.msg;
                                stad.udp_send(&minfo2,dp->shost,dp->sport,0) ;
                                }
                            }

                if (stad.server_param->recfile[0] != NULLCHAR && stad.server_param->rec_log == TRUE)
                        if (stad.mode != PM_TimeOver)
                                stad.WriteLog(&stad.minfo) ;
        }
}

void write_log(Stadium& stad, OnlineCoach& p, const char *message, int flag)
{
        displist_t *dp ;
        static char tmp[MaxStringSize] ;

        if (stad.server_param->rec_log || stad.server_param->send_log) {
                stad.minfo.body.msg.board = htons(LOG_BOARD) ;

                if (&p == NULL || ( p.side == LEFT && stad.team_l == NULL ) || ( p.side == RIGHT && stad.team_r == NULL ) || p.side == NEUTRAL )
                        sprintf(tmp, "%s", "Illegal\n") ;
                else
                        sprintf(tmp, "%s %s_Coach: %s\n", (flag == SEND) ? "Send" : "Recv", 
                                ( p.side == LEFT ) ? stad.team_l->name : stad.team_r->name,
                                message) ;

                strncpy(stad.minfo.body.msg.message, tmp, 128) ;

                if (stad.server_param->logfile[0] != NULLCHAR) {
                        static char tmp2[MaxStringSize] ;

                        sprintf(tmp2, "%d\t%s", stad.time, tmp ) ;
                        fputs(tmp2, stad.log_fp) ;
                        fflush(stad.log_fp) ;
                }

                if (stad.server_param->send_log)
                        for (dp = stad.top.next ; dp != NULL ; dp = dp->next)
                            {
                            if ( dp->version == 1 )
                                stad.udp_send(&(stad.minfo),dp->shost,dp->sport,0) ;
                            else
                                {
                                dispinfo_t2 minfo2;
                                minfo2.mode = stad.minfo.mode;
                                minfo2.body.msg = stad.minfo.body.msg;
                                stad.udp_send(&minfo2,dp->shost,dp->sport,0) ;
                                }
                            }

                if (stad.server_param->recfile[0] != NULLCHAR && stad.server_param->rec_log == TRUE)
                        if (stad.mode != PM_TimeOver)
                                stad.WriteLog(&stad.minfo) ;
        }    
}

Value distance(PVector *pos1, PVector *pos2)
{
        return sqrt(Pow(pos2->x - pos1->x) + Pow(pos2->y - pos1->y)) ;
}
