/*
 *Header:
 *File: object.h (for C++ & cc)
 *Author: Noda Itsuki
 *Date: 1995/02/21
 *EndHeader:
 */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */


/*
 *ModifyHistory
 *
 *EndModifyHistory
 */

#include "hetroplayer.h"

#ifndef _H_OBJECT
#define _H_OBJECT

#ifdef __cplusplus

#define SEPLINE	"--------------------"

inline Value square(Value x) { return x * x ; }

//extern ostream& operator<< (ostream& o, const Angle& ang) ;

/*
 *===================================================================
 *Part: Plain Vector
 *===================================================================
 */
class PVector;
PVector Polar2PVector(Value r, Angle ang);

class PVector {
 public:
	Value	x ;
	Value	y ;

	PVector(const double vx = 0.0, const double vy = 0.0) ;
	void	operator +=(const PVector& v) ;
	void	operator -=(const PVector& v) ;
	void	operator *=(const Value& a) ;
	void	operator /=(const Value& a) ;
	inline Value r() const { return sqrt(square(x)+square(y)) ;} ;

	inline void r( const Value& v ) { *this = Polar2PVector ( v, th () ) ;} 
	inline Value th() const { return Atan(y,x) ; } ;
	void normalize(const Value& l = 1.0) ;
	Value distance(const PVector& orig) ;
	Value angle() ;
	Value angle(const PVector& dir) ;
	void rotate(const Angle& ang) ;
	Value vangle(const PVector& target, const PVector& origin) ;
	Value vangle(const PVector& target, const Angle& origin) ;

} ;

PVector operator +(const PVector& a, const PVector& b) ;
PVector operator -(const PVector& a, const PVector& b) ;

inline bool operator== ( const PVector& a, const PVector& b )
{
  return ( a.x == b.x && a.y == b.y );
}

extern ostream& operator<< (ostream& o, const PVector& v) ;

/*
 *===================================================================
 *Part: Area
 *===================================================================
 */
class RArea {
 public:
	Value left ;
	Value right ;
	Value top ;
	Value bottom ;

	RArea(const Value left, const Value right, 
				const Value top, const Value bottom) ;
	RArea(const PVector center,const PVector size) ;

	Logical inArea(const PVector& p) ;
	PVector nearestHEdge(const PVector& p) ;
	PVector nearestVEdge(const PVector& p) ;
	PVector nearestEdge(const PVector& p) ;
	PVector randomize() ;

} ;

extern ostream& operator<< (ostream& o, const RArea& a) ;

class CArea {
 public:
	PVector center ;
	Value r ;

	CArea(const PVector center,const Value r) ;

	Logical inArea(const PVector& p) ;
	PVector nearestEdge(const PVector& p) ;
} ;


extern ostream& operator<< (ostream& o, const CArea& a) ;

/*
 *===================================================================
 *Part: Visual Info
 *===================================================================
 */
class PObject ;

class VisualInfo {
 public:
	PObject *object ;
	Name name ;
	/* th 19.05.00 */
       	Name short_name;
	Value distance ;
	Angle direction ;
	Value dist_change ;
	Angle dir_change ;
	Angle face_dir ;
	Angle body_dir ;
	int highquality ;
	Value qstep ;
#ifdef	NEW_QSTEP
	Value qstep_dir ;
#endif
	Value version ;

	VisualInfo(double) ;
	VisualInfo();
	void SetVersion(double);
} ;

extern ostream& operator<< (ostream& o, const VisualInfo& vi) ;

/*
 *===================================================================
 *Part: PObject
 *===================================================================
 */
class Stadium ;

class PObject {
 protected:
	static TheNumber Max ;
 public:
	ID		id ;
	Name	name ;
	/* th 19.05.00 */
	Name	short_name;
	PVector	pos ;
	Value	obj_ver ;
	Value	size ;
	Angle	angle ;
	int		enable ;
	Stadium	*stadium ;
	VisualInfo *vi ;

	PObject(const Name name=NULLCHAR, const Name short_name=NULLCHAR,
		const PVector& p=PVector(0.0,0.0), const Value v=3.0) ;
	
	void Set(const PVector& p, const Value& s) ; 
	Angle vangle(const PObject& obj) ;
	void reportScanField(ostream& o) ;
} ;

extern ostream& operator<< (ostream& o, const PObject& v) ;

class MPObjectTable ;
class Weather ;

class MPObject : public PObject {
 public:
	PVector	vel ;
	PVector	acc ;
	Value	decay ;
	Value	randp ;
	Value	weight ;
	Value	max_speed ;
	// th 6.3.00
	Value   max_accel ;
	//
	Weather *weather ;
	MPObjectType type ;
	
	MPObject() { MPObject(NULLCHAR, NULLCHAR); }
	MPObject(const Name name, const Name short_name) ;
	inline void	operator ++() { _inc(); } ;
	void _inc() ;
        void _turn() ;
        
	PVector noise() ;
	PVector wind() ;
	void Set(const MPObjectType& t, const PVector& p, const Value& s,
				const Angle& ang, const PVector& v, const PVector& a,
				const Value& d, const Value& r=0.0) ;
	Logical collisionP(const MPObject& obj) ;
	void collisionCheck(MPObjectTable& tbl) ;
	void collide(MPObject& obj) ;
	void step(MPObjectTable& tbl) ;
	void push(PVector f) ;
	void rotate(Angle a) ;
	void rotate_to(Angle dir) ;
	void dash(Value f) ;
	
	/* new collision stuff */
	PVector post_col_pos;
	int col_count;
	bool col;
} ;

extern ostream& operator<< (ostream& o, const MPObject& v) ;

bool detectCollision ( MPObject* a, MPObject* b );
void calcCollPos ( MPObject* a, MPObject* b, const PlayMode& playmode );
void moveToCollPos ( MPObject* obj );
void updateCollVel ( MPObject* obj );


/*
 *===================================================================
 *Part: Object Table
 *===================================================================
 */

class PObjectTable {
 public:
	TheNumber	Max ;
	TheNumber	n ;
	PObject** object ;

	PObjectTable(const TheNumber max = 128) ;
	~PObjectTable() ;
	PObject* AssignObject(PObject* obj) ;
	/* pfr 06/07/200 */
	PObject* NewObject() { return NewObject(NULLCHAR, NULLCHAR); }
	PObject* NewObject(const Name name,
			   const Name short_name) ;
} ;
extern ostream& operator<< (ostream& o, const PObjectTable& tbl) ;

class MPObjectTable {
  public:
	void _inc( const PlayMode& playmode ) ;
	TheNumber	Max ;
	TheNumber	n ;
	MPObject** object ;

	MPObjectTable(const TheNumber max = 128) ;
	~MPObjectTable() ;
	MPObject* AssignObject(MPObject* obj) ;
	/* pfr 06/07/200 */
	MPObject* NewObject() { return NewObject(NULLCHAR, NULLCHAR);}
	MPObject* NewObject(const Name name,
			    const Name short_name) ;
	//inline void	operator++ () { _inc(); } ;
        void _turn();
        
	/*inline void	operator ++(int dummy) { _inc(); } ;*/
	void collisionCheck(MPObject* obj) ;

	// new collision checks
	void collisions ( const PlayMode& playmode );
} ;
extern ostream& operator<< (ostream& o, const MPObjectTable& tbl) ;


/*
 *===================================================================
 *Part: Player & Team
 *===================================================================
 */

class Client {
 public:
	unsigned long host ;
	int	port ;
} ;

class Team ;

class Player : public MPObject {
 public:
	Team *team ;
	ID	unum ;		/* uniform number */
	Name unumstr ;
	Client client ;
	Name name_far ;
	Name name_toofar ;
	/* th 19.05.00 */
	Name short_name_far ;
	Name short_name_toofar ;
	Value stamina ;
	Value recovery ;
	Value effort ;
	Value vis_angle ;
	Value defangle ;
	Value vis_distance ;
	Value inertia_moment ;
	Value version ;
#ifdef	NEW_QSTEP
	Value dist_qstep_player ; /* each Player quantize step of distance */
	Value land_qstep_player ; /* each Player quantize step of distance for landmark */
	Value dir_qstep_player ; /* each Player quantize step of direction */
#endif

	Angle angle_neck ;
	Angle angle_neck_committed ;
	Angle angle_body ;

	// pfr 8/14/00: for RC2000 evaluation
	Value kick_rand;

	int vis_send ;
	int sendcnt ;
	int highquality ;
	int command_done ;
	int turn_neck_done ;
	int alive ;
	int hear_capacity_from_teammate ;
	int hear_capacity_from_opponent ;
	int goalie ;
	int goalie_catch_ban ;
	Value goalie_catch_probability ;
	int goalie_moves_since_catch;

	int count_kick ;
	int count_dash ;
	int count_turn ;
	int count_catch ;
	int count_move ;
	int count_turn_neck ;
	int count_change_view ;
	int count_say ;
	
	int offside_mark ;
	PVector offside_pos ;
	PVector relative_ball_ang ;

	const HetroPlayer *player_type;
	int player_type_id;
	
	int done_received; //pfr:SYNCH

	Player(Team *team, Stadium *stad, ID number) ;
	void init(Value version, int goalie_flag) ;
	void rotate(Angle a) ;
	void rotate_to(Angle a) ;
	void turn(Value a) ;
	void turn_neck(Value a) ;
	void kick(Value power, Angle dir) ;
	void goalie_catch(Angle dir) ;

	void parse_command(char *command) ;
	int scanField(void) ;
	void send_visual_info(void) ;
	void send_audio_info(char *message) ;	
	void say(char *message) ;
	void sense_body(void) ;
	void send_score(void) ;
	void udp_send(const char *message,
			const Logical logp=TRUE) ;

	void setPlayerType ( const int& );
	void substitute ( const int& );

	/* contributed by Artur Merke */
	void send_fullstate_information(void); 
} ;

class OnlineCoach ;

class Team {
 public:
	Stadium	*stadium ;
	int enable ;
	Name name ;
	Side side ;
	Name color ;
	TheNumber point ;
#ifdef	NEW_QSTEP
	Value	dist_qstep_team ;
	Value	land_qstep_team ;
	Value	dir_qstep_team ;
#endif
	// pfr 8/14/00: for RC2000 evaluation
	Value prand_factor_team; //factor to multiply prand by
        Value kick_rand_factor_team; //factor to multiple kick_rand by

	TheNumber Max ;
	TheNumber n ;
	int goalie ;
	Player **player ;

	int subs;
	int ptype_count [ 7 ];
	
	Team(Stadium *stad, const Side side) ;
	Player *newPlayer(Value version, int goalie_flag) ;
	OnlineCoach *olcoach ;
} ;

#endif

#endif // _H_OBJECT
