/* -*- Mode: C -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

#include <string.h>
#include <stdlib.h>
#include "paramreader.h"

ParamReader::ParamReader()
{
}

ParamReader::~ParamReader()
{
}


void ParamReader::readParams ( const char* path, const option_t *opt )
{
    FILE *fp ;

    if ( (fp = fopen ( path, "r" ) ) == NULL )
	{
	cerr << "cat't open config file " << path << endl ;
	return ;
	}

    char buf[MaxMesg] ;

    int line_no = 0;
    while( fgets ( buf, MaxMesg, fp ) != NULL)
      {
	line_no++;
      /* ignore remark line */
	if (buf[0] == '#' || buf[0] == '\n')
	  continue ;

	/* replace from ':' to ' ' */
	char *t = buf ;
	while ( *t != NULLCHAR )
	  {
	    if (*t == ':') *t = ' ' ;
	    t++ ;
	  }

	int n, p ;
	char com[256] ;
	char onoff[16] ;

	n = sscanf(buf,"%s", com) ;

	if (n < 1)
	  {
	    cerr << "Illegal line ( " << path << " line " << line_no << " ) : " << buf;
	    continue ;
	  }

	bool match = false;
	for ( p = 0 ; opt[p].optname[0] != 0 ; p++)
	  {
	    if ( strcmp ( com, opt[p].optname) )
	      continue ;

	    match = true;

	    if (opt[p].vptr == NULL)
	      continue;
	    
	    /* match */
	    switch(opt[p].vsize)
	      {
	      case V_INT:
		n = sscanf ( buf, "%s %d", com, (int *)opt[p].vptr ) ;
		break ;

	      case V_STRING:
		n = sscanf ( buf, "%s %s", com, (char *)opt[p].vptr ) ;  
		break ;

	      case V_DOUBLE:
		n = sscanf ( buf, "%s %lf", com, (double *)opt[p].vptr ) ;
		break ;

	      case V_BOOL:
		n = 2 ;
		*((int *)opt[p].vptr) = TRUE ;
		break ;

	      case V_ONOFF:
		n = sscanf(buf, "%s %s", com, onoff) ;
		*((int *)opt[p].vptr) = (!strcmp(onoff, "on")) ? TRUE : FALSE ;
		break ;
	      }

	    break ;
	  }

	if (!match) {
	  cerr << "Unrecognized parameter in file '"
	       << path << "': " << com << endl;
	}
	
	
      }

    fclose(fp) ;
}


void ParamReader::readOptions ( const int& orig_argc, const char* const orig_argv[], const option_t opt[] )
{
    int p ;
    int argc = orig_argc;
    const char * const *argv = orig_argv;
    char buf[MaxMesg];
    
    argc--;
    argv++;
    
    while (argc > 0) //D.N.
	{
	if ( sscanf ( *argv, "-%s", buf ) != 1 )
	    {
	    argv++ ;
	    argc-- ;
	    continue;
	    }

	bool match = false;
	for (p = 0 ; opt[p].optname[0] != 0 ; p++)
	    {
	    if (strcmp( buf, opt[p].optname))
		continue ;

	    /* match */
	    match = true;

	    if (opt[p].vptr == NULL)
	      continue;

	    argv++ ;
	    argc-- ;

		if(argc <= 0) break; //D.N

	    switch(opt[p].vsize)
		{
		case V_INT:
		    *((int *)opt[p].vptr) = atoi(*argv) ;
		    break ;

		case V_STRING:
		    strcpy((char *)opt[p].vptr, *argv) ;
		    break ;

		case V_DOUBLE:
		    *((double *)opt[p].vptr) = atof(*argv) ;
		    break ;

		case V_BOOL:
		    *((int *)opt[p].vptr) = TRUE ;
		    argv-- ;
		    argc++ ;
		    break ;

		case V_ONOFF:
		    *((int *)opt[p].vptr) = (!strcmp(*argv, "on")) ? TRUE : FALSE ;
		    break ;
		}

	    break ;
	    }

	if (!match) {
	  /* pfr: It's "okay" for options to be unrecognized on the command line because
	     it may be intended for the another parameter reader instantiation */
	  //cerr << "Unrecognized parameter on command line: " << buf << endl;
	}

	argv++ ;
	argc-- ;
	}

}
