/* -*- Mode: C -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */
#include <winsock.h>

#undef V_INT
#undef V_BOOL
#undef V_NONE
#undef INFINITE
#undef max
#undef min

#include <string.h>

#include "playerparam.h"

const int PlayerParam::DEFAULT_PLAYER_TYPES = 7;
const int PlayerParam::DEFAULT_SUBS_MAX = 3;
const int PlayerParam::DEFAULT_PT_MAX = 3;

const double PlayerParam::DEFAULT_PLAYER_SPEED_MAX_DELTA_MIN = 0.0;
const double PlayerParam::DEFAULT_PLAYER_SPEED_MAX_DELTA_MAX = 0.2;
const double PlayerParam::DEFAULT_STAMINA_INC_MAX_DELTA_FACTOR = -100.0;
    
const double PlayerParam::DEFAULT_PLAYER_DECAY_DELTA_MIN = 0.0;
const double PlayerParam::DEFAULT_PLAYER_DECAY_DELTA_MAX = 0.2;
const double PlayerParam::DEFAULT_INERTIA_MOMENT_DELTA_FACTOR = 25.0;

const double PlayerParam::DEFAULT_DASH_POWER_RATE_DELTA_MIN = 0.0;
const double PlayerParam::DEFAULT_DASH_POWER_RATE_DELTA_MAX = 0.002;
const double PlayerParam::DEFAULT_PLAYER_SIZE_DELTA_FACTOR = -100.0;

const double PlayerParam::DEFAULT_KICKABLE_MARGIN_DELTA_MIN = 0.0;
const double PlayerParam::DEFAULT_KICKABLE_MARGIN_DELTA_MAX = 0.2;
const double PlayerParam::DEFAULT_KICK_RAND_DELTA_FACTOR = 0.5;

const double PlayerParam::DEFAULT_EXTRA_STAMINA_DELTA_MIN = 0.0;
const double PlayerParam::DEFAULT_EXTRA_STAMINA_DELTA_MAX = 100.0;
const double PlayerParam::DEFAULT_EFFORT_MAX_DELTA_FACTOR = -0.002;
const double PlayerParam::DEFAULT_EFFORT_MIN_DELTA_FACTOR = -0.002;

const int    PlayerParam::DEFAULT_RANDOM_SEED = -1; //negative means generate a new seed

PlayerParam::PlayerParam(const int& argc, const char * const *argv ) 
{ 
    getOptions ( argc, argv ) ;
}

PlayerParam::~PlayerParam()
{
}

void PlayerParam::getOptions(const int& argc, const char * const *argv )
{
    option_t opt[] = {
	{ "player_types",			(void *)&player_types,			V_INT },
	{ "subs_max",				(void *)&subs_max, 			V_INT },		
	{ "pt_max",				(void *)&pt_max, 			V_INT },		
	{ "player_speed_max_delta_min",		(void *)&player_speed_max_delta_min,	V_DOUBLE },
	{ "player_speed_max_delta_max",		(void *)&player_speed_max_delta_max,	V_DOUBLE },
	{ "stamina_inc_max_delta_factor",	(void *)&stamina_inc_max_delta_factor,	V_DOUBLE },
	{ "player_decay_delta_min",		(void *)&player_decay_delta_min,	V_DOUBLE },
	{ "player_decay_delta_max",		(void *)&player_decay_delta_max,	V_DOUBLE },
	{ "inertia_moment_delta_factor",	(void *)&inertia_moment_delta_factor,	V_DOUBLE },
	{ "dash_power_rate_delta_min",		(void *)&dash_power_rate_delta_min,	V_DOUBLE },
	{ "dash_power_rate_delta_max",		(void *)&dash_power_rate_delta_max,	V_DOUBLE },
	{ "player_size_delta_factor",		(void *)&player_size_delta_factor,	V_DOUBLE },
	{ "kickable_margin_delta_min",		(void *)&kickable_margin_delta_min,	V_DOUBLE },
	{ "kickable_margin_delta_max",		(void *)&kickable_margin_delta_max,	V_DOUBLE },
	{ "kick_rand_delta_factor",		(void *)&kick_rand_delta_factor,	V_DOUBLE },
	{ "extra_stamina_delta_min",		(void *)&extra_stamina_delta_min,	V_DOUBLE },
	{ "extra_stamina_delta_max",		(void *)&extra_stamina_delta_max,	V_DOUBLE },
	{ "effort_max_delta_factor",		(void *)&effort_max_delta_factor,	V_DOUBLE },
	{ "effort_min_delta_factor",		(void *)&effort_min_delta_factor,	V_DOUBLE },
	{ "random_seed",              		(void *)&random_seed,                  	V_INT },
	/* these two lines are just to keep the server from complaining about
	   unknown options */
	{"sfile",                               NULL,                                   V_INT},
	{"pfile",                               NULL,                                   V_INT},
	{"\0",					NULL, 					0 }
	};

    player_types = PlayerParam::DEFAULT_PLAYER_TYPES;
    subs_max = PlayerParam::DEFAULT_SUBS_MAX;
    pt_max = PlayerParam::DEFAULT_PT_MAX;
    
    player_speed_max_delta_min = PlayerParam::DEFAULT_PLAYER_SPEED_MAX_DELTA_MIN;
    player_speed_max_delta_max = PlayerParam::DEFAULT_PLAYER_SPEED_MAX_DELTA_MAX;
    stamina_inc_max_delta_factor = PlayerParam::DEFAULT_STAMINA_INC_MAX_DELTA_FACTOR;
    
    player_decay_delta_min = PlayerParam::DEFAULT_PLAYER_DECAY_DELTA_MIN;
    player_decay_delta_max = PlayerParam::DEFAULT_PLAYER_DECAY_DELTA_MAX;
    inertia_moment_delta_factor = PlayerParam::DEFAULT_INERTIA_MOMENT_DELTA_FACTOR;

    dash_power_rate_delta_min = PlayerParam::DEFAULT_DASH_POWER_RATE_DELTA_MIN;
    dash_power_rate_delta_max = PlayerParam::DEFAULT_DASH_POWER_RATE_DELTA_MAX;
    player_size_delta_factor = PlayerParam::DEFAULT_PLAYER_SIZE_DELTA_FACTOR;

    kickable_margin_delta_min = PlayerParam::DEFAULT_KICKABLE_MARGIN_DELTA_MIN;
    kickable_margin_delta_max = PlayerParam::DEFAULT_KICKABLE_MARGIN_DELTA_MAX;
    kick_rand_delta_factor = PlayerParam::DEFAULT_KICK_RAND_DELTA_FACTOR;

    extra_stamina_delta_min = PlayerParam::DEFAULT_EXTRA_STAMINA_DELTA_MIN;
    extra_stamina_delta_max = PlayerParam::DEFAULT_EXTRA_STAMINA_DELTA_MAX;
    effort_max_delta_factor = PlayerParam::DEFAULT_EFFORT_MAX_DELTA_FACTOR;
    effort_min_delta_factor = PlayerParam::DEFAULT_EFFORT_MIN_DELTA_FACTOR;

    random_seed = PlayerParam::DEFAULT_RANDOM_SEED;
    
    for(int i = 1 ; i < argc ; i++)
	{
	if ( strcmp( argv [ i ], "-pfile" ) == 0)
	    {
	    i++;
		if(i < argc)
		{
			readParams ( argv [ i ], opt );
		}
		else //chyba
		{
			cerr << "Chyba nazov suboru! (Aj tu by to spadlo)" << endl;
		}

	    }
	}
    
    readOptions ( argc, argv, opt );

    if ( player_types < 1 )
      {
	cerr << "Error: must have at least one player type\n";
	exit ( EXIT_FAILURE );
      }
    
}

player_params_t PlayerParam::convertToStruct ()
{
    player_params_t tmp;

    tmp.player_types = htons( (short) player_types );
    tmp.subs_max = htons( (short) subs_max );
    tmp.pt_max = htons( (short) pt_max );
    
    tmp.player_speed_max_delta_min = htonl( (long)rint((SHOWINFO_SCALE2 * player_speed_max_delta_min) ));
    tmp.player_speed_max_delta_max = htonl( (long)rint((SHOWINFO_SCALE2 * player_speed_max_delta_max) ));
    tmp.stamina_inc_max_delta_factor = htonl( (long)rint((SHOWINFO_SCALE2 * stamina_inc_max_delta_factor) ));
    
    tmp.player_decay_delta_min = htonl( (long)rint((SHOWINFO_SCALE2 * player_decay_delta_min) ));
    tmp.player_decay_delta_max = htonl( (long)rint((SHOWINFO_SCALE2 * player_decay_delta_max) ));
    tmp.inertia_moment_delta_factor = htonl( (long)rint((SHOWINFO_SCALE2 * inertia_moment_delta_factor) ));

    tmp.dash_power_rate_delta_min = htonl( (long)rint((SHOWINFO_SCALE2 * dash_power_rate_delta_min) ));
    tmp.dash_power_rate_delta_max = htonl( (long)rint((SHOWINFO_SCALE2 * dash_power_rate_delta_max) ));
    tmp.player_size_delta_factor = htonl( (long)rint((SHOWINFO_SCALE2 * player_size_delta_factor) ));

    tmp.kickable_margin_delta_min = htonl( (long)rint((SHOWINFO_SCALE2 * kickable_margin_delta_min) ));
    tmp.kickable_margin_delta_max = htonl( (long)rint((SHOWINFO_SCALE2 * kickable_margin_delta_max) ));
    tmp.kick_rand_delta_factor = htonl( (long)rint((SHOWINFO_SCALE2 * kick_rand_delta_factor) ));

    tmp.extra_stamina_delta_min = htonl( (long)rint((SHOWINFO_SCALE2 * extra_stamina_delta_min) ));
    tmp.extra_stamina_delta_max = htonl( (long)rint((SHOWINFO_SCALE2 * extra_stamina_delta_max) ));
    tmp.effort_max_delta_factor = htonl( (long)rint((SHOWINFO_SCALE2 * effort_max_delta_factor) ));
    tmp.effort_min_delta_factor = htonl( (long)rint((SHOWINFO_SCALE2 * effort_min_delta_factor) ));
    tmp.random_seed = htonl( (long)random_seed );
    
    return tmp;
}


void PlayerParam::getSettingsString ( char* buf )
{
    ostrstream ost(buf, MaxStringSize) ;

    ost << "(player_param "
	<< player_types << " "
	<< subs_max << " "
	<< pt_max << " "
    
	<< player_speed_max_delta_min << " "
	<< player_speed_max_delta_max << " "
	<< stamina_inc_max_delta_factor << " "
    
	<< player_decay_delta_min << " "
	<< player_decay_delta_max << " "
	<< inertia_moment_delta_factor << " "

	<< dash_power_rate_delta_min << " "
	<< dash_power_rate_delta_max << " "
	<< player_size_delta_factor << " "

	<< kickable_margin_delta_min << " "
	<< kickable_margin_delta_max << " "
	<< kick_rand_delta_factor << " "

	<< extra_stamina_delta_min << " "
	<< extra_stamina_delta_max << " "
	<< effort_max_delta_factor << " "
	<< effort_min_delta_factor
	<< ")" << endl << ends;

    //do we want to add random_seed here? pfr 7/31/01
}
