/* -*- Mode: C -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

#ifndef _PLAYER_PARAM_H_
#define _PLAYER_PARAM_H_

//#include <sys/param.h> /* needed for htonl, htons, ... */
//#include <netinet/in.h>
#include <strstream>
using namespace std;

#include "paramreader.h"
#include "utility.h"
#include "types.h"

class PlayerParam : public ParamReader
{
public:

    PlayerParam ( const int& argc, const char * const *argv );
    virtual ~PlayerParam ();

    player_params_t convertToStruct ();
    void getSettingsString ( char* );

    int playerTypes () const { return player_types; }
    int subsMax () const { return subs_max; }
    int ptMax () const { return pt_max; }

    double playerSpeedMaxDeltaMin () const { return player_speed_max_delta_min; }
    
    double playerSpeedMaxDeltaMax () const { return player_speed_max_delta_max; }
    double staminaIncMaxDeltaFactor () const { return stamina_inc_max_delta_factor; }
    
    double playerDecayDeltaMin () const { return player_decay_delta_min; }
    double playerDecayDeltaMax () const { return player_decay_delta_max; }
    double inertiaMomentDeltaFactor () const { return inertia_moment_delta_factor; }

    double dashPowerRateDeltaMin () const { return dash_power_rate_delta_min; }
    double dashPowerRateDeltaMax () const { return dash_power_rate_delta_max; }
    double playerSizeDeltaFactor () const { return player_size_delta_factor; }

    double kickableMarginDeltaMin () const { return kickable_margin_delta_min; }
    double kickableMarginDeltaMax () const { return kickable_margin_delta_max; }
    double kickRandDeltaFactor () const { return kick_rand_delta_factor; }

    double extraStaminaDeltaMin () const { return extra_stamina_delta_min; }
    double extraStaminaDeltaMax () const { return extra_stamina_delta_max; }
    double effortMaxDeltaFactor () const { return effort_max_delta_factor; }
    double effortMinDeltaFactor () const { return effort_min_delta_factor; }

    int randomSeed () const { return random_seed; }
    
protected:
    virtual void getOptions(const int& argc, const char * const *argv );

    
private:
    static const int DEFAULT_PLAYER_TYPES;
    static const int DEFAULT_SUBS_MAX;
    static const int DEFAULT_PT_MAX;

    static const double DEFAULT_PLAYER_SPEED_MAX_DELTA_MIN;
    static const double DEFAULT_PLAYER_SPEED_MAX_DELTA_MAX;
    static const double DEFAULT_STAMINA_INC_MAX_DELTA_FACTOR;
    
    static const double DEFAULT_PLAYER_DECAY_DELTA_MIN;
    static const double DEFAULT_PLAYER_DECAY_DELTA_MAX;
    static const double DEFAULT_INERTIA_MOMENT_DELTA_FACTOR;

    static const double DEFAULT_DASH_POWER_RATE_DELTA_MIN;
    static const double DEFAULT_DASH_POWER_RATE_DELTA_MAX;
    static const double DEFAULT_PLAYER_SIZE_DELTA_FACTOR;

    static const double DEFAULT_KICKABLE_MARGIN_DELTA_MIN;
    static const double DEFAULT_KICKABLE_MARGIN_DELTA_MAX;
    static const double DEFAULT_KICK_RAND_DELTA_FACTOR;

    static const double DEFAULT_EXTRA_STAMINA_DELTA_MIN;
    static const double DEFAULT_EXTRA_STAMINA_DELTA_MAX;
    static const double DEFAULT_EFFORT_MAX_DELTA_FACTOR;
    static const double DEFAULT_EFFORT_MIN_DELTA_FACTOR;

    static const int    DEFAULT_RANDOM_SEED;

    int player_types;
    int subs_max;
    int pt_max;
  
    double player_speed_max_delta_min;
    double player_speed_max_delta_max;
    double stamina_inc_max_delta_factor;
    
    double player_decay_delta_min;
    double player_decay_delta_max;
    double inertia_moment_delta_factor;

    double dash_power_rate_delta_min;
    double dash_power_rate_delta_max;
    double player_size_delta_factor;

    double kickable_margin_delta_min;
    double kickable_margin_delta_max;
    double kick_rand_delta_factor;

    double extra_stamina_delta_min;
    double extra_stamina_delta_max;
    double effort_max_delta_factor;
    double effort_min_delta_factor;

    int random_seed;
};

#endif
