/* -*- Mode: C -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */
#include <winsock.h>

#undef V_INT
#undef V_BOOL
#undef V_NONE
#undef INFINITE
#undef max
#undef min

#include <string.h>

#include "serverparam.h"

ServerParam::ServerParam(const int& argc, const char * const *argv ) 
{ 
	getOptions ( argc, argv ) ;
}

ServerParam::~ServerParam()
{
}

void ServerParam::getOptions(const int& argc, const char * const *argv )
{
	option_t opt[] = {
		{"goal_width",		(void *)&gwidth,		V_DOUBLE},
		{"player_size",		(void *)&psize,			V_DOUBLE},
		{"player_decay",	(void *)&pdecay,		V_DOUBLE},
		{"player_rand",		(void *)&prand,			V_DOUBLE},
		{"player_weight",	(void *)&pweight,		V_DOUBLE},
		{"player_speed_max",	(void *)&pspeed_max,		V_DOUBLE},
		// th 6.3.00
		{"player_accel_max", 	(void *)&paccel_max,		V_DOUBLE},
		//
		{"stamina_max",		(void *)&stamina_max,	V_DOUBLE},
		{"stamina_inc_max",	(void *)&stamina_inc,	V_DOUBLE},
		{"recover_dec_thr",	(void *)&recover_dthr,	V_DOUBLE},
		{"recover_min",		(void *)&recover_min, 	V_DOUBLE},
		{"recover_dec",		(void *)&recover_dec, 	V_DOUBLE},
		{"effort_dec_thr",	(void *)&effort_dthr,	V_DOUBLE},
		{"effort_min",		(void *)&effort_min,	V_DOUBLE},
		{"effort_dec",		(void *)&effort_dec,	V_DOUBLE},
		{"effort_inc_thr",	(void *)&effort_ithr,	V_DOUBLE},
		{"effort_inc",		(void *)&effort_inc,	V_DOUBLE},
		// pfr 8/14/00: for RC2000 evaluation
                {"kick_rand",           (void *)&kick_rand,           V_DOUBLE},
                {"team_actuator_noise", (void *)&team_actuator_noise, V_ONOFF},
                {"prand_factor_l",      (void *)&prand_factor_l,      V_DOUBLE},
                {"prand_factor_r",      (void *)&prand_factor_r,      V_DOUBLE},
                {"kick_rand_factor_l",  (void *)&kick_rand_factor_l,  V_DOUBLE},
                {"kick_rand_factor_r",  (void *)&kick_rand_factor_r,  V_DOUBLE},

		{"ball_size",		(void *)&bsize,			V_DOUBLE},
		{"ball_decay",		(void *)&bdecay,		V_DOUBLE},
		{"ball_rand",		(void *)&brand,			V_DOUBLE},
		{"ball_weight",		(void *)&bweight,		V_DOUBLE},
		{"ball_speed_max",	(void *)&bspeed_max,	V_DOUBLE},
		// th 6.3.00
		{"ball_accel_max",	(void *)&baccel_max,		V_DOUBLE},
		//
		{"dash_power_rate",	(void *)&dprate,		V_DOUBLE},
		{"kick_power_rate",	(void *)&kprate,		V_DOUBLE},
		{"kickable_margin",	(void *)&kmargin,		V_DOUBLE},
		{"control_radius",	(void *)&ctlradius,		V_DOUBLE},

		{"catch_probability",	(void *)&catch_prob,	V_DOUBLE},
		{"catchable_area_l",    (void *)&catch_area_l,	V_DOUBLE},
		{"catchable_area_w",    (void *)&catch_area_w,	V_DOUBLE},
		{"goalie_max_moves",    (void *)&goalie_max_moves, V_INT},
		
		{"wind_dir",		(void *)&windir,		V_DOUBLE},
		{"maxpower",		(void *)&maxp,			V_DOUBLE},
		{"minpower",		(void *)&minp,			V_DOUBLE},
		{"maxmoment",		(void *)&maxm,			V_DOUBLE},
		{"minmoment",		(void *)&minm,			V_DOUBLE},
		{"maxneckmoment",	(void *)&maxnm,			V_DOUBLE},
		{"minneckmoment",	(void *)&minnm,			V_DOUBLE},
		{"maxneckang",		(void *)&maxn,			V_DOUBLE},
		{"minneckang",		(void *)&minn,			V_DOUBLE},
		{"visible_angle",	(void *)&visangle,		V_DOUBLE},
		{"visible_distance",(void *)&visdist,		V_DOUBLE},
		{"audio_cut_dist",	(void *)&audio_dist,	V_DOUBLE},
		{"quantize_step",	(void *)&dist_qstep,	V_DOUBLE},
		{"quantize_step_l",	(void *)&land_qstep,	V_DOUBLE},
#ifdef NEW_QSTEP
		{"quantize_step_dir",			(void *)&dir_qstep,	V_DOUBLE},
		{"quantize_step_dist_team_l",	(void *)&dist_qstep_l,	V_DOUBLE},
		{"quantize_step_dist_team_r",	(void *)&dist_qstep_r,	V_DOUBLE},
		{"quantize_step_dist_l_team_l",	(void *)&land_qstep_l,	V_DOUBLE},
		{"quantize_step_dist_l_team_r",	(void *)&land_qstep_r,	V_DOUBLE},
		{"quantize_step_dir_team_l",	(void *)&dir_qstep_l,	V_DOUBLE},
		{"quantize_step_dir_team_r",	(void *)&dir_qstep_r,	V_DOUBLE},
#endif		
		{"ckick_margin",	(void *)&ckmargin,		V_DOUBLE},
		{"wind_dir",		(void *)&windir,		V_DOUBLE},
		{"wind_force",		(void *)&winforce,		V_DOUBLE},
		{"wind_rand",		(void *)&winrand,		V_DOUBLE},
		{"inertia_moment",	(void *)&inertia_moment,V_DOUBLE},
		{"wind_none",		(void *)&win_no,		V_BOOL},
		{"wind_random",		(void *)&win_random,	V_BOOL},
		{"half_time",		(void *)&half_time,		V_INT},
		{"drop_ball_time",	(void *)&drop_time,		V_INT},
		{"port",			(void *)&portnum,		V_INT},
		{"coach_port",		(void *)&coach_pnum,	V_INT},
		{"olcoach_port",	(void *)&olcoach_pnum,	V_INT},
		{"say_coach_cnt_max",	(void *)&say_cnt_max,	V_INT},
		{"say_coach_msg_size",	(void *)&SayCoachMsgSize,	V_INT},
		{"simulator_step",	(void *)&sim_st,		V_INT},
		{"send_step",		(void *)&send_st,		V_INT},
		{"recv_step",		(void *)&recv_st,		V_INT},
		{"sense_body_step",	(void *)&sb_step,	V_INT},
		{"say_msg_size",	(void *)&SayMsgSize,	V_INT},
		{"clang_win_size",      (void *)&clang_win_size, V_INT},
		{"clang_define_win",    (void *)&clang_define_win, V_INT},
		{"clang_meta_win",      (void *)&clang_meta_win, V_INT},
		{"clang_advice_win",    (void *)&clang_advice_win, V_INT},
		{"clang_info_win",      (void *)&clang_info_win, V_INT},
		{"clang_mess_delay",    (void *)&clang_mess_delay, V_INT},
		{"clang_mess_per_cycle",      (void *)&clang_mess_per_cycle, V_INT},
		{"hear_max",		(void *)&hear_max,		V_INT},
		{"hear_inc",		(void *)&hear_inc,		V_INT},
		{"hear_decay",		(void *)&hear_decay,	V_INT},
		{"catch_ban_cycle", (void *)&cban_cycle,	V_INT},
		{"coach",			(void *)&CoachMode,		V_BOOL},
		{"coach_w_referee",	(void *)&CwRMode,		V_BOOL},
		{"old_coach_hear",	(void *)&old_hear,		V_BOOL},
		{"send_vi_step",        (void *)&sv_st,		V_INT},
		{"use_offside",		(void *)&useoffside,	V_ONOFF},
		{"offside_active_area_size",
							(void *)&offside_area,	V_DOUBLE},
		{"forbid_kick_off_offside",
							(void *)&kickoffoffside,V_ONOFF},
		{"log_file",		(void *)&logfile,		V_STRING},
		{"record",			(void *)&recfile,		V_STRING},
		{"record_version",	(void *)&rec_ver,		V_INT},
		{"record_log",		(void *)&rec_log,		V_ONOFF},
		{"record_messages",	(void *)&rec_msg_mode,	V_ONOFF},
		{"send_log",		(void *)&send_log,		V_ONOFF},
		// th 6.3.00
		{"log_times",		(void *)&log_times,		V_ONOFF},
		//
		{"verbose",			(void *)&verbose,		V_ONOFF},
		{"replay",			(void *)&replay,		V_STRING},
		{"offside_kick_margin",
				(void *)&offside_kick_margin,		V_DOUBLE},
		{"slow_down_factor",        (void *)&slow_down_factor,   V_INT},		
		{"synch_mode",              (void *)&synch_mode, V_ONOFF},//pfr:SYNCH
		{"synch_offset",            (void *)&synch_offset, V_INT},//pfr:SYNCH
		{"synch_micro_sleep",       (void *)&synch_micro_sleep, V_INT},//pfr:SYNCH  

		{"start_goal_l",            (void *)&start_goal_l,       V_INT},
		{"start_goal_r",            (void *)&start_goal_r,       V_INT},

		{"fullstate_l",             (void *)&fullstate_l,        V_ONOFF},
		{"fullstate_r",             (void *)&fullstate_r,        V_ONOFF},
		{"slowness_on_top_for_left_team", (void*)&slowness_on_top_for_left_team, V_DOUBLE},
		{"slowness_on_top_for_right_team", (void*)&slowness_on_top_for_right_team, V_DOUBLE},
		/* these two lines are just to keep the server from complaining about
		   unknown options */
		{"sfile",                   NULL, V_INT},
		{"pfile",                   NULL, V_INT},
		
		{"\0",				NULL, 					0}
	} ;

	/* set default parameter */
	effort_init = 0;
	recover_init = 0;
	winang = 0;

	gwidth = GOAL_WIDTH ;
	psize = PLAYER_SIZE ;
	pdecay = PLAYER_DECAY ;
	prand = PLAYER_RAND ;
	pweight = PLAYER_WEIGHT ;
	stamina_max = STAMINA_MAX ;
	stamina_inc = STAMINA_INC_MAX ;
	recover_dthr = RECOVERY_DEC_THR ;
	recover_min = RECOVERY_MIN ;
	recover_dec = RECOVERY_DEC ;
	effort_dthr = EFFORT_DEC_THR ;
	effort_min = EFFORT_MIN ;
	effort_dec = EFFORT_DEC ;
	effort_ithr = EFFORT_INC_THR ;
	effort_inc= EFFORT_INC ;
	inertia_moment = IMPARAM ;
	// pfr 8/14/00: for RC2000 evaluation
        kick_rand = KICK_RAND;
        team_actuator_noise = FALSE;
        prand_factor_l = PRAND_FACTOR_L;
        prand_factor_r = PRAND_FACTOR_R;
        kick_rand_factor_l = KICK_RAND_FACTOR_L;
        kick_rand_factor_r = KICK_RAND_FACTOR_R;

	bsize = BALL_SIZE ;
	bdecay = BALL_DECAY ;
	brand = BALL_RAND ;
	bweight = BALL_WEIGHT ;
	bspeed_max = BALL_SPEED_MAX ;
	pspeed_max = PLAYER_SPEED_MAX ;
	// th 6.3.00
	baccel_max = BALL_ACCEL_MAX ;
	paccel_max = PLAYER_ACCEL_MAX ;
	//
	dprate = DASHPOWERRATE ;
	kprate = KICKPOWERRATE ;
	kmargin = KICKABLE_MARGIN ;
	ctlradius = CONTROL_RADIUS ;
	ckmargin = CORNER_KICK_MARGIN ;
	catch_prob = GOALIE_CATCHABLE_POSSIBILITY ;
	catch_area_l = GOALIE_CATCHABLE_AREA_LENGTH ; 
	catch_area_w = GOALIE_CATCHABLE_AREA_WIDTH ;
	goalie_max_moves = GOALIE_MAX_MOVES;
	maxp = MAXPOWER ;
	minp = MINPOWER ;
	maxm = MAXMOMENT ;
	minm = MINMOMENT ;
	maxnm = MAX_NECK_MOMENT ;
	minnm = MIN_NECK_MOMENT ;
	maxn = MAX_NECK_ANGLE ;
	minn = MIN_NECK_ANGLE ;
	visangle = VisibleAngle ;
	visdist = VisibleDistance ;
	half_time = HALF_TIME ;
	drop_time = DROP_TIME ;
	portnum = DEFAULT_PORT_NUMBER ;
	coach_pnum = COACH_PORT_NUMBER ;
	olcoach_pnum = OLCOACH_PORT_NUMBER ;
	say_cnt_max = DEF_SAY_COACH_CNT_MAX ;
	SayCoachMsgSize = DEF_SAY_COACH_MSG_SIZE ;
	sim_st = SIMULATOR_STEP_INTERVAL_MSEC ;
	send_st = UDP_SEND_STEP_INTERVAL_MSEC ;
	recv_st = UDP_RECV_STEP_INTERVAL_MSEC ;
	sb_step = SENSE_BODY_INTERVAL_MSEC ;
	SayMsgSize = DEF_SAY_MSG_SIZE ;
	clang_win_size = 300;
	clang_define_win = 1;
	clang_meta_win = 1;
	clang_advice_win = 1;
	clang_info_win = 1;
	clang_mess_delay = 50;
	clang_mess_per_cycle = 1;
	CoachMode = FALSE ;
	CwRMode = FALSE ;
	old_hear = FALSE ;
	sv_st = SEND_VISUALINFO_INTERVAL_MSEC ;
	logfile[0] = NULLCHAR ;
	recfile[0] = NULLCHAR ;
	rec_ver = DEFAULT_REC_VERSION ;
	rec_log = FALSE ;
	rec_msg_mode = TRUE ;
	replay[0] = NULLCHAR ;
	audio_dist = AUDIO_CUT_OFF_DIST ;
	hear_max = HEAR_MAX ;
	hear_inc = HEAR_INC ;
	hear_decay = HEAR_DECAY ;
	cban_cycle = GOALIE_CATCH_BAN_CYCLE ;
	slow_down_factor = 1;
	verbose = FALSE ;

	dist_qstep = DIST_QSTEP ;
	land_qstep = LAND_QSTEP ;
#ifdef NEW_QSTEP
	dir_qstep  = DIR_QSTEP ;
#endif

	windir = WIND_DIR ;
	winforce = WIND_FORCE ;
	winrand = WIND_RAND ;
	win_no = FALSE ;
	win_random = FALSE ;
	useoffside = TRUE ;
	offside_area = OFFSIDE_ACTIVE_AREA_SIZE ;
	kickoffoffside = FALSE ; 
	send_log = TRUE ;
	log_times = FALSE;
	offside_kick_margin = OFFSIDE_KICK_MARGIN ;

	//pfr:SYNCH
	synch_mode = false;
	synch_offset = 60;
	synch_micro_sleep = 1;

	start_goal_l = 0;
	start_goal_r = 0;

	fullstate_l = FALSE;
	fullstate_r = FALSE;
	
/*  	// skip command name */
/*  	int tmp_argc = argc-1; */
/*  	argv++ ; argc-- ; */

	/* first, search option '-file' */
	int i ;
#ifdef NEW_QSTEP
	int defined_qstep_l = FALSE;
	int defined_qstep_r = FALSE;
	int defined_qstep_l_l = FALSE;
	int defined_qstep_l_r = FALSE;
	int defined_qstep_dir_l = FALSE;
	int defined_qstep_dir_r = FALSE;
#endif

	slowness_on_top_for_left_team = 1.0;
	slowness_on_top_for_right_team = 1.0;

	for(i = 1 ; i < argc ; i++)
	    {
	    if ( strcmp( argv [ i ], "-sfile" ) == 0)
	        {
		i++;
		if(i < argc)
		{
			readParams ( argv [ i ], opt );
		}
		else //chyba D.N.
		{
			cerr << "Chyba nazov suboru! (Inac by to spadlo)" << endl;
		}

		}
	    }

	readOptions ( argc, argv, opt );

#ifdef NEW_QSTEP
	if ( !defined_qstep_l )
		dist_qstep_l = dist_qstep ;
	if ( !defined_qstep_r )
		dist_qstep_r = dist_qstep ;
	if ( !defined_qstep_l_l )
		land_qstep_l = land_qstep ;
	if ( !defined_qstep_l_r )
		land_qstep_r = land_qstep ;
	if ( !defined_qstep_dir_l )
		dir_qstep_l = dir_qstep ;
	if ( !defined_qstep_dir_r )
		dir_qstep_r = dir_qstep ;
#endif

	visangle = Deg2Rad(visangle) ;
	half_time = half_time * 1000 / sim_st ;
	/*	send_st /= 4 ; */
	kickable_area = kmargin + bsize + psize ;
	ctlradius_width = ctlradius - psize ;
	/* apply the slow down factor */
	sim_st  *= slow_down_factor;
	sb_step *= slow_down_factor;
	sv_st   *= slow_down_factor;
	send_st *= slow_down_factor;
	synch_offset *= slow_down_factor; //pfr:SYNCH
	lcm_st = lcm (sim_st, lcm (send_st, lcm (recv_st, lcm (sb_step, sv_st) ) ) );
	lcm_st = lcm (lcm_st, (synch_mode ? synch_offset : 1)); //pfr:SYNCH

}

server_params_t ServerParam::convertToStruct ()
{
    server_params_t tmp;

    tmp.gwidth = htonl( (long)(SHOWINFO_SCALE2 * gwidth) );					
    tmp.inertia_moment = htonl( (long)(SHOWINFO_SCALE2 * inertia_moment) );			
    tmp.psize = htonl( (long)(SHOWINFO_SCALE2 * psize) );					
    tmp.pdecay = htonl( (long)(SHOWINFO_SCALE2 * pdecay) );					
    tmp.prand = htonl( (long)(SHOWINFO_SCALE2 * prand) );					
    tmp.pweight = htonl( (long)(SHOWINFO_SCALE2 * pweight) );					
    tmp.pspeed_max = htonl( (long)(SHOWINFO_SCALE2 * pspeed_max) );				
    tmp.paccel_max = htonl( (long)(SHOWINFO_SCALE2 * paccel_max) );				
    tmp.stamina_max = htonl( (long)(SHOWINFO_SCALE2 * stamina_max) );				
    tmp.stamina_inc = htonl( (long)(SHOWINFO_SCALE2 * stamina_inc) );				
    tmp.recover_init = htonl( (long)(SHOWINFO_SCALE2 * recover_init) );			
    tmp.recover_dthr = htonl( (long)(SHOWINFO_SCALE2 * recover_dthr) );			
    tmp.recover_min = htonl( (long)(SHOWINFO_SCALE2 * recover_min) );				
    tmp.recover_dec = htonl( (long)(SHOWINFO_SCALE2 * recover_dec) );				
    tmp.effort_init = htonl( (long)(SHOWINFO_SCALE2 * effort_init) );				
    tmp.effort_dthr = htonl( (long)(SHOWINFO_SCALE2 * effort_dthr) );				
    tmp.effort_min = htonl( (long)(SHOWINFO_SCALE2 * effort_min) );				
    tmp.effort_dec = htonl( (long)(SHOWINFO_SCALE2 * effort_dec) );				
    tmp.effort_ithr = htonl( (long)(SHOWINFO_SCALE2 * effort_ithr) );				
    tmp.effort_inc = htonl( (long)(SHOWINFO_SCALE2 * effort_inc) );				
    tmp.kick_rand = htonl( (long)(SHOWINFO_SCALE2 * kick_rand) );                                
    tmp.team_actuator_noise = htons( (short) team_actuator_noise );                        
    tmp.prand_factor_l = htonl( (long)(SHOWINFO_SCALE2 * prand_factor_l ) );                           
    tmp.prand_factor_r = htonl( (long)(SHOWINFO_SCALE2 * prand_factor_r ) );                           
    tmp.kick_rand_factor_l = htonl( (long)(SHOWINFO_SCALE2 * kick_rand_factor_l ) );                       
    tmp.kick_rand_factor_r = htonl( (long)(SHOWINFO_SCALE2 * kick_rand_factor_r ) );                       
    tmp.bsize = htonl( (long)(SHOWINFO_SCALE2 * bsize) );					
    tmp.bdecay = htonl( (long)(SHOWINFO_SCALE2 * bdecay) );					
    tmp.brand = htonl( (long)(SHOWINFO_SCALE2 * brand) );					
    tmp.bweight = htonl( (long)(SHOWINFO_SCALE2 * bweight) );					
    tmp.bspeed_max = htonl( (long)(SHOWINFO_SCALE2 * bspeed_max) );				
    tmp.baccel_max = htonl( (long)(SHOWINFO_SCALE2 * baccel_max ) );				
    tmp.dprate = htonl( (long)(SHOWINFO_SCALE2 * dprate) );					
    tmp.kprate = htonl( (long)(SHOWINFO_SCALE2 * kprate) );					
    tmp.kmargin = htonl( (long)(SHOWINFO_SCALE2 * kmargin) );					
    tmp.ctlradius = htonl( (long)(SHOWINFO_SCALE2 * ctlradius) );				
    tmp.ctlradius_width = htonl( (long)(SHOWINFO_SCALE2 * ctlradius_width) );			
    tmp.maxp = htonl( (long)(SHOWINFO_SCALE2 * maxp) );					
    tmp.minp = htonl( (long)(SHOWINFO_SCALE2 * minp) );					
    tmp.maxm = htonl( (long)(SHOWINFO_SCALE2 * maxm) );					
    tmp.minm = htonl( (long)(SHOWINFO_SCALE2 * minm) );					
    tmp.maxnm = htonl( (long)(SHOWINFO_SCALE2 * maxnm) );					
    tmp.minnm = htonl( (long)(SHOWINFO_SCALE2 * minnm) );					
    tmp.maxn = htonl( (long)(SHOWINFO_SCALE2 * maxn) );					
    tmp.minn = htonl( (long)(SHOWINFO_SCALE2 * minn) );					
    tmp.visangle = htonl( (long)(SHOWINFO_SCALE2 * visangle) );				
    tmp.visdist = htonl( (long)(SHOWINFO_SCALE2 * visdist) );					
    tmp.windir = htonl( (long)(SHOWINFO_SCALE2 * windir) );					
    tmp.winforce = htonl( (long)(SHOWINFO_SCALE2 * winforce) );				
    tmp.winang = htonl( (long)(SHOWINFO_SCALE2 * winang) );					
    tmp.winrand = htonl( (long)(SHOWINFO_SCALE2 * winrand) );					
    tmp.kickable_area = htonl( (long)(SHOWINFO_SCALE2 * kickable_area) );			
    tmp.catch_area_l = htonl( (long)(SHOWINFO_SCALE2 * catch_area_l) );			
    tmp.catch_area_w = htonl( (long)(SHOWINFO_SCALE2 * catch_area_w) );			
    tmp.catch_prob = htonl( (long)(SHOWINFO_SCALE2 * catch_prob) );				
    tmp.goalie_max_moves = htons( (short) goalie_max_moves );                 	
    tmp.ckmargin = htonl( (long)(SHOWINFO_SCALE2 * ckmargin) );				
    tmp.offside_area = htonl( (long)(SHOWINFO_SCALE2 * offside_area) );			
    tmp.win_no = htons( (short) win_no) ;					
    tmp.win_random = htons( (short) win_random) ;				
    tmp.say_cnt_max = htons( (short) say_cnt_max) ;				
    tmp.SayCoachMsgSize = htons( (short) SayCoachMsgSize) ;				
    tmp.clang_win_size = htons( (short) clang_win_size );
    tmp.clang_define_win = htons( (short) clang_define_win );
    tmp.clang_meta_win = htons( (short) clang_meta_win );
    tmp.clang_advice_win = htons( (short) clang_advice_win );
    tmp.clang_info_win = htons( (short) clang_info_win );
    tmp.clang_mess_delay = htons( (short) clang_mess_delay );
    tmp.clang_mess_per_cycle = htons( (short) clang_mess_per_cycle );
    tmp.half_time = htons( (short) half_time) ;					
    tmp.sim_st = htons( (short) sim_st) ;					
    tmp.send_st = htons( (short) send_st) ;					
    tmp.recv_st = htons( (short) recv_st) ;					
    tmp.sb_step = htons( (short) sb_step) ;					
    tmp.lcm_st = htons( (short) lcm_st) ;		                        
    tmp.SayMsgSize = htons( (short) SayMsgSize) ;				
    tmp.hear_max = htons( (short) hear_max) ;					
    tmp.hear_inc = htons( (short) hear_inc) ;					
    tmp.hear_decay = htons( (short) hear_decay) ;				
    tmp.cban_cycle = htons( (short) cban_cycle) ;				
    tmp.slow_down_factor = htons( (short) slow_down_factor) ;                          
    tmp.useoffside = htons( (short) useoffside) ;				
    tmp.kickoffoffside = htons( (short) kickoffoffside) ;			
    tmp.offside_kick_margin = htonl( (long)(SHOWINFO_SCALE2 * offside_kick_margin) );		
    tmp.audio_dist = htonl( (long)(SHOWINFO_SCALE2 * audio_dist) );				
    tmp.dist_qstep = htonl( (long)(SHOWINFO_SCALE2 * dist_qstep) );				
    tmp.land_qstep = htonl( (long)(SHOWINFO_SCALE2 * land_qstep) );				
#ifdef NEW_QSTEP
    tmp.dir_qstep = htonl( (long)(SHOWINFO_SCALE2 * dir_qstep) );				
    tmp.dist_qstep_l = htonl( (long)(SHOWINFO_SCALE2 * dist_qstep_l) );			
    tmp.dist_qstep_r = htonl( (long)(SHOWINFO_SCALE2 * dist_qstep_r) );			
    tmp.land_qstep_l = htonl( (long)(SHOWINFO_SCALE2 * land_qstep_l) );			
    tmp.land_qstep_r = htonl( (long)(SHOWINFO_SCALE2 * land_qstep_r) );			
    tmp.dir_qstep_l = htonl( (long)(SHOWINFO_SCALE2 * dir_qstep_l) );				
    tmp.dir_qstep_r = htonl( (long)(SHOWINFO_SCALE2 * dir_qstep_r) );				
#else
    tmp.dir_qstep = htonl( (long)(SHOWINFO_SCALE2 * -1) );				
    tmp.dist_qstep_l = htonl( (long)(SHOWINFO_SCALE2 * -1) );			
    tmp.dist_qstep_r = htonl( (long)(SHOWINFO_SCALE2 * -1) );			
    tmp.land_qstep_l = htonl( (long)(SHOWINFO_SCALE2 * -1) );			
    tmp.land_qstep_r = htonl( (long)(SHOWINFO_SCALE2 * -1) );			
    tmp.dir_qstep_l = htonl( (long)(SHOWINFO_SCALE2 * -1) );				
    tmp.dir_qstep_r = htonl( (long)(SHOWINFO_SCALE2 * -1) );				
#endif    
    tmp.CoachMode = htons( (short) CoachMode) ;				
    tmp.CwRMode = htons( (short) CwRMode) ;					
    tmp.old_hear = htons( (short) old_hear) ;					
    tmp.sv_st = htons( (short) sv_st) ;					

    tmp.synch_mode   = htons( (short)synch_mode);//pfr:SYNCH
    tmp.synch_offset = htons( (short)synch_offset);//pfr:SYNCH
    tmp.synch_micro_sleep = htons( (short)synch_micro_sleep);//pfr:SYNCH

    tmp.start_goal_l = htons( (short)start_goal_l );
    tmp.start_goal_r = htons( (short)start_goal_r );

    tmp.fullstate_l = htons( (short)fullstate_l );
    tmp.fullstate_r = htons( (short)fullstate_r );

    tmp.drop_time = htons( (short) drop_time) ;					

    tmp.slowness_on_top_for_left_team = htonl( (long)(SHOWINFO_SCALE2 * slowness_on_top_for_left_team) );
    tmp.slowness_on_top_for_right_team = htonl( (long)(SHOWINFO_SCALE2 * slowness_on_top_for_left_team) );
    return tmp;
}


void ServerParam::getSettingsString ( char* buf, const double& ver )
{
    ostrstream ost(buf, MaxStringSize) ;

    ost << "(server_param "
	<< gwidth << " " 					
	<< inertia_moment << " " 	
	<< psize << " " 			
	<< pdecay << " " 			
	<< prand << " " 			
	<< pweight << " " 				
	<< pspeed_max << " " 		
	<< paccel_max << " " 		
	<< stamina_max << " " 			
	<< stamina_inc << " " 			
	<< recover_init << " " 		
	<< recover_dthr << " " 		
	<< recover_min << " " 			
	<< recover_dec << " " 			
	<< effort_init << " " 			
	<< effort_dthr << " " 			
	<< effort_min << " " 		
	<< effort_dec << " " 			
	<< effort_ithr << " " 			
	<< effort_inc << " " 			
	<< kick_rand << " "                             
	<< team_actuator_noise << " "                     
	<< prand_factor_l << " "                          
	<< prand_factor_r << " "                            
	<< kick_rand_factor_l << " "                       
	<< kick_rand_factor_r << " "                     
	<< bsize << " " 			
	<< bdecay << " " 			
	<< brand << " " 			
	<< bweight << " " 				
	<< bspeed_max << " " 			
	<< baccel_max << " " 			
	<< dprate << " " 			
	<< kprate << " " 			
	<< kmargin << " " 				
	<< ctlradius << " " 			
	<< ctlradius_width << " " 		
	<< maxp << " "				
	<< minp << " " 				
	<< maxm << " " 				
	<< minm << " " 				
	<< maxnm << " " 				
	<< minnm << " " 				
	<< maxn << " " 				
	<< minn << " " 				
	<< Rad2Deg(visangle) << " " 			
	<< visdist << " " 					
	<< windir << " " 					
	<< winforce << " " 				
	<< winang << " " 					
	<< winrand << " " 					
	<< kickable_area << " " 			
	<< catch_area_l << " " 			
	<< catch_area_w << " " 			
	<< catch_prob << " " 				
	<< goalie_max_moves << " "                  	
	<< ckmargin << " " 				
	<< offside_area << " " 			
	<< win_no << " " 					
	<< win_random << " " 				
	<< say_cnt_max << " " 				
	<< SayCoachMsgSize << " " 				
	<< clang_win_size << " "
	<< clang_define_win << " "
	<< clang_meta_win << " "
	<< clang_advice_win << " "
	<< clang_info_win << " "
	<< clang_mess_delay << " "
	<< clang_mess_per_cycle << " "
	<< half_time << " " 					
	<< sim_st << " " 					
	<< send_st << " " 					
	<< recv_st << " " 					
	<< sb_step << " " 					
	<< lcm_st << " " 		                        
	<< SayMsgSize << " " 				
	<< hear_max << " " 					
	<< hear_inc << " " 					
	<< hear_decay << " " 				
	<< cban_cycle << " " 				
	<< slow_down_factor << " "                           
	<< useoffside << " " 				
	<< kickoffoffside << " " 			
	<< offside_kick_margin << " " 		
	<< audio_dist << " " 				
	<< dist_qstep << " " 				
	<< land_qstep << " " 				
#ifdef NEW_QSTEP
	<< dir_qstep << " " 				
	<< dist_qstep_l << " " 			
	<< dist_qstep_r << " " 			
	<< land_qstep_l << " " 			
	<< land_qstep_r << " " 			
	<< dir_qstep_l << " " 				
	<< dir_qstep_r << " " 				
#else
	<< -1 << " " 			
	<< -1 << " " 			
	<< -1 << " " 			
	<< -1 << " " 			
	<< -1 << " " 			
	<< -1 << " " 				
	<< -1 << " " 				
#endif	
	<< CoachMode << " " 				
	<< CwRMode << " " 					
	<< old_hear << " " 					
	<< sv_st << " "

	<< start_goal_l << " "
	<< start_goal_r << " "

	<< fullstate_l << " "
	<< fullstate_r << " "

	<< drop_time;

    if ( ver >= 7.11 )
      {
	ost << " " << slowness_on_top_for_left_team
	    << " " << slowness_on_top_for_right_team;
      }

    ost << ")" << endl << ends;
}

