/* -*- Mode: C -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */


#ifndef _SERVER_PARAM_H_
#define _SERVER_PARAM_H_

//#include <sys/param.h> /* needed for htonl, htons, ... */
//#include <netinet/in.h>
#include <strstream>
using namespace std;

#include "paramreader.h"
#include "utility.h"
#include "types.h"

class ServerParam : public ParamReader
{
public:

    ServerParam ( const int& argc, const char * const *argv );
    virtual ~ServerParam ();

    server_params_t convertToStruct ();
    void getSettingsString ( char*, const double& ver = 7.0 );
    
protected:
    virtual void getOptions(const int& argc, const char * const *argv );

    
public:

    Value gwidth ;					/* goal width */
    Value inertia_moment ;			/* intertia moment for turn */
    Value psize ;					/* player size */
    Value pdecay ;					/* player decay */
    Value prand ;					/* player rand */
    Value pweight ;					/* player weight */
    Value pspeed_max ;				/* player speed max */
    // th 6.3.00
    Value paccel_max ;				/* player acceleration max */
    //
    Value stamina_max ;				/* player stamina max */
    Value stamina_inc ;				/* player stamina inc */
    Value recover_init ;			/* player recovery init */
    Value recover_dthr ;			/* player recovery decriment threshold */
    Value recover_min ;				/* player recovery min */
    Value recover_dec ;				/* player recovery decriment */
    Value effort_init ;				/* player dash effort init */
    Value effort_dthr ;				/* player dash effort decriment threshold */
    Value effort_min ;				/* player dash effrot min */
    Value effort_dec ;				/* player dash effort decriment */
    Value effort_ithr ;				/* player dash effort incriment threshold */
    Value effort_inc ;				/* player dash effort incriment */
    // pfr 8/14/00: for RC2000 evaluation
    Value kick_rand;                                /* noise added directly to kicks */
    int team_actuator_noise;                        /* flag whether to use team specific actuator noise */
    Value prand_factor_l;                           /* factor to multiple prand for left team */
    Value prand_factor_r;                           /* factor to multiple prand for right team */
    Value kick_rand_factor_l;                       /* factor to multiple kick_rand for left team */
    Value kick_rand_factor_r;                       /* factor to multiple kick_rand for right team */

    Value bsize ;					/* ball size */
    Value bdecay ;					/* ball decay */
    Value brand ;					/* ball rand */
    Value bweight ;					/* ball weight */
    Value bspeed_max ;				/* ball speed max */
    // th 6.3.00
    Value baccel_max;				/* ball acceleration max */
    //
    Value dprate ;					/* dash power rate */
    Value kprate ;					/* kick power rate */
    Value kmargin ;					/* kickable margin */
    Value ctlradius ;				/* control radius */
    Value ctlradius_width ;			/* (control radius) - (plyaer size) */
    Value maxp ;					/* max power */
    Value minp ;					/* min power */
    Value maxm ;					/* max moment */
    Value minm ;					/* min moment */
    Value maxnm ;					/* max neck moment */
    Value minnm ;					/* min neck moment */
    Value maxn ;					/* max neck angle */
    Value minn ;					/* min neck angle */
    Value visangle ;				/* visible angle */
    Value visdist ;					/* visible distance */
    Angle windir ;					/* wind direction */
    Value winforce ;				/* wind force */
    Value winang ;					/* wind angle for rand */
    Value winrand ;					/* wind force for force */
    Value kickable_area ;			/* kickable_area */
    Value catch_area_l ;			/* goalie catchable area length */
    Value catch_area_w ;			/* goalie catchable area width */
    Value catch_prob ;				/* goalie catchable possibility */
    int   goalie_max_moves;                 /* goalie max moves after a catch */	
    Value ckmargin ;				/* corner kick margin */
    Value offside_area ;			/* offside active area size */
    int win_no ;					/* wind factor is none */
    int win_random ;				/* wind factor is random */
    int portnum ;					/* port number */
    int coach_pnum ;				/* coach port number */
    int olcoach_pnum ;				/* online coach port number */
    int say_cnt_max ;				/* max count of coach SAY (freeform) */
    int SayCoachMsgSize ;				/* max length of coach SAY (freeform) */
    int clang_win_size;                         /* coach language: time window size */
    int clang_define_win;                       /* coach language: define messages per window */
    int clang_meta_win;                         /* coach language: meta messages per window */
    int clang_advice_win;                       /* coach language: advice messages per window */
    int clang_info_win;                         /* coach language: info messages per window */
    int clang_mess_delay;                       /* coach language: delay between receive and send */
    int clang_mess_per_cycle;                   /* coach language: number of messages to flush per cycle */
  
    int half_time ;					/* half time */
    int drop_time;                                  /* cycles for dropping
						       the ball after a free kick,
						       corner kick message and
						       noone kicking the ball.
						       0 means don't drop
						       automatically  */

    int sim_st ;					/* simulator step interval msec */
    int send_st ;					/* udp send step interval msec */
    int recv_st ;					/* udp recv step interval msec */
    int sb_step ;					/* sense_body interval step msec */
    int lcm_st ;		                        /* lcm of all the above steps msec */
    int SayMsgSize ;				/* string size of say message */
    int hear_max ;					/* player hear_capacity_max */
    int hear_inc ;					/* player hear_capacity_inc */
    int hear_decay ;				/* player hear_capacity_decay */
    int cban_cycle ;				/* goalie catch ban cycle */
    int slow_down_factor ;                          /* factor to slow down simulator and send intervals */
    int useoffside ;				/* flag for using off side rule */
    int kickoffoffside ;			/* flag for permit kick off offside */
    Value offside_kick_margin ;		/* offside kick margin */
    Value audio_dist ;				/* audio cut off distance */
    Value dist_qstep ;				/* quantize step of distance */
    Value land_qstep ;				/* quantize step of distance for landmark */
#ifdef NEW_QSTEP
    Value dir_qstep ;				/* quantize step of direction */
    Value dist_qstep_l ;			/* team right quantize step of distance */
    Value dist_qstep_r ;			/* team left quantize step of distance */
    Value land_qstep_l ;			/* team right quantize step of distance for landmark */
    Value land_qstep_r ;			/* team left quantize step of distance for landmark */
    Value dir_qstep_l ;				/* team left quantize step of direction */
    Value dir_qstep_r ;				/* team right quantize step of direction */
#endif	
    int	verbose ;					/* flag for verbose mode */

    char logfile[MAX_FILE_LEN] ;	/* server log file name */
    char recfile[MAX_FILE_LEN] ;	/* log file name for recording */
    int  rec_ver ;					/* version for recording */
    int  rec_log ;					/* flag for log recording */
    int  rec_msg_mode ;				/* flag for whether to record msg_mode structures */
    int  send_log ;					/* flag for log sending */
    // th 6.3.00
    int  log_times ;				/* flag for logging server cycle times in the log file */
    //
    char replay[MAX_FILE_LEN] ;		/* log file name for replay */

    int CoachMode ;				/* coach mode */
    int CwRMode ;					/* coach with referee mode */
    int old_hear ;					/* old format for hear command (coach) */
    int sv_st ;					/* online coach's look interval step */
    
    int synch_mode;                               /* pfr:SYNCH whether to run in synchronized mode */
    int synch_offset;                        /* pfr:SYNCH the offset from the start of the cycle to tell players to run */
    int synch_micro_sleep;                   /* pfr:SYNCH the # of microseconds to sleep while waiting for players */

  int start_goal_l;                           /* The starting score of the left team */
  int start_goal_r;                           /* The starting score of the right team */

  int fullstate_l;                             /* flag to send fullstate messages to left team; supresses visual info */
  int fullstate_r;                             /* flag to send fullstate messages to right team; supresses visual info */

  double slowness_on_top_for_left_team; /* Kinda self explanatory */
  double slowness_on_top_for_right_team; /* ditto */
};



#endif
